/**
	$Id: Agent.java 6253 2013-03-14 14:48:27Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TimerTask;

import net.sf.basedb.clients.jobagent.executors.ProcessJobExecutor;
import net.sf.basedb.clients.jobagent.handlers.CustomRequestHandler;
import net.sf.basedb.clients.jobagent.handlers.DefaultRequestHandler;
import net.sf.basedb.clients.jobagent.handlers.MultiProtocolRequestHandler;
import net.sf.basedb.clients.jobagent.handlers.SignalRequestHandler;
import net.sf.basedb.clients.jobagent.slotmanager.InternalSlotManager;
import net.sf.basedb.clients.jobagent.slotmanager.Slot;
import net.sf.basedb.clients.jobagent.slotmanager.SlotManager;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.JobAgent;
import net.sf.basedb.core.JobAgentSettings;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.SessionControl;

import net.sf.basedb.util.SocketUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.jobagent.JobAgentServerConnection;
import net.sf.basedb.util.jobagent.JobInfo;
import net.sf.basedb.util.jobagent.RequestHandler;
import net.sf.basedb.util.timer.Scheduler;

/**
	This is the actual job agent application. It delegates the actual
	checking of the job queue to the {@link JobQueueChecker} class.
	It includes a listener service that can be used for remote control
	of the agent. It is for example possible to send <code>start</code>,
	<code>stop</code> and <code>pause</code> requests.
	<p>
	This class is responsible for creating a {@link JobExecutor}
	object and to delegate the actual execution of a job to it. The agent keep
	track of the running jobs and makes sure that only the configured
	number of jobs are running at the same time.
	<p>
	The agent is configured at construction time with parameters from
	a {@link Properties} object.
	
	<table>
	<tr>
		<th>Parameter</th>
		<th>Default value</th>
		<th>Description</th>
	</tr>
	<tr>
		<td>agent.user</td>
		<td>-</td>
		<td>The username to use for logging in to BASE (required)</td>
	</tr>
	<tr>
		<td>agent.password</td>
		<td>-</td>
		<td>The password to use for logging in to BASE (required)</td>
	</tr>
	<tr>
		<td>agent.id</td>
		<td>-</td>
		<td rowspan="3">
			The external ID (required), name and description properties of the 
			corresponding {@link JobAgent} item in the BASE database
		</td>
	</tr>
	<tr>
		<td>agent.name</td>
		<td>-</td>
	</tr>
	<tr>
		<td>agent.description</td>
		<td>-</td>
	</tr>
	<tr>
		<td>agent.port</td>
		<td>47822</td>
		<td>The port the remote control service listener is listening to</td>
	</tr>
	<tr>
		<td>agent.remotecontrol</td>
		<td>-</td>
		<td>
			A comma-separated list of computer that are allowed remote control. It is recommended
			that the web server is put in this list. The local host is always allowed control
			and doesn't have to be in this list.
		</td>
	</tr>
	<tr>
		<td>agent.allowremote.stop</td>
		<td>false</td>
		<td>
			If <code>stop</code> requests should be allowed from remote hosts or not. 
			Note! A stop request shuts down the agent making it impossible to start it
			again using remote control.
		</td>
	</tr>
	<tr>
		<td>agent.allowremote.start</td>
		<td>true</td>
		<td>If <code>start</code> requests should be allowed from remote hosts or not</td>
	</tr>
	<tr>
		<td>agent.allowremote.pause</td>
		<td>true</td>
		<td>If <code>pause</code> requests should be allowed from remote hosts or not</td>
	</tr>
	<tr>
		<td>agent.request-handler.*</td>
		<td>-</td>
		<td>
			One or more entries that can be used to register custom remote control request handlers.
			The * should be replaced with the name of the protocol which means that requests
			should take the form: protocol://custom-data....
			<p>
			The value is the name of a class that implements the {@link CustomRequestHandler}
			interface. The implementation must provide a public no-argument constructor. The
			implementation must be thread-safe and be able to handle multiple requests at the
			time.
		</td>
	</tr>
	<tr>
		<td>agent.executor.class</td>
		<td>{@link ProcessJobExecutor}</td>
		<td>
			The name of a class that is responsible for starting the a job once
			the agent has determined that is allowed to be exected. The class must
			implement the {@link JobExecutor} interface and provide a public noargument
			constructor. Note that only one instance of this class exists for an agent.
			It must be thread-safe since the jobs are executed in parallel threads.
		</td>
	</tr>
	<tr>
		<td>agent.checkinterval</td>
		<td>30</td>
		<td>Number of seconds between checks to the database for new jobs.</td>
	</tr>
	<tr>
		<td>agent.shortest.slots</td>
		<td>1</td>
		<td>Number of slots to reserve for jobs that take &lt; 1 minute to execute</td>
	</tr>
	<tr>
		<td>agent.shortest.priority</td>
		<td>4</td>
		<td>The thread priority of jobs in this slot. See {@link Thread#setPriority(int)}.</td>
	</tr>
	<tr>
		<td>agent.short.slots</td>
		<td>1</td>
		<td>Number of slots to reserve for jobs that take &lt; 10 minute to execute</td>
	</tr>
	<tr>
		<td>agent.short.priority</td>
		<td>4</td>
		<td>The thread priority of jobs in this slot. See {@link Thread#setPriority(int)}.</td>
	</tr>
	<tr>
		<td>agent.medium.slots</td>
		<td>2</td>
		<td>Number of slots to reserve for jobs that take &lt; 1 hour to execute</td>
	</tr>
	<tr>
		<td>agent.medium.priority</td>
		<td>3</td>
		<td>The thread priority of jobs in this slot. See {@link Thread#setPriority(int)}.</td>
	</tr>
	<tr>
		<td>agent.long.slots</td>
		<td>2</td>
		<td>Number of slots to reserve for jobs that take &gt; 1 hour to execute</td>
	</tr>
	<tr>
		<td>agent.long.priority</td>
		<td>3</td>
		<td>The thread priority of jobs in this slot. See {@link Thread#setPriority(int)}.</td>
	</tr>
	</table>

	@author nicklas
	@version 2.0
	@base.modified $Date: 2013-03-14 15:48:27 +0100 (Thu, 14 Mar 2013) $
*/
public class Agent
{

	/**
		The default job executor to use if none has been specified
		in the configuration file.
	*/
	public static final Class<? extends JobExecutor> DEFAULT_JOB_EXECUTOR = 
		ProcessJobExecutor.class;
	
	/**
		The default slot manager to use if none has been specified
		in the configuration file.
	*/
	public static final Class<? extends SlotManager> DEFAULT_SLOT_MANAGER = 
		InternalSlotManager.class;
	
	/**
		The default check interval in seconds.
	*/
	public static final int DEFAULT_CHECK_INTERVAL = 30;
	
	/**
		Log job agent events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.Agent");

	/**
		Log job agent server events.
	*/
	private static final org.apache.log4j.Logger logServer = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.JobAgentServerConnection");
	
	private final Properties properties;
	
	// Base settings
	private final String login;
	private final String password;
	private final String externalId;
	private final String name;
	private final String description;

	// Service listener settings
	private final Integer port;
	private final Set<InetAddress> remote;
	private final boolean allowRemoteStop;
	private final boolean allowRemotePause;
	private final boolean allowRemoteStart;
	private final Map<String, Class<? extends CustomRequestHandler>> customRequestHandlerClasses;
	
	// Job execution settings
	private final long checkInterval;
	private final Class<? extends JobExecutor> executorClass;
	
	// Slot manager
	private final Class<? extends SlotManager> slotManagerClass;
	
	private InetAddress serverAddress;
	private JobAgentServerConnection server;
	private MultiProtocolRequestHandler requestHandler;
	private AgentSignalReceiver signalReceiver;
	private JobExecutor jobExecutor;
	private SlotManager slotManager;

	private TimerTask jobQueueChecker;
	
	
	private boolean isRunning;
	private SessionControl sc;
	
	/**
		The thread priority to use when executing jobs in each slot.
	*/
	private final Map<Job.ExecutionTime, Integer> priorities;
	
	private final Set<JobInfo> activeJobs;

	/**
		Timeout to wait for jobs to act on the ABORT signal when stopping.
		Default value is 20 seconds.
	*/
	private int closeTimeout = 20000;

	/**
		The group were all job runners are placed.
	*/
	private final ThreadGroup runnersGroup;

	/**
		Create a new job agent. See class documentation for information about
		configuration parameters.

		@param properties A properties object containing the configuration 
			parameters for the agent
	*/
	public Agent(Properties properties)
	{
		this.properties = properties;
		
		if (log.isDebugEnabled())
		{
			log.debug("Creating new agent");
			for (Map.Entry<Object, Object> entry : properties.entrySet())
			{
				Object value = entry.getValue();
				Object key = entry.getKey();
				if ("agent.password".equals(key))
				{
					value = "******************";
				}
				log.debug("Property [" + key + " = " + value + "]");
			}
		}

		// BASE settings
		this.login = properties.getProperty("agent.user");
		this.password = properties.getProperty("agent.password");
		this.externalId = properties.getProperty("agent.id");
		this.name = properties.getProperty("agent.name");
		this.description = properties.getProperty("agent.description");

		// Listener service settings
		this.port = Values.getInt(properties.getProperty("agent.port"), JobAgent.DEFAULT_PORT);
		this.remote = getRemoteAddresses(properties.getProperty("agent.remotecontrol"));
		this.allowRemoteStop = Values.getBoolean(properties.getProperty("agent.allowremote.stop"), false);
		this.allowRemotePause = Values.getBoolean(properties.getProperty("agent.allowremote.pause"), true);
		this.allowRemoteStart = Values.getBoolean(properties.getProperty("agent.allowremote.start"), true);
		
		// Additional request handlers
		this.customRequestHandlerClasses = getCustomRequestHandlerClasses();
		
		// Job execution settings
		this.checkInterval = Values.getInt(properties.getProperty("agent.checkinterval"), DEFAULT_CHECK_INTERVAL);
		this.executorClass = getJobExecutorClass(properties.getProperty("agent.executor.class"));
		
		// Slot manager
		this.slotManagerClass = getSlotManagerClass(properties.getProperty("agent.slotmanager.class"));
		
		// Slots and priorities
		this.priorities = new HashMap<Job.ExecutionTime, Integer>();
		for (Job.ExecutionTime et : Job.ExecutionTime.values())
		{
			String configName = "agent."+et.name().toLowerCase();
			int priority = Values.getInt(properties.getProperty(configName+".priority"), et.getDefaultPriority());
			priorities.put(et, priority);
		}

		validateConfiguration();
		
		// Configure the job runners thread group
		runnersGroup = new ThreadGroup("Agent."+externalId);
		runnersGroup.setDaemon(false);

		this.activeJobs = new HashSet<JobInfo>();
		isRunning = false;
	}
	
	/**
		Get a configuration property.
		@param key The key
		@return The configured value or null if not found
	*/
	public String getProperty(String key)
	{
		return properties.getProperty(key);
	}
	
	/**
		Get the <code>agent.id</code> configuration value.
		@return The ID
	*/
	public String getId()
	{
		return externalId;
	}
	
	/**
		Get the <code>agent.user</code> configuration value.
		@return The login
	*/
	public String getLogin()
	{
		return login;
	}
	
	/**
		Get the <code>agent.password</code> configuration value.
		@return The password
	*/
	public String getPassword()
	{
		return password;
	}
	
	/**
		Get the <code>agent.name</code> configuration value.
		@return The name or the ID if the name is null
	*/
	public String getName()
	{
		return name == null ? externalId : name;
	}
	
	/**
		Get the <code>agent.description</code> configuration value.
		@return The description
	*/
	public String getDescription()
	{
		return description;
	}
	
	/**
		Get the <code>agent.port</code> configuration value.
		@return The port number
	*/
	public int getPort()
	{
		return port;
	}
	
	/**
		Get the thread priority to use for the specified execution time slot.
		@param slot
	 	@return Priority as an int. High number shows a high priority.
	*/
	public int getThreadPriority(Job.ExecutionTime slot)
	{
		return priorities.get(slot);
	}
	
	/**
		Get the host name of the server where the job agent is running.
	 	@return Server's name as a String. 
		@see SocketUtil#getPublicLocalHost()
	*/
	public String getServerName()
	{
		if (serverAddress == null)
		{
			serverAddress = SocketUtil.getPublicLocalHost();
		}
		return serverAddress.getHostName();
	}
	
	/**
		Split the string at commas and try to create an {@link InetAddress}
		for each part. Parts that turn out to be invalid are skipped and 
		a warning message is logged.
		@param config The string to split
		@return A set of <code>InetAdress</code> object
	*/
	private Set<InetAddress> getRemoteAddresses(String config)
	{
		Set<InetAddress> remote = new HashSet<InetAddress>();
		if (config != null)
		{
			for (String host : config.split(","))
			{
				try
				{
					remote.add(InetAddress.getByName(host));
				}
				catch (UnknownHostException ex)
				{
					log.warn("Unknown host: " + host, ex);
				}
			}
		}
		return remote;
	}
	
	@SuppressWarnings("unchecked")
	private Map<String, Class<? extends CustomRequestHandler>> getCustomRequestHandlerClasses()
	{
		log.debug("Loading custom request handler classes");
		Map<String, Class<? extends CustomRequestHandler>> handlers = 
			new HashMap<String, Class<? extends CustomRequestHandler>>();
		
		for (String property : properties.stringPropertyNames())
		{
			if (property.startsWith("agent.request-handler."))
			{
				String protocol = property.substring("agent.request-handler.".length());
				String className = properties.getProperty(property);
				try
				{
					Class<? extends CustomRequestHandler> clazz = (Class<? extends CustomRequestHandler>)Class.forName(className);
					if (!CustomRequestHandler.class.isAssignableFrom(clazz))
					{
						log.warn("Class '" + className + "' doesn't implement the CustomRequestHandler interface, ignored.");
					}
					else
					{
						log.info("Loaded custom request handler for protocol '" + protocol + "': " + clazz.getName());
						handlers.put(protocol, clazz);
					}
				}
				catch (Throwable t)
				{
					log.warn("Custom request handler class '" + className + "' not found, ignored.");
				}
			}
		}
		log.debug("Custom request handler classes lodaed: " + handlers.size());
		return handlers;
	}
	
	private void registerCustomRequestHandlers(MultiProtocolRequestHandler master)
	{
		Map<Class<? extends CustomRequestHandler>, CustomRequestHandler> created = 
			new HashMap<Class<? extends CustomRequestHandler>, CustomRequestHandler>();
		for (Map.Entry<String, Class<? extends CustomRequestHandler>> entry : customRequestHandlerClasses.entrySet())
		{
			String protocol = entry.getKey();
			Class<? extends CustomRequestHandler> clazz = entry.getValue();
			CustomRequestHandler handler = created.get(clazz);
			if (handler == null)
			{
				try
				{
					handler = clazz.newInstance();
					handler.init(this);
					created.put(clazz, handler);
					log.info("Created request handler for protocol '" + protocol + ": " + clazz.getName());
				}
				catch (Throwable t)
				{
					log.error("Could not create request handler for protocol '" + protocol + ": " + clazz.getName(), t);
				}
			}
			master.registerProtocols(handler, protocol);
		}
	}
	
	/**
		Get the class object for the configured job executor. If the 
		specified class can't be found or doesn't implement the 
		{@link JobExecutor} interface a warning message is logged and 
		the {@link #DEFAULT_JOB_EXECUTOR} is used instead.
		
		@param className The name of the job executor class
		@return The class object for that class or the default job executor
	*/
	@SuppressWarnings("unchecked")
	private Class<? extends JobExecutor> getJobExecutorClass(String className)
	{
		Class<? extends JobExecutor> executor = DEFAULT_JOB_EXECUTOR;
		try
		{
			if (className != null)
			{
				executor = (Class<JobExecutor>)Class.forName(className);
				if (!JobExecutor.class.isAssignableFrom(executor))
				{
					log.warn("Class '" + className + "' doesn't implement the JobExecutor interface, using '" + 
						DEFAULT_JOB_EXECUTOR.getName() + "' instead");
					executor =  DEFAULT_JOB_EXECUTOR;
				}
			}
		}
		catch (Throwable t)
		{
			log.warn("Job executor class '" + className + "' not found, using '" + DEFAULT_JOB_EXECUTOR.getName() + 
				"' instead", t);
		}
		log.info("Job executor class '" + executor.getName() + "' loaded successfully");
		return executor;
	}
	
	/**
		Get the class object for the configured slot manager. If the 
		specified class can't be found or doesn't implement the 
		{@link SlotManager} interface a warning message is logged and 
		the {@link #DEFAULT_SLOT_MANAGER} is used instead.
		
		@param className The name of the slot manager class
		@return The class object for that class or the default job executor
	*/
	@SuppressWarnings("unchecked")
	private Class<? extends SlotManager> getSlotManagerClass(String className)
	{
		Class<? extends SlotManager> manager = DEFAULT_SLOT_MANAGER;
		try
		{
			if (className != null)
			{
				manager = (Class<SlotManager>)Class.forName(className);	
			}
			if (!SlotManager.class.isAssignableFrom(manager))
			{
				log.warn("Class '" + className + "' doesn't implement the SlotManager interface, using '" + 
					DEFAULT_SLOT_MANAGER.getName() + "' instead");
				manager =  DEFAULT_SLOT_MANAGER;
			}
		}
		catch (Throwable t)
		{
			log.warn("Slot manager class '" + className + "' not found, using '" + DEFAULT_SLOT_MANAGER.getName() + 
				"' instead", t);
		}
		log.info("Slot manager class '" + manager.getName() + "' loaded successfully");
		return manager;
	}

	
	/**
		Validate that all required configuration parameters have been specified.
		@throws InvalidDataException If parameters are missing or have incorrect values
	*/
	private void validateConfiguration()
		throws InvalidDataException
	{
		if (login == null) throw new InvalidUseOfNullException("agent.user");
		if (password == null) throw new InvalidUseOfNullException("agent.password");
		if (externalId == null) throw new InvalidUseOfNullException("agent.id");
	}

	/**
		Start the listener service that listens for control commands such
		as <code>start</code>, <code>stop</code>, <code>pause</code> and
		<code>info</code>. The listener service is only required if the job
		agent needs to respond to remote control commands. 
		<p>
		Note! The {@link AgentController} which is the default controller for
		agents always uses the listener service for communicating with the job
		agent.
		<p>
		Note! The listener service is started in a separate thread and this method
		returns as soon as the network connections are set up.
		<p>
		Note! The default handler supplied as an argument is used as a fallback 
		handler for unregistered protocols. Additional request handlers can be
		set up by calling {@link #registerRequestHandler(RequestHandler, String...)}.
		
		@param defaultHandler A {@link RequestHandler} that handles the 
			incoming requsts, or null to use the {@link DefaultRequestHandler}
		@throws IOException If there is an error when starting the service
	*/
	public synchronized void service(RequestHandler defaultHandler)
		throws IOException
	{
		log.info("Starting listener service for job agent on port " + port);
		if (server == null)
		{
			this.signalReceiver = new AgentSignalReceiver(this);
			this.signalReceiver.init(null);
			if (defaultHandler == null) defaultHandler = new DefaultRequestHandler(this);
			this.requestHandler = new MultiProtocolRequestHandler(this, defaultHandler);
			registerRequestHandler(new SignalRequestHandler(this), "signal");
			registerCustomRequestHandlers(requestHandler);
			this.server = new JobAgentServerConnection(port, this.requestHandler, logServer);
			server.open();
		}
	}
	
	/**
		Start the job agent. This method will register a {@link JobQueueChecker}
		object with the BASE core scheduler {@link Application#getScheduler()}. The 
		timer will call the {@link JobQueueChecker#run()} method at intervals
		specified by the <code>agent.checkinterval</code> configuration settings.
		<p>
		Note! The <code>JobQueueChecker</code> will run in a separate thread
		and this method return immediately after registering the object with the
		scheduler.
		<p>
		Note! This method also creats a single instance of a {@link JobExecutor}.
		The actual class to use is specified by the <code>agent.jobexecutor.class</code>
		configuration setting. The default job executor is {@link ProcessJobExecutor}.
	*/
	public synchronized void start()
	{
		log.info("Starting job agent on port " + port);
		if (!isRunning)
		{
			// Start BASE application and login
			getSessionControl();;
			// Create executor and job queue checker
			slotManager = createSlotManager();
			jobExecutor = createJobExecutor();
			jobQueueChecker = createJobQueueChecker();
			isRunning = true;
		}
	}
	
	/**
		Stop the job agent. This method will:

		<ul>
		<li>Cancel the {@link JobQueueChecker} that was registered with the BASE 
			core scheduler by the {@link #start()} method.
		<li>Call the {@link JobExecutor#close()} method on the job executor 
			created by the {@link #start()} method.
		<li>Close the service listener started by the {@link #service(RequestHandler)}
			method.
		<li>Try to stop all running jobs by calling {@link Thread#interrupt()}
			on all job threads.
		<li>Logout from BASE.
		</ul>
		
		Unless no other things are running in the same virtual machine as this 
		job agent the virtual machine should exit as a result from calling this 
		method.
	*/
	public synchronized void stop()
	{
		log.info("Stopping job agent on port " + port);
		isRunning = false;
		closeJobQueueChecker();
		maybeStopRunningJobs();
		closeJobExecutor();
		closeSlotManager();
		closeServer();
		if (sc != null) sc.logout();
	}
	
	/**
		Pause the job agent. This method will:

		<ul>
		<li>Cancel the {@link JobQueueChecker} that was registered with the BASE 
			core scheduler by the {@link #start()} method.
		<li>Call the {@link JobExecutor#close()} method on the job executor 
			created by the {@link #start()} method.
		<li>Logout and stop the BASE {@link Application}.
		</ul>
		
		This method will not try to stop the running jobs or shut down the
		BASE application. Calling {@link #start()} again will start the job
		agent again.
		
		@see #stop()
		@see #start()
	*/
	public synchronized void pause()
	{
		log.info("Pausing job agent on port " + port);
		isRunning = false;
		closeJobQueueChecker();
		closeJobExecutor();
		if (sc != null) sc.logout();
	}
	
	/**
		Check if the job agent is running or not.
		@return TRUE if the job agent is running, FALSE otherwise
	*/
	public boolean isRunning()
	{
		return isRunning;
	}
	
	/**
		Get a set containing the ID:s of the jobs that are currently
		beeing executed by this job agent.
		@return A set of JobInfo objects
		@since 2.6
	*/
	public Set<JobInfo> getRunningJobs()
	{
		if (log.isDebugEnabled()) log.debug("Active jobs: " + activeJobs);
		return Collections.unmodifiableSet(activeJobs);
	}
	
	/**
		Check if the computer specified by the given address is allowed to
		control this job agent. A computer is allowed to control this
		job agent if it's name or ip-address is listed in the 
		<code>agent.remotecontrol</code> property. This method is called from
		the {@link DefaultRequestHandler#handleCmd(java.net.Socket, String)} method
		to determine if a service request should be accepted or not.
		<p>
		Note! The <code>stop</code>, <code>start</code> and <code>pause</code>
		commands are only allowed from the local host unless the <code>agent.allowremote.stop</code>,
		<code>agent.allowremote.start</code> and <code>agent.allowremote.pause</code> 
		are set to a true values.
		<p>
		Note! The local host doesn't have to be listed in the <code>agent.remotecontrol</code>
		property. Requests are always allowed from the local host.
		
		@param remote The address to the remote computer
		@param cmd The command the remote computer wants to execute
		@return TRUE if the computer is allowed, FALSE otherwise
	*/
	public boolean isAllowedControl(InetAddress remote, String cmd)
	{
		boolean allow = false;
		boolean isLocal = SocketUtil.isLocalHost(remote);
		if ("stop".equals(cmd) && !allowRemoteStop)
		{
			allow = isLocal;
		}
		else if ("start".equals(cmd) && !allowRemoteStart)
		{
			allow = isLocal;
		}
		else if ("pause".equals(cmd) && !allowRemotePause)
		{
			allow = isLocal;
		}
		else 
		{
			allow = isLocal || this.remote.contains(remote);
		}
		return allow;
	}
	
	/**
		Register a request handler for one or more protocols. See
		{@link MultiProtocolRequestHandler} for more information.
		Calls to this method is ignored if the job agent has no listener
		service. See {@link #service(RequestHandler)}.
		
		@param handler The request handler
		@param protocols The name of the protocols the handler should handle
		@since 2.16
	*/
	public void registerRequestHandler(RequestHandler handler, String... protocols)
	{
		if (this.requestHandler == null) return;
		if (handler == null) throw new NullPointerException("handler");
		if (log.isDebugEnabled())
		{
			log.debug("Registering protocol handler " + handler + " for protocols: " + Arrays.asList(protocols));
		}
		requestHandler.registerProtocols(handler, protocols);
	}

	/**
		Unregister one or more protocols. See
		{@link MultiProtocolRequestHandler} for more information.
		Calls to this method is ignored if the job agent has no listener
		service. See {@link #service(RequestHandler)}.
		
		@param protocols The name of the protocols the handler should handle
		@since 2.16
	*/
	public void unregisterRequestHandler(String... protocols)
	{
		if (this.requestHandler == null) return;
		if (log.isDebugEnabled())
		{
			log.debug("Unregistering protocol handler for protocols: " + Arrays.asList(protocols));
		}
		requestHandler.unregisterProtocols(protocols);
	}
	
	/**
		Get the signal receiver that is processing signal messages
		on behalf of this job agent.
		@since 2.6
	*/
	public AgentSignalReceiver getSignalReceiver()
	{
		return signalReceiver;
	}
	
	/**
		Get a session control with the configured user logged in.
		@return A session control
	*/
	public SessionControl getSessionControl()
	{
		if (sc == null)
		{
			log.info("Starting BASE application");
			// Disable database cleanup script, it is only needed on the web client
			Config.setProperty("db.cleanup.interval", "0");
			Application.start(false);
			sc = Application.newSessionControl("net.sf.basedb.clients.jobagent", 
				SocketUtil.getLocalHost().toString(), null);
		}
		if (!sc.isLoggedIn())
		{
			log.info("Logging in as user: " + login);
			sc.login(login, password, "Job agent running on host " + getServerName());
		}
		return sc;
	}
	
	/**
		Get a session control where the owner of the job has been impersonated and
		the active project has been set if needed.
		@param job The job to get the impersonated session control for
		@return A session control object
	*/
	public SessionControl getImpersonatedSessionControl(Job job)
	{
		SessionControl sc = getSessionControl();
		SessionControl impersonated = null;
		DbControl dc = null;
		try
		{
			// Reload job and impersonate the owner
			dc = sc.newDbControl();
			job = Job.getById(dc, job.getId());
			impersonated = sc.impersonateLogin(job, "Running job: " + job.getName());
			dc.close();
			
			// Set the active project if any
			int projectId = job.getActiveProjectId();
			if (projectId != 0)
			{
				try
				{
					dc = impersonated.newDbControl();
					Project activeProject = Project.getById(dc, projectId);
					impersonated.setActiveProject(activeProject);
					dc.close();
				}
				catch (Throwable t)
				{
					log.warn("Exception while setting active project to Project[id=" + projectId + 
						"]. Continuing with no active project.", t);
				}
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return impersonated;
	}
	
	/**
		Get the {@link JobAgent} item corresponding to this agent. The
		job agent is looked up by the {@link JobAgent#getByExternalId(DbControl, String)}
		method where the external ID is given by the <code>agent.id</code> property.
		@param dc The DbControl to use for database acces
		@return A JobAgent item
	*/
	public JobAgent getJobAgent(DbControl dc)
	{
		return JobAgent.getByExternalId(dc, externalId);
	}
	
	/**
		Find a free slot for executing a job. If there is a free slot
		in the requested pool the requested slot is returned, otherwise
		the method checks the slower-running pools for free slots
		and returns one of those. If there are no free slot available
		null is returned.
		<p>
		Note! This method reserves the slot for the job. It is important that
		the {@link #jobDone(Job, Slot)} method is called once
		the job has completed to return the slot to the pool. Failure to do
		so may result in that the agent thinks that all slots are
		used when they are not.
		
		@param job The job that wants a slot
		@return The assigned slot or null if no slot is available
	*/
	synchronized Slot getSlot(Job job)
	{
		log.debug("Requesting slot for job: " + job);
		Slot slotToUse = slotManager.getSlot(job);

		if (slotToUse != null)
		{
			log.debug("Got slot: " + slotToUse);
			activeJobs.add(new JobInfo(job, slotToUse.getEstimatedExecutionTime()));
		}
		else
		{
			log.debug("No available slot");
		}

		// If null we couldn't find a free slot
		return slotToUse;
	}

	/**
		Start a job. This method will create a new thread for running the job
		and return immediately. If the job for some reason can't be started,
		for example, if there are no available slots, a log messsage will be
		written but nothing else will happen. No exception will be thrown to
		the caller of this method.		
		@param job The job to start
	 	@param settings Settings for the job agent to use. 
	*/
	public synchronized void startJob(Job job, JobAgentSettings settings)
	{
		JobRunner runner = new JobRunner(this, job, settings, jobExecutor);
		Thread t = new Thread(runnersGroup, runner);
		t.setDaemon(false);
		t.setPriority(priorities.get(job.getEstimatedExecutionTime()));
		t.start();
	}

	/**
		Used by {@link JobRunner} to tell that a job has finished executing and
		that the used slot should be released.
		
		@param job The job that has finished
		@param usedSlot The slot that was used
	*/
	synchronized void jobDone(Job job, Slot usedSlot)
	{
		if (slotManager != null)
		{
			slotManager.releaseSlot(usedSlot);
			activeJobs.remove(new JobInfo(job, usedSlot.getEstimatedExecutionTime()));
		}
	}
	
	/**
		Close the service listener.
	*/
	private synchronized void closeServer()
	{
		log.info("Closing service listener: " + server);
		if (server != null) 
		{
			server.close();
			server = null;
		}
		if (requestHandler != null)
		{
			requestHandler.close();
			requestHandler = null;
		}
	}
	
	/**
		Create a job queue checker. and register it with the BASE core
		scheduler.
		@see Application#getScheduler()
		@see JobQueueChecker
	*/
	private TimerTask createJobQueueChecker()
	{
		log.info("Creating job queue checker; checkInterval=" + checkInterval + " s");
		Scheduler scheduler = Application.getScheduler();
		return scheduler.schedule(
			new JobQueueChecker(this), 1000*checkInterval, 1000*checkInterval, false);
	}
	
	/**
		Close the job queue checker. 
	*/
	private synchronized void closeJobQueueChecker()
	{
		log.info("Closing job queue checker: " + jobQueueChecker);
		if (jobQueueChecker != null)
		{
			jobQueueChecker.cancel();
			jobQueueChecker = null;
		}
	}

	/**
		Create a job executor and initialise it.
		@return A JobExecutor instance or null if none could be created
	*/
	private JobExecutor createJobExecutor()
	{
		log.info("Creating job executor: " + executorClass.getName());
		JobExecutor executor = null;
		try
		{
			executor = executorClass.newInstance();
			executor.init(this);
		}
		catch (RuntimeException t)
		{
			log.error("Could not create job executor: " + executorClass.getName(), t);
			executor = null;
			throw t;
		}
		catch (Exception e)
		{
			log.error("Could not create job executor: " + executorClass.getName(), e);
			executor = null;
			throw new RuntimeException(e);
		}
		return executor;
	}
	

	/**
		Close the job executor.
	*/
	private synchronized void closeJobExecutor()
	{
		log.info("Closing job executor: " + jobExecutor);
		if (jobExecutor != null)
		{
			jobExecutor.close();
			jobExecutor = null;
		}
	}
	
	
	/**
		Create a slot manager and initialize it.
		@return A SlotManager instance
		@since 2.16
	*/
	private SlotManager createSlotManager()
	{
		log.info("Creating slot manager: " + slotManagerClass.getName());
		SlotManager manager = null;
		try
		{
			manager = slotManagerClass.newInstance();
			manager.init(this);
		}
		catch (RuntimeException t)
		{
			log.error("Could not create slot manager: " + slotManagerClass.getName(), t);
			manager = null;
			throw t;
		}
		catch (Exception e)
		{
			log.error("Could not create slot manager: " + slotManagerClass.getName(), e);
			manager = null;
			throw new RuntimeException(e);
		}
		return manager;
	}

	/**
		Close the slot manager.
	*/
	private synchronized void closeSlotManager()
	{
		log.info("Closing slot manager: " + slotManager);
		if (slotManager != null)
		{
			slotManager.close();
			slotManager = null;
		}
	}

	/**
		Try to stop running jobs by interrupting the threads they are executing in.
	*/
	private synchronized void maybeStopRunningJobs()
	{
		log.info("Stopping running jobs. " + activeJobs.size() + " job(s) still active.");
		
		// Send SHUTDOWN/ABORT to all jobs, that support signals
		signalReceiver.close(closeTimeout);
		signalReceiver = null;
		unregisterRequestHandler("signal");
	}
	
	
}
