/*
	$Id: ExtensionsControl.java 6028 2012-03-26 13:42:47Z nicklas $

	Copyright (C) Nicklas Nordborg
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.File;
import java.net.URI;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.PageContext;

import net.sf.basedb.clients.web.extensions.edit.EditUtil;
import net.sf.basedb.clients.web.extensions.list.ListColumnUtil;
import net.sf.basedb.clients.web.extensions.service.Services;
import net.sf.basedb.clients.web.extensions.toolbar.ToolbarUtil;
import net.sf.basedb.clients.web.servlet.ExtensionsServlet;
import net.sf.basedb.core.AccessControlled;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Client;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.extensions.ErrorHandlerFactory;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.events.EventType;
import net.sf.basedb.util.extensions.manager.ExtensionKey;
import net.sf.basedb.util.extensions.manager.ExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.ObjectKey;
import net.sf.basedb.util.extensions.manager.ProcessResults;
import net.sf.basedb.util.extensions.manager.Settings;
import net.sf.basedb.util.extensions.manager.filter.InstalledFilter;
import net.sf.basedb.util.extensions.manager.processor.DeleteResourcesProcessor;
import net.sf.basedb.util.extensions.manager.processor.DisablePluginsProcessor;
import net.sf.basedb.util.extensions.manager.processor.ExtractResourcesProcessor;
import net.sf.basedb.util.extensions.manager.processor.PluginInstallationProcessor;
import net.sf.basedb.util.extensions.manager.processor.UnregisterExtensionsProcessor;
import net.sf.basedb.util.extensions.xml.ExtensionPointFilter;
import net.sf.basedb.util.extensions.xml.PluginInfo;
import net.sf.basedb.util.filter.Filter;

/**
	Controller class for the extension system in the web client. This is the 
	main class to use when working with extensions from the JSP code. 
	<p>
	
	To get all registered extensions for an extension point call 
	{@link #useExtensions(JspContext, String...)}.
	<p>

	Use {@link #get(DbControl)} to access the management API wich allows
	you to enable/disable extensions and change settings for the extension
	system.
	
	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class ExtensionsControl
	implements AccessControlled
{

	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.web.extensions.ExtensionsControl");
	
	/**
		The location where the extensions resources are extracted.
	*/
	public static final String RESOURCES_URL = "/extensions";

	/**
		The location where extension servlets are registered.
		@since 2.13
	*/
	public static final String SERVLET_URL = RESOURCES_URL + "/servlet";
	
	/**
		This event is sent before an extension or extension point is
		disabled by {@link #enableExtensionPoint(String, boolean)},
		{@link #enableExtension(String, boolean)}, or 
		{@link #enableAllInFile(String, boolean)}.
	*/
	public static final EventType BEFORE_DISABLE = new EventType("Before disable");
	
	/**
		This event is sent after an extension or extension point is
		enabled by {@link #enableExtensionPoint(String, boolean)},
		{@link #enableExtension(String, boolean)}, or 
		{@link #enableAllInFile(String, boolean)}.
	*/
	public static final EventType AFTER_ENABLE = new EventType("After enable");
	
	// If the extension system has been initialised or not
	private static boolean initialised = false;
	
	// The main extensions manager, registry and settings
	private static ExtensionsManager manager;
	private static Registry registry;
	private static Settings settings;
	
	// The Servlet Context provided by the web server
	private static ServletContext servletContext;
		
	// The result of the last automatic update
	private static ProcessResults lastScanResults;
		
	/**
		Initialise the extension system. This method is called once from the
		{@link ExtensionsServlet} when the web server starts up.
	*/
	public static synchronized void init(ServletContext context)
	{
		if (initialised) return;
		log.info("Initialising extensions controller");
		servletContext = context;
		
		// Hook into core manager and registry
		manager = Application.getExtensionsManager();
		registry = manager.getRegistry();
		settings = manager.getSettings();
		
		// Register some extension points programmatically. This
		// must be done before scanning files otherwise it is not
		// possible to register extensions for for them
		ToolbarUtil.registerExtensionPoints(registry);  // Toolbars
		EditUtil.registerExtensionPoints(registry); // TabControls
		ListColumnUtil.registerExtensionPoints(registry); // List columns
		
		// Add default web-client extension points and extensions
		manager.addURI(URI.create(ExtensionsControl.class.getResource("/web-extensions.xml").toString()));

		// Install extensions
		lastScanResults = new ProcessResults();
		
		// Initial scan
		scan(null, null);
		initialised = true;
		
		Services.init();
		log.info("Extensions controller initialised successfully");
	}
	
	/**
		Shut down the extensions system.
		@since 2.8
	*/
	public static synchronized void close()
	{
		if (!initialised) return;
		log.info("Closing extensions controller");
		Services.close();
		registry = null;
		servletContext = null;
		lastScanResults = null;
		initialised = false;
		log.info("Extensions controller closed successfully");
	}
	
	
	private static ProcessResults scan(DbControl dc, HttpServletRequest request)
	{
		log.info("Starting scan");
		String rootPath = servletContext.getContextPath();
		File resourceDir = new File(servletContext.getRealPath(RESOURCES_URL));
		boolean initialScan = request == null;

		// Collect the results from this scan
		ProcessResults results = new ProcessResults();
		
		// File statistics
		int numInstalled = 0;
		int numUninstalled = 0;
		
		// Extensions+extension points statistics
		int numUnregistered = 0;
		int numRegistered = 0;
		int numInstalledPlugins = 0;
		int numDisabledPlugins = 0;
		
		synchronized (manager)
		{
			if (!initialScan)
			{
				// 1. Process files that has been selected for uninstallation
				log.debug("Uninstalling extensions");
				UninstallFilter uninstallFilter = new UninstallFilter(request);
				
				// 1a. Remove resources extracted from JAR files
				DeleteResourcesProcessor deleteResources = new DeleteResourcesProcessor(resourceDir, results);
				manager.processFiles(deleteResources, uninstallFilter);
				
				// 1b. Unregister all extensions from deleted files
				UnregisterExtensionsProcessor unregister = new UnregisterExtensionsProcessor(results);
				manager.processFiles(unregister, uninstallFilter);
				numUnregistered = unregister.getNumUnregistered();
				
				// 1c. Mark all plug-ins as disabled
				DisablePluginsProcessor disablePlugins = new DisablePluginsProcessor(dc, results);
				manager.processFiles(disablePlugins, uninstallFilter);
				numDisabledPlugins = disablePlugins.getNumPlugins();
				
				// 1d. Mark files as uninstalled and remove from the manager
				UninstalledFileProcessor uninstall = new UninstalledFileProcessor(results);
				manager.processFiles(uninstall, uninstallFilter);
				numUninstalled = uninstall.getNumFiles();
				log.debug("Uninstalled " + numUninstalled + " extensions");
			}
			
			// 2. Process files that has been selected for installation
			log.debug("Installing/updating extensions");
			// In the initial scan we only process already installed extensions
			Filter<ExtensionsFile> installFilter = initialScan ? 
				new InstalledFilter(settings) : new InstallFilter(request);
			
			// 2a. Load extension definitions
			WebClientRegisterExtensionsProcessor registerExtensions = 
				new WebClientRegisterExtensionsProcessor(rootPath, rootPath + RESOURCES_URL, rootPath + SERVLET_URL, results);
			registerExtensions.setForceUpdate(!initialScan);
			if (initialScan)
			{
				// Filter that only load web extension points
				registerExtensions.getXmlLoader().setFilter(new ExtensionPointFilter("net\\.sf\\.basedb\\.clients\\.web\\..*"));
			}
			manager.processFiles(registerExtensions, installFilter);

			// 2b. Load plug-in definitions information
			PluginInstallationProcessor pluginInstaller = new PluginInstallationProcessor(dc, results);
			manager.processFiles(pluginInstaller, installFilter);
			numInstalledPlugins = pluginInstaller.getNumPlugins();
			
			// 2c. Extract web resources
			ExtractResourcesProcessor extractResources = new ExtractResourcesProcessor(resourceDir, Pattern.compile("resources/(.*)"), "$1", results);
			manager.processFiles(extractResources, installFilter);

			// 2d. Servlets
			LoadServletsProcessor loadServlets = new LoadServletsProcessor(servletContext, results);
			manager.processFiles(loadServlets, installFilter);
				
			// 2e. Actual registration of all extensions
			registerExtensions.finalizeRegistration(manager, !initialScan);
			numRegistered = registerExtensions.getNumRegistered();
			
			// 3. Reset the 'last modified' status of all files and generated summary per file
			InstalledFileProcessor install = new InstalledFileProcessor(results);
			manager.processFiles(install, installFilter);
			numInstalled = install.getNumFiles();
			results.setEnded();
		}
		
		// Generate summary message, etc.
		StringBuilder summary = new StringBuilder();
		if (numInstalled > 0)
		{
			summary.append(numInstalled).append(" installed/updated extension file(s)\n");
			summary.append(numRegistered).append(" extension(s) registered\n");
		}
		if (numInstalledPlugins > 0) 
		{
			summary.append(numInstalledPlugins).append(" plug-in(s) installed or updated\n");
		}
		if (numUninstalled > 0)
		{
			summary.append(numUninstalled).append(" uninstalled extension file(s) containing ");
			summary.append(numUnregistered).append(" extension(s)\n");
		}
		if (numDisabledPlugins > 0) 
		{
			summary.append(numDisabledPlugins).append(" plug-in(s) have been disabled\n");
		}
		if (results.getNumErrorFiles() > 0)
		{
			summary.append(results.getNumErrorFiles()).append(" file(s) with errors");
		}
		results.setSummary(summary.toString());

		lastScanResults = results;
		return results;
	}
	
	

	/**
		Get an extension control object for managing the
		installed extensions. Everybody has permission to
		read information. To update, WRITE permission for the
		web client application is needed.
		
		@param dc The DbControl to use for database access and 
			permission checks, or null to only get read permission
		@return An extensions control instance
	*/
	public static ExtensionsControl get(DbControl dc)
	{
		Set<Permission> permissions = null;
		if (dc != null)
		{
			SessionControl sc = dc.getSessionControl();
			Client c = Client.getById(dc, sc.getClientId());
			permissions = c.getPermissions();
		}
		else
		{
			permissions = EnumSet.of(Permission.READ);
		}
		return new ExtensionsControl(dc, permissions);
	}
	
	/**
		Use extensions from one or more extension points. This method will return 
		an invoker object that allows the use and rendering of the extensions.
		The returned invoker will not reflect changes made to the registry by the 
		same or other threads after this call has returned. Each use/invokation of
		an extension point should call this method to get a new invoker object.
		
		@param context The current context
		@param extensionPoints An array with the ID values of the extension points.
			Extension points that are not found are ignored.
		@return An invoker instance
		@see Registry#useExtensions(net.sf.basedb.util.extensions.ClientContext, net.sf.basedb.util.extensions.ExtensionsFilter, String...)
	*/
	public static ExtensionsInvoker<?> useExtensions(JspContext context, String...extensionPoints)
	{
		return registry.useExtensions(context, settings, extensionPoints);
	}
	
	public static ExtensionsInvoker<?> useExtensions(JspContext context, 
		ExtensionsFilter filter, String...extensionPoints)
	{
		return registry.useExtensions(context, filter, extensionPoints);
	}
	
	/**
		Create a new {@link JspContext} object without an active DbControl.
		NOTE! Unless there is a very good reason we recommend that an 
		active DbControl is always supplied. Use the 
		{@link #createContext(DbControl, PageContext)} method instead.
		
		@param sc The current session control
		@param pageContext Page context for the executing JSP page
		@see #createContext(DbControl, PageContext)
	*/
	public static JspContext createContext(SessionControl sc, PageContext pageContext)
	{
		return new JspContext(sc, null, pageContext, null, null);
	}
	
	/**
		Create a new {@link JspContext} object with an active DbControl.
		@param dc An open DbControl
		@param pageContext Page context for the executing JSP page
	*/
	public static JspContext createContext(DbControl dc, PageContext pageContext)
	{
		return new JspContext(dc.getSessionControl(), dc, pageContext,
			null, null);
	}
	
	/**
		Create a new {@link JspContext} object with an active DbControl 
		and a current {@link GuiContext} and item.
		@param dc An open DbControl
		@param pageContext Page context for the executing JSP page
		@since 2.12
	*/
	public static JspContext createContext(DbControl dc, PageContext pageContext,
		GuiContext guiContext, Object currentItem)
	{
		return new JspContext(dc.getSessionControl(), dc,
			pageContext, guiContext, currentItem);
	}

	/**
		Get the URL to the home directory of extension with the
		given ID. The URL is a local absolute URL, eg. it includes the
		full path but not the protocol, server or port.
		
		@param extensionId The ID of a registered extension
		@return The local absolute URL to the home directory, or null
			if an extension with the given ID is not found
	*/
	public static String getHomeUrl(String extensionId)
	{
		ExtensionsFile xtFile = manager.getFileByObjectKey(new ExtensionKey(extensionId));
		return xtFile == null ? null : servletContext.getContextPath() + RESOURCES_URL + "/" + xtFile.getName();
	}
	
	/**
		Get the base URL for servlets in the extension with the given ID. The
		URL is a local absolute URL, eg. it includes the full path but not the
		protocol, server or port. The URL is usually something like:
		<p>
		<code>/extensions/servlet/[jar-file-name]/[servlet-name]</code>
		<p>
		NOTE! Servlets can also be invoked by using the alternate path:
		<p>
		<code>/extensions/[jar-file-name]/[servlet-name].servlet</code>
		
		
		@param extensionId The ID of a registered extension
		@param servletName The name of the servlet
		@return The local absolute URL to the servlet, or null
			if an extension with the given ID is not found
		@since 2.13
	*/
	public static String getServletUrl(String extensionId, String servletName)
	{
		ExtensionsFile xtFile = manager.getFileByObjectKey(new ExtensionKey(extensionId));
		String servletPath = null;
		if (xtFile != null)
		{
			servletPath = servletContext.getContextPath() + SERVLET_URL + "/" + xtFile.getName() + "/" + servletName;
		}
		return servletPath;
	}
	
	private final DbControl dc;
	private final Set<Permission> permissions;
	private ExtensionsControl(DbControl dc, Set<Permission> permissions)
	{
		this.dc = dc;
		this.permissions = permissions;
	}
	
	/*
		From the AccessControlled interface
		-----------------------------------
	*/
	@Override
	public void checkPermission(Permission permission)
		throws PermissionDeniedException 
	{
		if (!hasPermission(permission))
		{
			throw new PermissionDeniedException(permission);
		}
	}
	@Override
	public Set<Permission> getPermissions() 
	{
		return Collections.unmodifiableSet(permissions);
	}
	@Override
	public boolean hasPermission(Permission permission) 
	{
		return permissions.contains(permission);
	}
	// ----------------------------------------------
	
	/**
		For checking the permission and adding a custom message to
		the PermissionDeniedException if the required permission is
		missing.
	*/
	private void checkPermission(Permission permission, String what)
		throws PermissionDeniedException
	{
		if (!hasPermission(permission))
		{
			throw new PermissionDeniedException(permission, what);
		}
	}
	
	/**
		Scan the managed directories for new/updated/deleted extensions.
		Call {@link #getFiles()} to get an updated list of files.
		
		@throws PermissionDeniedException If the logged in user doesn't
			have WRITE permission
		@return The combined number of new/modified/deleted files
		@since 3.0
	*/
	public int scanForChanges()
	{
		return manager.scanForChanges();
	}
	
	/**
		Perform the requested actions (install/uninstall) as given 
		in the request form.
		
		@param request The HTTP request with parameters
		@throws PermissionDeniedException If the logged in user doesn't
			have WRITE permission
		@since 3.0
	*/
	public ProcessResults performActions(HttpServletRequest request)
	{
		checkPermission(Permission.WRITE, "extensions");
		ProcessResults results = scan(dc, request);
		saveSettings();
		return results;
	}
	
	/**
		Get the results of the last scan (manual or automatic).
		@return A {@link ProcessResults} object, or null if no scan
			has taken place
		@since 3.0
	*/
	public ProcessResults getLastScanResults()
	{
		return lastScanResults;
	}
	
	/**
		Get the default error handling factory.
		@since 2.17
	*/
	public ErrorHandlerFactory<Action> getDefaultErrorHandlerFactory()
	{
		return registry.getDefaultErrorHandlerFactory();
	}
	
	/**
		Get an list returning all registered extension points.
		@see Registry#getExtensionPoints()
		@since 3.0 (Returned an Iterator in BASE 2.x)
	*/
	public List<ExtensionPoint<?>> getExtensionPoints()
	{
		return registry.getExtensionPoints();
	}
	
	/**
		Get the extension point with a given ID.
		@param id The ID of the extension point
		@return An {@link ExtensionPoint} object or null if that 
			extension point doesn't exists
		@see Registry#getExtensionPoint(String)
	*/
	public ExtensionPoint<?> getExtensionPoint(String id)
	{
		return registry.getExtensionPoint(id);
	}	
	
	/**
		Check if an extension point is enabled of disabled.
		@param extensionPoint The extension point
		@return TRUE if the extension point is enabled, FALSE if it is disabled
		@see Settings#isEnabled(ExtensionPoint)
	*/
	public boolean isEnabled(ExtensionPoint extensionPoint) 
	{
		return settings.isEnabled(extensionPoint);
	}

	/**
		Get information about the last error that happened
		when rendering an extension point.
		@param id The id of the extension point
		@return The error, or null if no error information is available
		@since 2.9
	*/
	public Throwable getLastExtensionPointError(String id)
	{
		return registry.getLastExtensionPointError(id);
	}

	/**
		Enable/disable an extension point.
		@param extensionPointId The ID of the extension point to enable/disable
		@param enable TRUE to enable the extension point, FALSE to disable it
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission
	*/
	public void enableExtensionPoint(String extensionPointId, boolean enable)
	{
		checkPermission(Permission.WRITE, "ExtensionPoint[id=" + extensionPointId + "]");
		ExtensionPoint ep = getExtensionPoint(extensionPointId);
		if (ep == null)
		{
			log.warn("Extension point '" + extensionPointId + "' is not registered.");
			return;
		}
		if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, null);
		settings.enableExtensionPoint(extensionPointId, enable);
		if (enable) registry.handleEvent(AFTER_ENABLE, ep, null);
	}	
	
	/**
		Get a list returning all registered extensions.
		@see Registry#getExtensions()
		@since 3.0 (Returned an Iterator in BASE 2.x)
	*/
	public List<Extension<?>> getExtensions()
	{
		return registry.getExtensions();
	}
	
	/**
		Get a list returning all registered extensions for a
		specific extension point.
		@param id The ID of the extension point
		@see Registry#getExtensions(String)
		@since 3.0 (Returned an Iterator in BASE 2.x)
	*/
	public List<Extension<?>> getExtensions(String id)
	{
		return registry.getExtensions(id);
	}	
	
	/**
		Get the extension with a given ID.
		@param id The ID of the extension point
		@return An {@link Extension} object or null if that 
			extension doesn't exists
		@see Registry#getExtension(String)
	*/
	public Extension<?> getExtension(String id)
	{
		return registry.getExtension(id);
	}
	
	/**
		Get information about the last error that happened
		when rendering an extension.
		@param id The id of the extension
		@return The error, or null if no error information is available
		@since 2.9
	*/
	public Throwable getLastExtensionError(String id)
	{
		return registry.getLastExtensionError(id);
	}
	
	/**
		Check if an extension is enabled of disabled.
		@param extension The extension
		@return TRUE if the extension is enabled, FALSE if it is disabled
		@see Settings#isEnabled(Extension)
	*/
	public boolean isEnabled(Extension extension) 
	{
		return settings.isEnabled(extension);
	}
	
	/**
		Enable/disable an extension.
		@param extensionId The ID of the extension to enable/disable
		@param enable TRUE to enable the extension, FALSE to disable it
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission
	*/
	public void enableExtension(String extensionId, boolean enable)
	{
		checkPermission(Permission.WRITE, "Extension[id=" + extensionId + "]");
		Extension e = getExtension(extensionId);
		if (e == null)
		{
			log.warn("Extension '" + extensionId + "' is not registered.");
			return;
		}
		ExtensionPoint ep = getExtensionPoint(e.getExtends());
		if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, e);
		settings.enableExtension(extensionId, enable);
		if (enable) registry.handleEvent(AFTER_ENABLE, ep, e);
	}

	/**
		Get an list returning all XML/JAR files which contains installed
		extensions.
	*/
	public List<ExtensionsFile> getFiles()
	{
		return manager.getFiles();
	}
	
	/**
		Get information about an installed extensions file.
		@param fileuri The URI to the file
		@return An {@link ExtensionsFile} object or null if
			the given file doesn't exists or isn't an extensions file
	*/
	public ExtensionsFile getFile(String fileuri)
	{
		return manager.getFileByURI(fileuri);
	}
	
	/**
		Find out which file the given object is defined in. 
		
		@param key An object key
		@return Information about the file, or null if
			the object can't be found
		@since 3.0
	*/
	public ExtensionsFile getFileByObjectKey(ObjectKey key)
	{
		return manager.getFileByObjectKey(key);
	}
	
	/**
		Get the object that was registered for the given key.
		@param key An object key
		@return The object or null if no object was found
		@since 3.0
	*/
	public <O> O getObjectForKey(ObjectKey<O> key)
	{
		ExtensionsFile xtFile = getFileByObjectKey(key);
		return xtFile == null ? null : xtFile.getObjectForKey(key);
	}
	
	/**
		Enable or disable all extensions and extension points in a file.
		If no file with the given name is found or if that file isn't an
		extensions file, nothing is done.
		@param fileuri The fileuri which contains the extensions
		@param enable TRUE to enable, FALSE to disable
	*/
	public void enableAllInFile(String fileuri, boolean enable)
	{
		checkPermission(Permission.WRITE, "File[" + fileuri + "]");

		ExtensionsFile file = getFile(fileuri);
		if (file == null) 
		{
			log.warn("File '" + fileuri + "' is not an extensions file");
			return;
		}
		
		for (ExtensionPoint ep : file.getObjectsOfClass(ExtensionPoint.class))
		{
			if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, null);
			settings.enableExtensionPoint(ep.getId(), enable);
			if (enable) registry.handleEvent(AFTER_ENABLE, ep, null);
		}

		for (Extension ext : file.getObjectsOfClass(Extension.class))
		{
			ExtensionPoint ep = getExtensionPoint(ext.getExtends());
			if (!enable) registry.handleEvent(BEFORE_DISABLE, ep, ext);
			settings.enableExtension(ext.getId(), enable);
			if (enable) registry.handleEvent(AFTER_ENABLE, ep, ext);
		}
		
		for (PluginInfo info : file.getObjectsOfClass(PluginInfo.class))
		{
			if (info.isInstalled() && info.isDisabled() == enable)
			{
				PluginDefinition plugin = PluginDefinition.getById(dc, info.getInternalId());
				plugin.setDisabled(!enable);
			}
			info.setDisabled(!enable);
		}
	}
	
	/**
		Save the settings. This method should be called when changes 
		have been made to the settings. Otherwise, the settings are lost
		if the server is stopped.
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission
		@see Settings#save()
	*/
	public void saveSettings()
	{
		checkPermission(Permission.WRITE, "settings");
		settings.save();
	}
	
}
