/**
	$Id: IncludeContentTabFactory.java 6096 2012-09-03 07:06:01Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.tabcontrol;

import javax.servlet.ServletRequest;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.BodyContent;

import net.sf.basedb.clients.web.extensions.AbstractJspActionFactory;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.ThrowableUtil;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.PathType;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	A tab action factory implementation that includes another resource 
	as the content of tab. The other resource can be a JSP
	script, servlet, a static file, or anything else that makes sense.
	<p>
	The current {@link JspContext} is stored as a {@link ServletRequest} attribute
	with key {@link JspContext#ATTRIBUTE_KEY}. If the included page is a JSP page it
	can get the JspContext by calling 
	<code>request.getAttribute(JspContext.ATTRIBUTE_KEY)</code>.
	
	<p>
	If the extension point supports scripts and stylesheets, use the
	{@link #setScript(String)} and {@link #setStylesheet(String)} methods.
	Each call to these methods will add the argument to a set. Use
	the {@link #getScripts()} and {@link #getStylesheets()} methods to gain
	access to the sets, for example, to remove added items.
	
	@author nicklas
	@since 2.17
	@base.modified $Date: 2012-09-03 09:06:01 +0200 (Mon, 03 Sep 2012) $
*/
public class IncludeContentTabFactory 
	extends AbstractJspActionFactory<TabAction>
{

	private String id;
	private String clazz;
	private String style;
	private String title;
	private String tooltip;
	private String helpId;
	private boolean visible;
	private String onValidate;
	private String onActivate;
	private String include;
	private int tabIndex;
	
	/**
		Create a new factory instance.
	*/
	public IncludeContentTabFactory()
	{}
	
	/*
		From the ActionFactory interface
		--------------------------------
	*/
	@Override
	public TabAction[] getActions(InvokationContext context) 
	{
		Tab tab = new Tab((JspContext)context.getClientContext());
		return new TabAction[] { tab };
	}
	// ------------------------------------

	public void setId(String id) 
	{
		this.id = id;
	}

	public void setClazz(String clazz) 
	{
		this.clazz = clazz;
	}

	public void setStyle(String style) 
	{
		this.style = style;
	}

	public void setTitle(String title) 
	{
		this.title = title;
	}

	public void setTooltip(String tooltip) 
	{
		this.tooltip = tooltip;
	}

	public void setVisible(boolean visible) 
	{
		this.visible = visible;
	}

	public void setVisible(String visible) 
	{
		this.visible = Values.getBoolean(visible);
	}

	public void setHelpId(String helpId)
	{
		this.helpId = helpId;
	}
	
	public void setTabIndex(String tabIndex)
	{
		this.tabIndex = Values.getInt(tabIndex);
	}

	@VariableSetter
	public void setOnValidate(String onValidate) 
	{
		this.onValidate = onValidate;
	}

	@VariableSetter
	public void setOnActivate(String onActivate) 
	{
		this.onActivate = onActivate;
	}

	@PathSetter(pathType=PathType.CONTEXT_RELATIVE)
	public void setInclude(String include)
	{
		this.include = include;
	}

	class Tab
		implements TabAction
	{

		private final JspContext jspContext;
		
		Tab(JspContext jspContext)
		{
			this.jspContext = jspContext;
		}
		
		/*
			From the TabAction interface
			--------------------------------
		*/
		@Override
		public String getId() 
		{
			return id;
		}
	
		@Override
		public String getClazz() 
		{
			return clazz;
		}
	
		@Override
		public String getStyle() 
		{
			return style;
		}
	
		@Override
		public String getTitle() 
		{
			return title;
		}
	
		@Override
		public String getTooltip() 
		{
			return tooltip;
		}
	
		@Override
		public boolean isVisible() 
		{
			return visible;
		}
		
		@Override
		public String getHelpId()
		{
			return helpId;
		}
	
		@Override
		public String getOnValidate()
		{
			return onValidate;
		}

		@Override
		public String getOnActivate()
		{
			return onActivate;
		}
		
		@Override
		public int getTabIndex()
		{
			return tabIndex;
		}

		@Override
		public String getContent()
		{
			PageContext pageContext = jspContext.getPageContext();
			String tabContent = null;
			BodyContent tabBody = null;
			try
			{
				tabBody = pageContext.pushBody();
				pageContext.getRequest().setAttribute(JspContext.ATTRIBUTE_KEY, jspContext);
				pageContext.include(include);
				tabContent = tabBody.getString();
			}
			catch (Exception ex)
			{
				tabContent = "<div class=\"stacktrace\">" + 
					HTML.encodeTags(ThrowableUtil.stackTraceToString(ex)) + 
					"</div>";
			}
			finally
			{
				if (tabBody != null) pageContext.popBody();
			}
			return tabContent;
		}
	}
	
}
