/*
	$Id: ButtonGroup.java 5906 2011-12-12 11:59:03Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import java.util.ArrayList;
import java.util.List;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

/**
	This tag is used to group buttons together (horizontally) 
	and make them all have the same size.
	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:buttongroup
      id=...
      class=...
      subclass=...
      style=...
      &gt;
      &lt;base:button ... &gt;
      &lt;base:button ... &gt;
	&lt;/base:buttongroup&gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>The ID of the button. The value of this attribute
		goes directly into the standard <code>id</code> attribute.
		This means that if you need a reference to the button object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		buttonGroup = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>buttongroup</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different buttons.
		</td>
	</tr>
	<tr>
		<td>subclass</td>
		<td>-</td>
		<td>no</td>
		<td>
			Additional classes to put in the HTML class attribute. Doesn't
			replace the main class attribute.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different buttons.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class ButtonGroup
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = -8808469470232672680L;


	/**
		The buttons in this group.
	*/
	private List<String> buttons;
	
	/**
		If the buttons should be vertically grouped.
	*/
	private boolean vertical;
	
	public ButtonGroup()
	{
		super("buttongroup");
	}
	
	public void setVertical(boolean vertical)
	{
		this.vertical = vertical;
	}
	
	void addButton(String html)
	{
		buttons.add(html);
	}
	
	public int doStartTag()
		throws JspException
	{
		buttons = new ArrayList<String>();
		return EVAL_BODY_INCLUDE;
	}
	
	public int doEndTag()
		throws JspException
	{
		StringBuilder sb = new StringBuilder();
		int numButtons = buttons.size();
		if (numButtons > 0)
		{
			int percent = 100 / numButtons;
			sb.append("<div");
			addIdAndStyles(sb, vertical ? "vertical" : null);
			sb.append(">\n");
			sb.append("<table>\n");
			if (!vertical) sb.append("<tr>\n");
			for (String b : buttons)
			{
				if (vertical) sb.append("<tr>");
				sb.append("<td style=\"width: ").append(percent).append("%;\">").append(b).append("</td>\n");
				if (vertical) sb.append("</tr>");
			}
			if (!vertical) sb.append("</tr>\n");
			sb.append("</table>\n");
			sb.append("</div>\n");
		}
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}

}




