/*
	$Id: Annotation.java 6101 2012-09-04 09:34:05Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.PrefixSuffixFormatter;
import net.sf.basedb.util.units.LinearUnitConverter;
import net.sf.basedb.util.units.UnitConverter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Collections;
import java.util.Set;

/**
	This class represents an annotation. Annotations are user-defined
	properties that can be attached to {@link Annotatable}
	items. They can be very useful in the filtering and analysis
	process.
	<p>
	An annotation must have an {@link AnnotationType},
	which controls the type of and how many values that can be attached 
	to an annotation.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2012-09-04 11:34:05 +0200 (Tue, 04 Sep 2012) $
*/
public class Annotation
	extends BasicItem<AnnotationData>
{
	/**
		The type of item represented by this class.
		@see Item#ANNOTATION
		@see #getType()
	*/
	public static final Item TYPE = Item.ANNOTATION;
	
	/**
		Create a new annotation.
		@param dc The DbControl object
		@param annotationType The annotation type of the new annotation
		@param annotationSet The annotation set of the new annotation
		@see AnnotationSet#getAnnotation(AnnotationType)
	*/
	static Annotation getNew(DbControl dc, AnnotationType annotationType, AnnotationSet annotationSet)
		throws BaseException
	{
		assert annotationSet != null : "annotationSet == null";
		assert annotationType != null : "annotationType == null";
		Annotation a = dc.newItem(Annotation.class, annotationSet);
		AnnotationData data = a.getData();
		data.setAnnotationSet(annotationSet.getData());
		data.setAnnotationType(annotationType.getData());
		data.setUnit(annotationType.getData().getDefaultUnit());
		data.setValues(annotationType.getValueType().newParameterValueData());
		return a;
	}
	
	/**
		Get an <code>Annotation</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Annotation</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Annotation getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Annotation a = dc.loadItem(Annotation.class, id);
		if (a == null) throw new ItemNotFoundException("Annotation[id="+id+"]");
		return a;
	}
	
	/**
		Get a query configured to retrieve annotations. This query may return items 
		which the logged in user doesn't have 
		read permission to. At the moment there is no way to solve this problem.
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<Annotation> getQuery()
	{
		return new ItemQuery<Annotation>(Annotation.class, new QueryRuntimeFilterImpl());
	}
	
	private AnnotationSet annotationSet;
	private boolean disableATPermissionCheck;
	
	Annotation(AnnotationData annotationData)
	{
		super(annotationData);
	}
	
	Annotation(AnnotationData annotationData, AnnotationSet annotationSet)
	{
		super(annotationData);	
		this.annotationSet = annotationSet;
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		READ permission is granted if the logged in user has READ permission on 
		the associated item and annotation type. CREATE, WRITE and DELETE permissions
		are granted if the logged in user has WRITE permission on the associated item.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		AnnotationSet as = getAnnotationSet();
		if (!disableATPermissionCheck)
		{
			AnnotationType at = getAnnotationType();
		}
		if (as.hasPermission(Permission.WRITE))
		{
			granted |= Permission.grant(Permission.CREATE, Permission.READ, Permission.WRITE, Permission.DELETE);
		}
		else if (as.hasPermission(Permission.USE))
		{
			granted |= Permission.grant(Permission.USE);
		}
		else
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	// -------------------------------------------

	void disableATPermissionCheck()
	{
		this.disableATPermissionCheck = true;
	}
	
	/**
		Get the annotation set this annotation belongs to.
		@return An <code>AnnotationSet</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the annotation set (should never happen)
		@throws BaseException If anything else goes wrong when getting the annotation set.
	*/
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		if (annotationSet == null)
		{
			return getDbControl().getItem(AnnotationSet.class, getData().getAnnotationSet());
		}
		else
		{
			return annotationSet;
		}		
	}

	/**
		Get the date and time the values in this annotation was last
		updated.
		@since 2.14
	*/
	public Date getLastUpdate()
	{
		return DateUtil.copy(getData().getLastUpdate());
	}
	
	/**
		Check if this annotation belongs to the primaray annotations for a specified annotation set.
		This method is useful to check if an annotation has been inherited or not.
		@param as The annotation set to check
		@return TRUE if this annotation is a primary annotation of the specified annotation set,
			FALSE otherwise
		@since 2.10
	*/
	public boolean belongsTo(AnnotationSet as)
	{
		return getData().getAnnotationSet().equals(as.getData());
	}
	
	/**
		Get the annotation type of this annotation.
		@return An <code>AnnotationType</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the annotation type
	 * @throws BaseException If anything else goes wrong when getting the annotation type.
	*/
	public AnnotationType getAnnotationType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(AnnotationType.class, getData().getAnnotationType());
	}
	
	/**
		Checks if the values in this annotation has a unit.
		@return TRUE if a unit is used, FALSE otherwise
		@since 2.9
	*/
	public boolean hasUnit()
	{
		return getData().getUnit() != null;
	}
	
	/**
		Get the unit used for the values in this annotation.
		@return A unit, or null if this annotation doesn't use units
		@since 2.9
	*/
	public Unit getUnit()
	{
		return getDbControl().getItem(Unit.class, getData().getUnit());
	}
	
	/**
		Get a converter for converting the values in this annotation to
		a specific unit. 
		
		@param unit The unit to convert the values to, or null
			to convert the values to the unit specified by {@link #getUnit()}
		@return A unit converter, or null if this annotation doesn't use units, or
			if the specified unit is the same as the default unit of the annotation
			type (see {@link AnnotationType#getDefaultUnit()}
		@since 2.9
	*/
	public UnitConverter getUnitConverter(Unit unit)
	{
		UnitConverter converter = null;
		UnitData unitData = unit == null ? getData().getUnit() : unit.getData();
		if (unitData != null)
		{
			UnitData reference = getData().getAnnotationType().getDefaultUnit();
			if (!unitData.equals(reference))
			{
				converter = new LinearUnitConverter(
						unitData.getReferenceFactor(), unitData.getReferenceOffset(),
						reference.getReferenceFactor(), reference.getReferenceOffset());
			}
		}
		return converter;
	}

	/**
		Wrap a parent formatter with a formatter for units. NOTE!
		Values should be retrived with the same unit using 
		{@link #getValues(Unit)} to be correctly formatted with the
		returned formatter.
		
		@param parent The parent formatter, which should be of the
			correct type for the annotation values
		@param unit The unit to use for formatting, or null to
			use the annotation's unit
		@return A formatter object
		@since 2.9
	*/
	public <T> Formatter<T> getUnitFormatter(Formatter<T> parent, Unit unit)
	{
		Formatter<T> formatter = parent;
		UnitData unitData = unit == null ? getData().getUnit() : unit.getData();
		if (unitData == null) 
		{
			unitData = getData().getAnnotationType().getDefaultUnit();
		}
		if (unitData != null)
		{
			formatter = new PrefixSuffixFormatter<T>("", parent, 
					PrefixSuffixFormatter.NBSP + unitData.getDisplaySymbol());
		}
		return formatter;
	}
	
	/**
		Shortcut for <code>Unit.getDisplaySymbol()</code>, but also
		checks the default unit of the annotation type.
		@param unit The unit to get the symbol for, or null to use
			the annotation's unit
		@return The unit symbol, or null if this annotation doesn't use
			units
		@since 2.9
	*/
	public String getUnitSymbol(Unit unit)
	{
		UnitData unitData = unit == null ? getData().getUnit() : unit.getData();
		if (unitData == null) 
		{
			unitData = getData().getAnnotationType().getDefaultUnit();
		}
		return unitData == null ? null : unitData.getDisplaySymbol();
	}
	
	/**
		Get the value type for this annotation type. It can't be change once
		the object has been created.
	  	@return Type value type as an enum. 
		@since 2.2
	*/
	public Type getValueType()
	{
		return Type.fromValue(getData().getAnnotationType().getValueType());
	}
	
	/**
		Get the values this annotation contains. The values are of a
		{@link Type} appropriate for the given annotation type, ie. Integer object
		for an integer annotation type. All objects in the list are of the same type.
		If this is an annotation with units, the values are returned in the unit
		of the annotation type ({@link AnnotationType#getDefaultUnit()}). Use
		{@link #getValues(Unit)} to get the values in a different unit, including the
		specific unit used on this annotation.
		@see AnnotationType#getValueType()
		@return A list of objects
		@throws BaseException If there is an error
	*/
	public List<?> getValues()
		throws BaseException
	{
		List<?> values = Values.getItemValues(getDbControl(), getData().getValues().getValues());
		return Collections.unmodifiableList(values);
	}
	
	/**
		Get the values for this annotation in a specific unit. The values are of a
		{@link Type} appropriate for the given annotation type, ie. Integer object
		for an integer annotation type. All objects in the list are of the same type.
		
		@param unit The unit to convert the values to, or null to get the values
			in the unit used on this annotation
		@see AnnotationType#getValueType()
		@return A list of objects
		@throws BaseException If there is an error
		@since 2.9
	*/
	public List<?> getValues(Unit unit)
		throws BaseException
	{
		List<?> values = Values.getItemValues(getDbControl(), getData().getValues().getValues());
		UnitConverter converter = getUnitConverter(unit);
		if (converter != null)
		{
			List<Object> convertedValues = new ArrayList<Object>(values.size());
			Type valueType = getValueType();
			for (Object value : values)
			{
				if (value instanceof Number)
				{
					value = valueType.convertNumber(converter.convertToSpecificUnit(((Number)value).doubleValue()));
				}
				convertedValues.add(value);
			}
			values = convertedValues;
		}
		return Collections.unmodifiableList(values);
	}
	
	
	/**
		Get the version number of the item keeping the values. This is useful when you need
		to know if the annotation values has been modified by another transaction since
		they were read from the database. Unlike other items the version number of the
		annotation item doesn't change when the values are modified.
	  	@return The version number as an int. 
	*/
	public int getValuesVersion()
	{
		return getData().getValues().getVersion();
	}

	public void setValue(Object value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		setValue(value, null);
	}
	
	/**
		Set the value of the annotation, replacing any previous values.
		@param value The new value
		@param unit The unit of the value, or null to use the default unit of
			the annotation type
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the annotation or read permission for the annotation
			type
		@throws InvalidDataException If the value isn't a valid value for the 
			annotation type, see {@link AnnotationType#validateAnnotationValue(Object)}
		@throws BaseException If there is another error
		@since 2.9
	*/
	public void setValue(Object value, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationType annotationType = getAnnotationType();
		UnitConverter converter = null;
		if (annotationType.supportUnits() && unit != null)
		{
			converter = getUnitConverter(unit);
			Type valueType = getValueType();
			if (value instanceof Number)
			{
				Number numericValue = (Number)value;
				if (converter != null) 
				{
					numericValue = converter.convertToReferenceUnit(numericValue.doubleValue());
				}
				value = valueType.convertNumber(numericValue);
			}
		}
		annotationType.validateAnnotationValue(value);
		if (annotationType.supportUnits())
		{
			getData().setUnit(unit == null ? annotationType.getData().getDefaultUnit() : unit.getData());
		}
		getData().getValues().setSingleValue(Values.getDataValue(value));
		getData().setLastUpdate(new Date());
		getAnnotationSet().setSnapshotInvalid();
	}

	public void setValues(List<?> values)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		setValues(values, null);
	}
	
	/**
		Set the values of the annotation, replacing any previous values.
		@param values A list containing the new values
		@param unit The unit of the value, or null to use the default unit of
			the annotation type
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the annotation or read permission for the annotation
			type
		@throws InvalidDataException If the number of values are too many according
			to the {@link AnnotationType#getMultiplicity()} setting any of the values 
			isn't a valid value for the annotation type, see 
			{@link AnnotationType#validateAnnotationValue(Object)}
		@throws BaseException If there is another error
	*/
	public void setValues(List<?> values, Unit unit)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationType annotationType = getAnnotationType();
		int multiplicity = annotationType.getMultiplicity();
		if (multiplicity > 0 && values.size() > multiplicity)
		{
			throw new InvalidDataException("Too many values. Max "+multiplicity+" value(s) are allowed for this annotation type");
		}
		UnitConverter converter = null;
		if (annotationType.supportUnits() && unit != null)
		{
			converter = getUnitConverter(unit);
		}
		List<Object> convertedValues = new ArrayList<Object>(values.size());
		Type valueType = getValueType();
		for (Object value : values)
		{
			if (value instanceof Number)
			{
				Number numericValue = (Number)value;
				if (converter != null) 
				{
					numericValue = converter.convertToReferenceUnit(numericValue.doubleValue());
				}
				value = valueType.convertNumber(numericValue);
			}
			annotationType.validateAnnotationValue(value);
			convertedValues.add(value);
		}
		if (annotationType.supportUnits())
		{
			getData().setUnit(unit == null ? annotationType.getData().getDefaultUnit() : unit.getData());
		}
		getData().getValues().replaceValues(Values.getDataValues(convertedValues));
		getData().setLastUpdate(new Date());
		getAnnotationSet().setSnapshotInvalid();
	}

	/*
		Filter implementation that make sure that only annotations with annotation types
		that the logged in user has permission to read are loaded.
	 */
	static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		
		private Collection<Integer> annotationTypeIds;
		
		QueryRuntimeFilterImpl() 
		{}
		
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.hasPermission(Permission.READ, Item.ANNOTATIONTYPE))
			{
				// Do not add any filters.
			}
			else
			{
				// The user doesn't have generic permission to read annotation types
				// Load all that annotation types that the user owns or has been shared
				if (annotationTypeIds == null)
				{
					org.hibernate.Query atQuery = HibernateUtil.getPredefinedSQLQuery(dc.getHibernateSession(), 
						"GET_ALL_ANNOTATIONTYPE_IDS_FOR_USER");
					
					// Owner
					atQuery.setInteger("owner", sc.getLoggedInUserId());
					
					// All items keys for the logged in user
					Set<Integer> itemKeys = sc.getItemKeys();
					if (itemKeys == null || itemKeys.size() == 0) itemKeys = QueryRuntimeFilterFactory.ZERO_SET;
					atQuery.setParameterList("itemKeys", itemKeys);
					
					// All project keys for the active project and logged in user
					Set<Integer> projectKeys = sc.getProjectKeys();
					if (projectKeys == null || projectKeys.size() == 0) projectKeys = QueryRuntimeFilterFactory.ZERO_SET;
					atQuery.setParameterList("projectKeys", projectKeys);
					
					annotationTypeIds = HibernateUtil.loadList(Integer.class, atQuery, null);
					if (annotationTypeIds.size() == 0)
					{
						annotationTypeIds = QueryRuntimeFilterFactory.ZERO_SET;
					}
				}
				org.hibernate.Filter filter = manager.enableFilter("annotationTypes");
				filter.setParameterList("annotationTypes", annotationTypeIds);
			}
		}
	}
}

