/*
	$Id: BioMaterialList.java 5014 2009-06-25 13:11:37Z martin $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.BioMaterialData;
import net.sf.basedb.core.data.BioMaterialListData;
import net.sf.basedb.util.biomaterial.ListUtil;

/**
	This class represents a list of biomaterials.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-06-25 15:11:37 +0200 (Thu, 25 Jun 2009) $
*/
public class BioMaterialList
	extends CommonItem<BioMaterialListData>
{
	/**
		The type of item represented by this class.
		@see Item#BIOMATERIALLIST
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOMATERIALLIST;
	
	/**
		The maximum length of the external id variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setExternalId(String)} method to avoid
		exceptions.
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = BioMaterialListData.MAX_EXTERNAL_ID_LENGTH;

	
	/**
		Create a new <code>BioMaterialList</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param memberType The item type of members in this list
		@return The new <code>BioMaterialList</code> item
		@throws BaseException If there is an error
	*/
	public static BioMaterialList getNew(DbControl dc, Item memberType)
		throws BaseException
	{
		BioMaterialList l = dc.newItem(BioMaterialList.class);
		l.setName("New biomaterial list");
		l.setMemberType(memberType);
		return l;
	}

	/**
		Get a <code>BioMaterialList</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioMaterialList</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static BioMaterialList getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioMaterialList l = dc.loadItem(BioMaterialList.class, id);
		if (l == null) throw new ItemNotFoundException("BioMaterialList[id="+id+"]");
		return l;
	}

	/**
		Get a query that returns biomaterial lists.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<BioMaterialList> getQuery()
	{
		return new ItemQuery<BioMaterialList>(BioMaterialList.class);
	}

	/**
		Create a set with the intersection of the items
		in the given lists.
		@return A set with items. The set is always a new
			set, modifications to it will not affect anything else
	*/
	private static Set<BioMaterialData> intersection(Collection<BioMaterialList> lists)
	{
		Set<BioMaterialData> intersection = new HashSet<BioMaterialData>();
		if (lists != null)
		{
			boolean first = true;
			Iterator<BioMaterialList> it = lists.iterator();
			while (it.hasNext())
			{
				Set<BioMaterialData> other = it.next().getData().getBioMaterials();
				if (first)
				{
					intersection.addAll(other);
					first = false;
				}
				else
				{
					intersection.retainAll(other);
				}
			}
		}
		return intersection;
	}
	
	/**
		Create a set with the union of the items 
		in the given lists.
		@return A set with items. The set is always a new
			set, modifications to it will not affect anything else
	*/
	private static Set<BioMaterialData> union(Collection<BioMaterialList> lists)
	{
		Set<BioMaterialData> union = new HashSet<BioMaterialData>();
		if (lists != null)
		{
			for (BioMaterialList list : lists)
			{
					union.addAll(list.getData().getBioMaterials());
			}
		}
		return union;
	}
	
	/**
		Count the number of times each item is present in the
		given lists and return those items that are found in
		&gt=minCount and &lt;=maxCount lists.
		<p>
		NOTE! 
		<ul>
			<li>'minCount=1' and 'maxCount=number of lists' is the same
				as {@link #union(Collection)}
			<li>'minCount=maxCount=number of lists' is the same
				as {@link #intersection(Collection)}.
		</ul>
	
		@param lists The lists to look in
		@param minCount The minimum number of list an item
			is required to be present in
		@param maxCount The maximum number of lists an item
			is allowed to the present in
		@return A set with items. It is always a new
			set, modifications to it will not affect anything else
	*/
	private static Set<BioMaterialData> count(Collection<BioMaterialList> lists, final int minCount, final int maxCount)
	{
		Map<BioMaterialData, Integer> union = new HashMap<BioMaterialData, Integer>();
		if (lists != null)
		{
			Iterator<BioMaterialList> it = lists.iterator();
			while (it.hasNext())
			{
				Iterator<BioMaterialData> e = it.next().getData().getBioMaterials().iterator();
				while (e.hasNext())
				{
					BioMaterialData bm = e.next();
					Integer count = union.get(bm);
					count = count == null ? 1 : count+1;
					union.put(bm, count);
				}
			}
		}
		CollectionUtils.filter(union.values(), 
			new Predicate()
		{
			@Override
			public boolean evaluate(Object o)
			{
				int count = (Integer)o;
				return minCount <= count && count <= maxCount;
			}
		});
		return union.keySet();
	}
	
	BioMaterialList(BioMaterialListData listData)
	{
		super(listData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/**
		Get the external id for this list.
		@return A string with the external id of this item, or null
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	/**
		Set the external id for this list. Null is allowed, but
		the value must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant. The value doesn't have to be unique.
		@param externalId The new external id for this item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	/**
		Get the number of members in this list.
	*/
	public int getSize()
	{
		return getData().getSize();
	}

	/**
		Get the item type of members on this list.
		@return An Item object
	*/
	public Item getMemberType()
	{
		return Item.fromValue(getData().getMemberType());
	}
	
	private void setMemberType(Item memberType)
	{
		if (memberType == null) throw new InvalidUseOfNullException("memberType");
		if (!BioMaterialData.class.isAssignableFrom(memberType.getDataClass()))
		{
			throw new InvalidDataException("Unsupported member type: " + memberType);
		}
		getData().setMemberType(memberType.getValue());
	}
	
	/**
		Checks if the other type is the same type as the member type of
		this list.
	*/
	private void checkMemberType(Item otherType)
	{
		if (getMemberType() != otherType)
		{
			throw new InvalidDataException("List '" + getName() + "' can't contain " + otherType);
		}
	}
	
	/**
		Checks that the lists have the same member type as this list.
	*/
	private void checkMemberType(Collection<BioMaterialList> lists)
	{
		Item memberType = getMemberType();
		for (BioMaterialList list : lists)
		{
			if (memberType != list.getMemberType())
			{
				throw new InvalidDataException("List '" + getName() + "' can't contain " + list.getMemberType());
			}
		}
	}
	
	/**
		Add a biomaterial to this list.
	 
		@param bioMaterial The biomaterial to add to the list
		@return TRUE if the biomaterial was added, FALSE if it already existed
			in this list
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the biomaterial is null or not of the
			correct member type
	*/
	public boolean add(BioMaterial bioMaterial)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (bioMaterial == null) throw new InvalidUseOfNullException("bioMaterial");
		checkMemberType(bioMaterial.getType());
		boolean added = getData().getBioMaterials().add((BioMaterialData)bioMaterial.getData());
		if (added) getData().setSize(getSize() + 1);
		return added;
	}
	
	/**
		Add multiple biomaterials to this list in one go.
		@param it An iterator that returns all biomaterials to be added,
			null items are ignored
		@return The numer of added biomaterials
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the iterator returns a biomaterial
			of a different type than {@link #getMemberType()}
	*/
	public int add(Iterator<? extends BioMaterial> it)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (it == null) return 0;
		int added = 0;
		Set<BioMaterialData> data = getData().getBioMaterials();
		while (it.hasNext())
		{
			BioMaterial bm = it.next();
						
			if (bm == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}
			
			checkMemberType(bm.getType());
			if (data.add((BioMaterialData)bm.getData())) added++;
		}
		getData().setSize(getSize() + added);
		return added;
	}
	
	/**
		Remove a biomaterial from the list.
		@param bioMaterial The biomaterial to remove.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the biomaterial is null
	*/
	public void remove(BioMaterial bioMaterial)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (bioMaterial == null) throw new InvalidUseOfNullException("bioMaterial");
		boolean removed = getData().getBioMaterials().remove(bioMaterial.getData());
		if (removed) getData().setSize(getData().getSize()-1);
	}
	
	/**
		Get a query that returns all biomaterials that are
		members of this list.
		@return An {@link ItemQuery} object
		@throws BaseException If the query could not be created and configured.
	*/
	public ItemQuery<? extends BioMaterial> getMemberBioMaterials()
		throws BaseException
	{
		ItemQuery<? extends BioMaterial> query = getAllBioMaterials();
		query.joinPermanent(Hql.innerJoin("bioMaterialLists", Item.BIOMATERIALLIST.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.BIOMATERIALLIST.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Get a query that returns all biomaterials that can possibly be 
		members of this list. NOTE! This is just a convenience function
		that calls {@link BioSource#getQuery()}, etc. depending on the
		member type of the list.
		@return An {@link ItemQuery} object
		@throws BaseException If the query could not be created and configured
		@see ListUtil#getQuery(Item)
	*/
	public ItemQuery<? extends BioMaterial> getAllBioMaterials()
		throws BaseException
	{
		return ListUtil.getQuery(getMemberType());
	}

	/**
		Add members to this list.
		@param items The items to add
		@return The number of new items
	*/
	private int addMembers(Set<BioMaterialData> items)
	{
		int numAdded = 0;
		BioMaterialListData data = getData();
		Set<BioMaterialData> target = data.getBioMaterials();
		for (BioMaterialData item : items)
		{
			if (target.add(item)) numAdded++;
		}
		data.setSize(getSize() + numAdded);
		return numAdded;
	}
	
	/**
		Remove members from this list
		@param items The items to remove
		@return The number of removed items
	*/
	private int removeMembers(Set<BioMaterialData> items)
	{
		int numRemoved = 0;
		BioMaterialListData data = getData();
		Set<BioMaterialData> target = data.getBioMaterials();
		
		// We loop over the list that is smallest
		if (items.size() <= this.getSize())
		{
			for (BioMaterialData item : items)
			{
				if (target.remove(item)) numRemoved++;
			}
		}
		else
		{
			Iterator<BioMaterialData> it = target.iterator();
			while (it.hasNext())
			{
				if (items.contains(it.next()))
				{
					it.remove();
					numRemoved++;
				}
			}
		}
		data.setSize(getSize() - numRemoved);
		return numRemoved;
	}
	
	/**
		Keep only specified members in this list
		@param items The items to keep
		@return The number of removed items
	*/
	private int retainMembers(Set<BioMaterialData> items)
	{
		int numRemoved = 0;
		BioMaterialListData data = getData();
		Set<BioMaterialData> target = data.getBioMaterials();

		if (items.size() == 0)
		{
			target.clear();
			numRemoved = getSize();
		}
		else
		{
			Iterator<BioMaterialData> it = target.iterator();
			while (it.hasNext())
			{
				if (!items.contains(it.next()))
				{
					it.remove();
					numRemoved++;
				}
			}
		}
		data.setSize(getSize() - numRemoved);
		return numRemoved;
	}
	
	/**
		Add all items from another biomaterial list to this
		list. The new list is the union of the existing list
		and the added list. The other list must have the 
		same member type as this list.
		
		@param list The list to add items from
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addAll(BioMaterialList list)
	{
		checkPermission(Permission.WRITE);
		if (list == null) return 0;
		checkMemberType(list.getMemberType());
		return addMembers(list.getData().getBioMaterials());
	}

	/**
		Add items that are present in at least ONE OF the given lists to
		this list. Eg. the added items is the union from 
		the given lists. 
	
		@param lists The lists to add items from
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addUnion(Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		if (lists == null) return 0;
		checkMemberType(lists);

		int numAdded = 0;
		if (lists.size() == 1)
		{
			numAdded = addMembers(lists.iterator().next().getData().getBioMaterials());
		}
		else if (lists.size() > 1)
		{
			numAdded = addMembers(union(lists));
		}
		return numAdded;
	}

	/**
		Add items that are present in the specified number of lists to 
		this list.
		<ul>
		<li>Delegate to {@link #addUnion(Collection)} if
			<code>minCount &lt;= 1 && maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #addIntersection(Collection)}
			if <code>minCount = size(lists) && maxCount >= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount > maxCount</code> no 
			items will be added.
		</ul>
		
		@param lists The lists to add items from
		@param minCount The minimum number of lists an item is required to
			be present in
		@param maxCount The maximum number of lists an item is allowed to 
			be present in
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addIfPresentIn(int minCount, int maxCount, Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		if (lists == null) return 0;
		checkMemberType(lists);
		
		int numAdded = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numAdded = addUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numAdded = addIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			numAdded = addMembers(count(lists, minCount, maxCount));
		}
		return numAdded;
	}

	/**
		Add items that are present in ALL OF the given lists to
		this list. Eg. the added items is the intersection from 
		the given lists. 
	
		@param lists The lists to add items from
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addIntersection(Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		if (lists == null) return 0;
		checkMemberType(lists);
		
		int numAdded = 0;
		if (lists.size() == 1)
		{
			numAdded = addMembers(lists.iterator().next().getData().getBioMaterials());
		}
		else if (lists.size() > 1)
		{
			numAdded = addMembers(intersection(lists));
		}
		return numAdded;
	}

	/**
		Remove all items from this list that are present
		in the other list. 
		
		@param list The list that contains the items
			to remove
		@return The number of items removed from this list
	*/
	public int removeAll(BioMaterialList list)
	{
		checkPermission(Permission.WRITE);
		if (list == null) return 0;
		return removeMembers(list.getData().getBioMaterials());
	}

	/**
		Remove items that are present in at least ONE OF the given lists from
		this list. Eg. the removed items is the union from 
		the given lists. 
	
		@param lists The lists that contains the items to remove
		@return The number of items removed from this list
	*/
	public int removeUnion(Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		if (lists == null) return 0;
		
		int numRemoved = 0;
		if (lists.size() == 1)
		{
			numRemoved = removeMembers(lists.iterator().next().getData().getBioMaterials());
		}
		else if (lists.size() > 1)
		{
			numRemoved = removeMembers(union(lists));
		}
		return numRemoved;
	}

	/**
		Remove items that are present in the specified number of lists from 
		this list.
		<ul>
		<li>Delegate to {@link #removeUnion(Collection)} if
			<code>minCount &lt;= 1 && maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #removeIntersection(Collection)}
			if <code>minCount = size(lists) && maxCount >= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount > maxCount</code> no 
			items will be removed.
		</ul>
		
		@param lists The lists with the items
		@param minCount The minimum number of lists an item is required to
			be present in
		@param maxCount The maximum number of lists an item is allowed to 
			be present in
		@return The number of items removed
	*/
	public int removeIfPresentIn(int minCount, int maxCount, Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		if (lists == null) return 0;
		
		int numRemoved = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numRemoved = removeUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numRemoved = removeIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			numRemoved = removeMembers(count(lists, minCount, maxCount));
		}
		return numRemoved;
	}

	/**
		Remove items that are present in ALL OF the given lists from
		this list. Eg. the removed items is the intersection from 
		the given lists. 
	
		@param lists The lists that contains the items to remove
		@return The number of items removed from this list
	*/
	public int removeIntersection(Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		if (lists == null) return 0;
		
		int numRemoved = 0;
		if (lists.size() == 1)
		{
			numRemoved = removeMembers(lists.iterator().next().getData().getBioMaterials());
		}
		else if (lists.size() > 1)
		{
			numRemoved = removeMembers(intersection(lists));
		}
		return numRemoved;
	}

	/**
		Remove all items from this list that are NOT present
		in the other list. The new list is the intersection
		of the existing list and the other list. NOTE! If the
		other list is null or empty, all items will be removed
		from this list.
		
		@param list The list that contains the items
			to be retained
		@return The number of items removed from this list
	*/
	public int retainAll(BioMaterialList list)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (list == null || list.getSize() == 0)
		{
			getData().getBioMaterials().clear();
			numRemoved = getSize();
			getData().setSize(0);
		}
		else
		{
			numRemoved = retainMembers(list.getData().getBioMaterials());
		}
		return numRemoved;
	}

	/**
		Retain items that are present in at least ONE OF the given lists 
		in this list. Eg. items that are not found in any of the given
		lists are removed from this list.
	
		@param lists The lists that contains the items to retain
		@return The number of items removed from this list
	*/
	public int retainUnion(Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (lists == null || lists.size() == 0)
		{
			getData().getBioMaterials().clear();
			numRemoved = getSize();
			getData().setSize(0);
		}
		else if (lists.size() == 1)
		{
			numRemoved = retainMembers(lists.iterator().next().getData().getBioMaterials());
		}
		else
		{
			numRemoved = retainMembers(union(lists));
		}
		return numRemoved;
	}

	/**
		Retain items that are present in ALL OF the given lists 
		in this list. Eg. items that are not found in all 
		lists of the given lists are removed from this list.
	
		@param lists The lists that contains the items to retain
		@return The number of items removed from this list
	*/
	public int retainIntersection(Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (lists == null || lists.size() == 0)
		{
			getData().getBioMaterials().clear();
			numRemoved = getSize();
			getData().setSize(0);
		}
		else if (lists.size() == 1)
		{
			numRemoved = retainMembers(lists.iterator().next().getData().getBioMaterials());
		}
		else
		{
			numRemoved = retainMembers(intersection(lists));
		}
		return numRemoved;
	}

	/**
		Retain items that are present in the specified number of lists in 
		this list.
		<ul>
		<li>Delegate to {@link #retainUnion(Collection)} if
			<code>minCount &lt;= 1 && maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #retainIntersection(Collection)}
			if <code>minCount = size(lists) && maxCount >= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount > maxCount</code> no 
			items will be removed.
		</ul>
		
		@param lists The lists with the items
		@param minCount The minimum number of lists an item is required to
			be present in
		@param maxCount The maximum number of lists an item is allowed to 
			be present in
		@return The number of items removed
	*/
	public int retainIfPresentIn(int minCount, int maxCount, Collection<BioMaterialList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		int numLists = lists == null ? 0 : lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numRemoved = retainUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numRemoved = retainIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			numRemoved = retainMembers(count(lists, minCount, maxCount));
		}
		return numRemoved;
	}

}
