/*
	$Id: DbControl.java 6079 2012-08-06 13:20:43Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.QuotaData;
import net.sf.basedb.core.data.QuotaIndex;
import net.sf.basedb.core.data.QuotaTypeData;
import net.sf.basedb.core.data.DiskConsumableData;
import net.sf.basedb.core.data.DiskUsageData;
import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.hibernate.SessionWrapper;
import net.sf.basedb.core.hibernate.StatelessSessionWrapper;
import net.sf.basedb.core.log.LogManagerFactory;
import net.sf.basedb.core.log.LoggingInterceptor;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;

import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.lang.reflect.Constructor;

/**
	A <code>DbControl</code> object is the main object used for communicating
	with the database. It contains methods for saving and deleting
	items, handling transactions, etc. <code>DbControl</code> objects
	are created by the {@link SessionControl#newDbControl()} method.

	@author Nicklas, Samuel
	@version 2.0
	@base.modified $Date: 2012-08-06 15:20:43 +0200 (Mon, 06 Aug 2012) $
*/
public final class DbControl
{

	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.DbControl");

	/*
		The <code>SessionControl</code> that created this object.
	*/
	private final SessionControl sc;
	
	/**
		To keep track of the code that created this DbControl.
	*/
	private final Throwable calledFrom;
	
	/**
		Handles logging of changes to items.
	*/
	private LogControl logControl;

	/**
		The Hibernate session.
	*/
	private SessionWrapper hSession;
	
	/**
		A Hibernate stateless session.
	*/
	private StatelessSessionWrapper hStatelessSession;
	
	/**
		The Hibernate transaction.
	*/
	private org.hibernate.Transaction hTransaction;
	
	/**
		A cache of already loaded items. Maps from data object --> item object
	*/
	private Map<BasicData, BasicItem> itemCache;
	
	/**
		A list of items that need special handling when commit is called.
	*/
	private Map<BasicItem, Transactional.Action> commitQueue;

	/**
		Lists items that are saved only if the parent item is also saved.
		@see #saveItemIf(BasicItem, BasicItem, boolean)
	*/
	private Map<BasicItem, List<SaveIf>> saveIfQueue;
	
	/**
		A list of batchers that will be closed when <code>DbControl</code> is
		closed and will be flushed when commit is called.
	*/
	private List<Batcher> batchers;
	
	/**
		List of actions that should be performed when the
		transaction commits or rollbacks.
	*/
	private List<TransactionalAction> transactionalActions;
	
	/**
		Is the connection closed or not?
	*/
	private boolean isClosed;
	
	/**
		If Hibernate is connected to the database or not.
	*/
	private boolean isConnected;

	/**
		If the current transaction has been rolled back or not.
	*/
	private boolean isRolledBack;
	
	/**
		Create a new object.
	*/
	@SuppressWarnings("unchecked")
	DbControl(SessionControl sc)
		throws BaseException
	{
		this.sc = sc;
		
		// Initialize logging extensions...
		ExtensionsManager xtManager = Application.getExtensionsManager();
		ClientContext context = new ClientContext(sc);
		ExtensionsInvoker<LogManagerFactory> invoker = (ExtensionsInvoker<LogManagerFactory>)xtManager.getRegistry().useExtensions(
			context, xtManager.getSettings(), "net.sf.basedb.core.log-manager");		
		LoggingInterceptor interceptor = null;
		if (invoker.getNumExtensions() > 0)
		{
			logControl = new LogControl(this);
			interceptor = new LoggingInterceptor(logControl, invoker);
		}

		hSession = new SessionWrapper(HibernateUtil.newSession(interceptor, null));
		hTransaction = HibernateUtil.newTransaction(hSession);
		itemCache = new IdentityHashMap<BasicData, BasicItem>();
		commitQueue = new LinkedHashMap<BasicItem,Transactional.Action>();
		isClosed = false;
		isConnected = true;
		calledFrom = new Throwable("Please check the code to make sure that DbControl.close() " +
			"is always called. Stacktrace of code that created this DbControl:");
	}

	/**
		Check if the connection to the database has been closed.
		To get a new connection, you must create a new <code>DbControl</code>
		object.
		@return TRUE if the connection is closed, FALSE otherwise
		@see SessionControl#newDbControl()
	*/
	public boolean isClosed()
	{
		return isClosed;
	}

	/**
		Get the {@link SessionControl} object that owns
		this <code>DbControl</code> object.
		@return a <code>SessionControl</code> object
	*/
	public SessionControl getSessionControl()
	{
		return sc;
	}
	
	/**
		Get the log controller for this db control. May be null if
		logging has been disabled.
		@since 2.13
	*/
	LogControl getLogControl()
	{
		return logControl;
	}
	
	/**
		Get the underlying Hibernate session.
	*/
	org.hibernate.Session getHibernateSession()
	{
		return hSession;
	}

	/**
		Get a stateless Hibernate session using the same database connection as
		the regular session. If no stateless session exists a new one is created.
	*/
	org.hibernate.StatelessSession getStatelessSession()
	{
		if (hSession != null && hStatelessSession == null)
		{
			hStatelessSession = new StatelessSessionWrapper(HibernateUtil.newStatelessSession(hSession));
		}
		return hStatelessSession;
	}
	
	/**
		Temporarily disconnect from the database. Uncommitted changes of regular
		items are not lost, but you must {@link #reconnect()} before calling 
		{@link #commit()}. If you have used batch items, for example
		{@link ReporterBatcher}, you should call {@link #commit()} or you will lose
		everything that has already been {@link BasicBatcher#flush()}:ed to the
		database.
	  	@throws ConnectionClosedException If connection to database has been closed. 
	 	@throws BaseException If there is another error
	*/
	public void disconnect()
		throws ConnectionClosedException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		HibernateUtil.rollback(hTransaction);
		hSession.disconnect();
		isConnected = false;
	}
	
	/**
		Reconnects to the database after a {@link #disconnect()}.
	 	@throws ConnectionClosedException If connection to database has been closed.
	 	@throws BaseException If there is another error.
	*/
	public void reconnect()
		throws ConnectionClosedException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		hTransaction = HibernateUtil.newTransaction(hSession);
		isConnected = true;
	}
	
	/**
		Checks if there is an active connection to the database or not.
	*/
	public boolean isConnected()
	{
		return isConnected;
	}
	
	/**
		Close the connection to the database. Uncommitted changes
		will be rolled back. If the connection is already closed
		this method does nothing. Any errors in this method
		are written to the log file. After calling this method the 
		<code>DbControl</code> cannot be used anymore.
	*/
	public void close()
	{
		if (isClosed()) return;
		rollback();
		cleanUp();
	}
	
	/**
		If the transaction was rolled back, either as a call to {@link #close()}
		or because of an error during {@link #commit()}. To check if a transaction
		was successfully committed use: <code>dc.isClosed() && !dc.isRolledBack()</code>.
		
		@return TRUE if the transaction was rolled back, FALSE if it was committed
			or is still open
	*/
	public boolean isRolledBack()
	{
		return isRolledBack;
	}
	
	/**
		Rollback the transaction.
	*/
	private void rollback()
	{
		isRolledBack = true;
		// Close up batchers
		// Make copy because close will remove it from the queue
		if (batchers != null)
		{
			ArrayList<Batcher> batchCopy = new ArrayList<Batcher>(batchers);
			for (Batcher batcher : batchCopy)
			{
				try
				{
					batcher.close();
				}
				catch (Throwable ex)
				{
					log.warn("Exception during rollback in batcher.close()", ex);
				}
			}
		}
		try
		{
			if (isConnected) HibernateUtil.rollback(hTransaction);
			HibernateUtil.close(hSession);
			if (hStatelessSession != null) HibernateUtil.close(hStatelessSession);
		}
		catch (Throwable ex)
		{
			log.warn("Exception during rollback in Hibernate", ex);
		}
		hStatelessSession = null;
		hSession = null;
		hTransaction = null;

		// Call TransactionalAction.onRollback()
		if (transactionalActions != null)
		{
			for (TransactionalAction action : transactionalActions)
			{
				try
				{
					action.onRollback();
				}
				catch (RuntimeException ex)
				{
					log.warn("Action failed: " + action, ex);
				}
			}
		}
		for (Map.Entry<BasicItem,Transactional.Action> entry : commitQueue.entrySet())
		{
			BasicItem item = entry.getKey();
			Transactional.Action action = entry.getValue();
			if (item instanceof Transactional) item.onRollback(action);
		}
	}
	
	/**
		Clear and close all objects.
	*/
	private void cleanUp()
	{
		commitQueue.clear();
		itemCache.clear();
		if (batchers != null) batchers.clear();
		if (saveIfQueue != null) saveIfQueue.clear();
		if (transactionalActions != null) transactionalActions.clear();
		saveIfQueue = null;
		commitQueue = null;
		transactionalActions = null;
		itemCache = null;
		batchers = null;
		isClosed = true;
		isConnected = false;
	}
	
	/**
		Add a transactional action to this DbControl.
		@since 3.1
	*/
	public void addTransactionalAction(TransactionalAction action)
	{
		if (transactionalActions == null)
		{
			transactionalActions = new ArrayList<TransactionalAction>();
		}
		transactionalActions.add(action);
	}
	
	/**
		Commit all changes made to items and close the connection to the database. 
		If there is an error, a rollback is issued and the connection is closed.

		@throws BaseException If there is an error
	*/
	public void commit()
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();

		Project activeProject = null;
		if (sc.getActiveProjectId() != 0)
		{
			try
			{
				activeProject = Project.getById(this, sc.getActiveProjectId());
			}
			catch (ItemNotFoundException ex)
			{
				// The project was deleted
				sc.setActiveProject(null);
			}
		}
		
		Map<BasicItem, Transactional.Action> tempQueue = null;
		Map<BasicItem, Transactional.Action> afterCommitQueue = new HashMap<BasicItem, Transactional.Action>();

		try
		{
			// Flush and close all batchers
			// Make copy because close will remove it from the queue
			if (batchers != null)
			{
				ArrayList<Batcher> batchCopy = new ArrayList<Batcher>(batchers);
				for (Batcher batcher : batchCopy)
				{
					batcher.close();
				}
				batchers = null;
			}

			/*
				We must make a copy of and clear the commitQueue since the onBeforeCommit may add new 
				items to it. Therefore we must also check if new items have been added to the queue
				and process them as well.
			*/
			while (commitQueue.size() > 0)
			{
				tempQueue = commitQueue;
				commitQueue = new HashMap<BasicItem, Transactional.Action>();
				Iterator<Map.Entry<BasicItem,Transactional.Action>> iterator = tempQueue.entrySet().iterator();

				while (iterator.hasNext())
				{
					Map.Entry<BasicItem,Transactional.Action> entry = iterator.next();
					BasicItem item = entry.getKey();
					Transactional.Action action = entry.getValue();
					
					// Don't process the same item twice for the same action
					if (afterCommitQueue.get(item) == action)
					{
						/* #### CONTINUE-STATEMENT #### */						
						continue;
					}
					BasicData data = item.getData();
					boolean controlled = item instanceof Controlled;
					boolean transactional = controlled && item instanceof Transactional;
					if (controlled && (item instanceof Validatable)) item.validate();
					if (action != Transactional.Action.UPDATE || transactional) 
					{
						item.onBeforeCommit(action);
					}
					if (controlled && (item instanceof DiskConsumable)) updateDiskUsage(item);
					if (action == Transactional.Action.CREATE)
					{
						if (activeProject != null) item.setProjectDefaults(activeProject);
						HibernateUtil.saveData(hSession, data);
						item.onAfterInsert();
					}
					else if (action == Transactional.Action.DELETE)
					{
						if (HibernateUtil.exists(hSession, item.getType().getDataClass(), item.getId()))
						{
							if (item.isUsed()) throw new ItemInUseException(item.toString());
							HibernateUtil.deleteData(hSession, data);
							// Need flush so directories can be deleted recursiveley
							if (item instanceof Directory || item instanceof File) HibernateUtil.flush(hSession);
						}
					}
					if (action != Transactional.Action.UPDATE || transactional) afterCommitQueue.put(item, action);
				}
			}
			// Call TransactionalAction.onBeforeCommit()
			if (transactionalActions != null)
			{
				for (TransactionalAction action : transactionalActions)
				{
					action.onBeforeCommit();
				}
			}
			HibernateUtil.commit(hTransaction);
			HibernateUtil.close(hSession);
			if (hStatelessSession != null) HibernateUtil.close(hStatelessSession);
			hSession = null;
			hTransaction = null;
		}
		catch (RuntimeException ex)
		{
			if (afterCommitQueue != null) commitQueue.putAll(afterCommitQueue);
			if (tempQueue != null) commitQueue.putAll(tempQueue);
			close();
			throw ex;
		}
		// Call TransactionalAction.onAfterCommit()
		if (transactionalActions != null)
		{
			for (TransactionalAction action : transactionalActions)
			{
				try
				{
					action.onAfterCommit();
				}
				catch (RuntimeException ex)
				{
					log.warn("Action failed: " + action, ex);
				}
			}
		}
		cleanUp();
		for (Map.Entry<BasicItem,Transactional.Action> entry : afterCommitQueue.entrySet())
		{
			BasicItem item = entry.getKey();
			Transactional.Action action = entry.getValue();
			item.onAfterCommit(action);
		}
	}

	/**
		Update the disk usage information for a DiskConsumable item.
		It is expected that the <code>item</code> parameter impements the
		<code>DiskConsumable</code> interface and that it's data object implements the
		<code>DiskConsumableData</code> interface.
		@throws QuotaException If the quota is exceeded
		@throws BaseException If there is another error
	*/
	private void updateDiskUsage(BasicItem item)
		throws QuotaException, BaseException
	{
		DiskConsumable dcItem = (DiskConsumable)item;  // The item
		DiskConsumableData dcData = (DiskConsumableData)item.getData();  // The data object for the item
		DiskUsageData currentDiskUsage = dcData.getDiskUsage();  // Current disk usage information

		QuotaTypeData quotaType = HibernateUtil.loadData(hSession, QuotaTypeData.class, SystemItems.getId(dcItem.getQuotaTypeSystemId()));
		Location location = dcItem.getLocation();
		long bytes = dcItem.getBytes();
		boolean useDifference = dcItem.useDifference();
		long difference = dcItem.getDifference();
		
		// Validate that correct values are returned by the DiskConsumable item
		if (quotaType == null) throw new InvalidUseOfNullException(item+".quotaType");
		if (location == null) throw new InvalidUseOfNullException(item+".location");
		if (bytes < 0) throw new NumberOutOfRangeException(item+".bytes", bytes, 0, false);
		if (quotaType.getId() == SystemItems.getId(QuotaType.TOTAL))
		{
			throw new InvalidDataException("QuotaType.TOTAL is not a valid quota type for item " + item + ".");
		}
		if (location == Location.SECONDARY && !quotaType.getSecondaryLocation())
		{
			throw new InvalidDataException(quotaType.getName()+" cannot use the secondary location");
		}
		
		// Ignore quota check for some locations
		if (location.isQuotaLimited())
		{
			long deltaBytes = bytes;
			if (location.getValue() == currentDiskUsage.getLocation()) 
			{
				if (useDifference)
				{
					deltaBytes = difference;
					bytes = currentDiskUsage.getBytes() + deltaBytes;
				}
				else
				{
					deltaBytes = bytes - currentDiskUsage.getBytes();
				}
			}
			// Ignore quota check if the item now takes less space than before (on the same location)
			if (deltaBytes > 0)
			{
				QuotaTypeData typeTotal = HibernateUtil.loadData(hSession, QuotaTypeData.class, SystemItems.getId(QuotaType.TOTAL));
				QuotaTypeData typeSpecific = quotaType;
				
				/*
					Check the owner's quota
					We use the loadData method since it is possible that the dcData item has been
					loaded in another Hibernate session, which would cause a LazyInitialisationException
					when we call getQuotaValues below.
				*/
				UserData owner = HibernateUtil.loadData(hSession, UserData.class, dcData.getOwner().getId());
				if (owner.getQuota() != null)
				{
					QuotaData ownerQuota = owner.getQuota();

					// Check the owner's total quota
					Long maxTotalQuota = ownerQuota.getQuotaValues().get(new QuotaIndex(typeTotal, location.getValue()));
					if (maxTotalQuota == null)
					{
						maxTotalQuota = Quota.UNDEFINED; 
					}
					if (maxTotalQuota != Quota.UNLIMITED)
					{
						long usedTotalQuota = getDiskUsage(owner, typeTotal, location);
						if (usedTotalQuota >= maxTotalQuota) 
						{
							throw new QuotaException(location, typeTotal, maxTotalQuota, usedTotalQuota);
						}
					}
					// Check the owner's type specific quota if it exists
					Long maxSpecificQuota = ownerQuota.getQuotaValues().get(new QuotaIndex(typeSpecific, location.getValue()));
					if (maxSpecificQuota != null && maxSpecificQuota != Quota.UNLIMITED)
					{
						long usedSpecificQuota = getDiskUsage(owner, typeSpecific, location);
						if (usedSpecificQuota >= maxSpecificQuota) 
						{
							throw new QuotaException(location, typeSpecific, maxSpecificQuota, usedSpecificQuota);
						}
					}
				}

				// If the user belongs to a quota group, check the groups quota
				GroupData quotaGroup = owner.getQuotaGroup();
				if (quotaGroup != null && quotaGroup.getQuota() != null)
				{
					QuotaData groupQuota = quotaGroup.getQuota();
					// Check the group's total quota
					Long maxTotalQuota = groupQuota.getQuotaValues().get(new QuotaIndex(typeTotal, location.getValue()));
					if (maxTotalQuota != null && maxTotalQuota != Quota.UNLIMITED)
					{
						long usedTotalQuota = getDiskUsage(quotaGroup, typeTotal, location);
						if (usedTotalQuota >= maxTotalQuota) 
						{
							throw new QuotaException(location, typeTotal, maxTotalQuota, usedTotalQuota);
						}
					}
					// Check the group's type specific quota if it exists
					Long maxSpecificQuota = groupQuota.getQuotaValues().get(new QuotaIndex(typeSpecific, location.getValue()));
					if (maxSpecificQuota != null && maxSpecificQuota != Quota.UNLIMITED && maxSpecificQuota != Quota.UNDEFINED)
					{
						long usedSpecificQuota = getDiskUsage(quotaGroup, typeSpecific, location);
						if (usedSpecificQuota >= maxSpecificQuota) 
						{
							throw new QuotaException(location, typeSpecific, maxSpecificQuota, usedSpecificQuota);
						}
					}
				}
			}
		}
		
		// Update the disk usage information
		currentDiskUsage.setItemType(dcItem.getType().getValue());
		currentDiskUsage.setBytes(bytes);
		currentDiskUsage.setLocation(location.getValue());
		currentDiskUsage.setQuotaType(quotaType);
		currentDiskUsage.setUser(dcData.getOwner());
		currentDiskUsage.setGroup(dcData.getOwner().getQuotaGroup());
	}

	/**
		Get the disk usage for the specified user at the specified
		location and quota type.
	*/
	long getDiskUsage(UserData user, QuotaTypeData quotaType, Location location)
		throws BaseException
	{
		assert user != null : "user == null";
		assert location != null : "location == null";
		assert quotaType != null : "quotaType == null";
		org.hibernate.Query query;
		if (quotaType.getId() == SystemItems.getId(QuotaType.TOTAL))
		{
			query = HibernateUtil.getPredefinedQuery(hSession, "GET_TOTAL_DISKUSAGE_FOR_USER");
			/*
				SELECT SUM(du.bytes)
				FROM DiskUsageData du
				WHERE du.user = :user
				AND du.location = :location
			*/
		}
		else
		{
			query = HibernateUtil.getPredefinedQuery(hSession, "GET_SPECIFIC_DISKUSAGE_FOR_USER");
			/*
				SELECT SUM(du.bytes)
				FROM DiskUsageData du
				WHERE du.user = :user
				AND du.location = :location
				AND du.quotaType = :quotaType
			*/
			query.setEntity("quotaType", quotaType);
		}
		query.setEntity("user", user);
		query.setInteger("location", location.getValue());
		Long bytes = HibernateUtil.loadData(Long.class, query);
		return bytes == null ? 0 : bytes;
	}

	/**
		Get the disk usage for the specified group at the specified
		location and quota type.
	*/
	long getDiskUsage(GroupData group, QuotaTypeData quotaType, Location location)
		throws BaseException
	{
		assert group != null : "group == null";
		assert location != null : "location == null";
		assert quotaType != null : "quotaType == null";
		org.hibernate.Query query;
		if (quotaType.getId() == SystemItems.getId(QuotaType.TOTAL))
		{
			query = HibernateUtil.getPredefinedQuery(hSession, "GET_TOTAL_DISKUSAGE_FOR_GROUP");
			/*
				SELECT SUM(du.bytes)
				FROM DiskUsageData du
				WHERE du.group = :group
				AND du.location = :location
			*/
		}
		else
		{
			query = HibernateUtil.getPredefinedQuery(hSession, "GET_SPECIFIC_DISKUSAGE_FOR_GROUP");
			/*
				SELECT SUM(du.bytes)
				FROM DiskUsageData du
				WHERE du.group = :group
				AND du.location = :location
				AND du.quotaType = :quotaType
			*/
			query.setEntity("quotaType", quotaType);
		}
		query.setEntity("group", group);
		query.setInteger("location", location.getValue());
		Long bytes = HibernateUtil.loadData(Long.class, query);
		return bytes == null ? 0 : bytes;
	}

	/**
		Create a new item.
		@param itemClass The class of the item		
		@return An object of the <code>itemClass</code> class
		@throws BaseException If there is an error
	*/
	@SuppressWarnings("unchecked") 	
	<I extends BasicItem> I newItem(Class<I> itemClass, Object... extraParameters)
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
				
		Class<? extends BasicData> dataClass = Item.fromItemClass(itemClass).getDataClass();
		assert dataClass != null : "dataClass == null";		
				
		Item itemType = Item.fromDataClass(dataClass);
		Class<? extends I> realItemClass = (Class<? extends I>)itemType.getItemClass();
		Constructor<? extends I> itemConstructor;
		Object[] constructorParams;
		
		try
		{	
			BasicData data = dataClass.newInstance();
			if (extraParameters != null && extraParameters.length > 0)
			{
				constructorParams = new Object[1+extraParameters.length];
				for (int i = 0; i < extraParameters.length; i++)
				{
					constructorParams[1+i] = extraParameters[i];
				}
			}
			else
			{
				constructorParams = new Object[1];				
			}
			constructorParams[0] = data;
			itemConstructor = ClassUtil.findConstructor(realItemClass, constructorParams);
			assert itemConstructor != null : "itemConstructor == null";			
			I item = itemConstructor.newInstance(constructorParams);
			item.setDbControl(this);	
			itemCache.put(data, item);
			return item;			
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Load an item from the database. If the item is already attached to this
		DbControl the same instance is returned, otherwise a new instance is 
		created an attached to this DbControl.
		@param item The item to load
		@return An item that is attached to this DbControl (or null if item is null)
		@since 3.1
	*/
	@SuppressWarnings("unchecked")
	public <I extends BasicItem> I loadItem(I item)
	{
		if (item == null) return null;
		return (I)loadItem(item.getClass(), item.getId());
	}
	
	/**
		Load an item from the database when you now the id. If the item
		has already been loaded the item in the cache will be used.

		@param itemClass The <code>Class</code> of the item object
		@param id The id of the object
		@return An object of the <code>itemClass</code> class
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission
		@throws BaseException If there is another error
	*/
	<I extends BasicItem> I loadItem(Class<I> itemClass, int id)
		throws PermissionDeniedException, BaseException
	{
		Class<? extends BasicData> dataClass = Item.fromItemClass(itemClass).getDataClass();
		return loadItemByDataClass(itemClass, dataClass, id);
	}
	
	/**
		Load an item from the database when you know the id. If the item
		has already been loaded the item in the cache will be used.

		@param itemClass The item class for the object 
		@param dataClass The actual data class for the object which must be compatible 
			with the item class
		@param id The id of the object
		@return An object of the <code>itemClass</code> class
		@since 2.17
	*/
	<I extends BasicItem, D extends BasicData> I loadItemByDataClass(Class<I> itemClass, Class<D> dataClass, int id)
		throws PermissionDeniedException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		BasicData data = HibernateUtil.loadData(hSession, dataClass, id);
		return getItem(itemClass, data);
	}
	

	/**
		Get an item object for a known data object, using the cache if
		possible. If the item already exists in the cache that object
		is returned, otherwise a new item object is created assuming
		that a constructor that takes a single data object as parameter exists.
		
		@param itemClass The <code>Class</code> of the item object
		@param data The data object, or null
		@return An object of the <code>itemClass</code> class, or null if null was passed
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission
		@throws ItemNotFoundException If the data object is a Hibernate proxy
			which points to a non-existing row in the database
		@throws BaseException If there is another error
	*/
	@SuppressWarnings({"unchecked"})
	<I extends BasicItem> I getItem(Class<I> itemClass, BasicData data, Object... extraParameters)
		throws PermissionDeniedException, ItemNotFoundException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		if (data == null) return null;
		
		I item = null;
		try
		{
			Class<? extends BasicData> realDataClass = null;
			// The 'data' object may be a proxy, we need the real data class
			if (data instanceof org.hibernate.proxy.HibernateProxy)
			{
				// We are dealing with a proxy instead of the real data class
				// To get the real instance we...
				
				// 1. Find out the real data-layer class
				realDataClass = 
					(Class<? extends BasicData>)org.hibernate.proxy.HibernateProxyHelper.getClassWithoutInitializingProxy(data);
				
				// 2. Initialise the proxy using the stateful session (the proxy may come from the stateless session)				 
				int id = data.getId();
				org.hibernate.proxy.HibernateProxy hp = (org.hibernate.proxy.HibernateProxy)data;
				org.hibernate.proxy.LazyInitializer li = hp.getHibernateLazyInitializer();
				if (li.isUninitialized())
				{
					li.setSession((org.hibernate.impl.SessionImpl)hSession.getParentSession());
				}
				
				// 3. Get the real instance from the proxy, but it may have pointed to a non-existing row in the database
				try
				{
					data = (BasicData)li.getImplementation();
				}
				catch (org.hibernate.ObjectNotFoundException ex)
				{
					throw new ItemNotFoundException(HibernateUtil.getShortEntityName(ex.getEntityName())+
						"[id="+ex.getIdentifier()+"]");
				}
			}
			else
			{
				realDataClass = (Class<? extends BasicData>)data.getClass();
			}
			
			// Check if the item is in the cache.
			item = itemClass.cast(itemCache.get(data));
			if (item == null)
			{
				// The actual item class may be a subclass of the one specified
				// We use the real data class to find out which real item class to create
				Item itemType = Item.fromDataClass(realDataClass);
				Class<? extends I> realItemClass = (Class<? extends I>)itemType.getItemClass();
				
				// The constructor always take a BasicData as first parameter
				// but may take additional parameters passed in the extraParameters array
				Object[] constructorParams;
				Constructor<? extends I> c;
				if (extraParameters != null && extraParameters.length > 0)
				{
					constructorParams = new Object[1+extraParameters.length];
					for (int i = 0; i < extraParameters.length; ++i)
					{
						constructorParams[i+1] = extraParameters[i];
					}
					c = ClassUtil.findConstructor(realItemClass, constructorParams);
				}
				else
				{
					constructorParams = new Object[1];
					c = (Constructor<? extends I>)itemType.getConstructor();
				}
				constructorParams[0] = data;				
				try
				{
					item = c.newInstance(constructorParams);
					itemCache.put(data, item);
					if (item instanceof Controlled)
					{
						commitQueue.put(item, Transactional.Action.UPDATE);
					}
				}
				catch (Exception ex)
				{
					throw new BaseException(ex);
				}
			}
		}
		catch (org.hibernate.HibernateException hex)
		{
			throw new BaseException(hex);
		}
		// Initialise the new item and check permissions
		item.setDbControl(this);
		item.initPermissions(0, 0);
		item.checkPermission(Permission.READ);
		return item;
	}
	
	/**
		Schedule a new item to be saved in the database. The item is 
		saved the next time the {@link #commit()} method is called.

		@param item The item to be saved
		@throws PermissionDeniedException If the logged in user doesn't have
			create permission
		@throws ItemAlreadyExistsException If the item already exists in the
			database
		@throws BaseException If there is another error
	*/
	public void saveItem(BasicItem item)
		throws PermissionDeniedException, ItemAlreadyExistsException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		if (item.isInDatabase())
		{
			throw new ItemAlreadyExistsException(item.toString());
		}
		item.setDbControl(this);
		item.initPermissions(0, 0);
		item.checkPermission(Permission.CREATE);

		// Check save-if queue
		List<SaveIf> saveIfItems = saveIfQueue == null ? null : saveIfQueue.remove(item);
		
		// Save items with before=TRUE
		if (saveIfItems != null)
		{
			for (SaveIf ifItem : saveIfItems)
			{
				if (ifItem.before && !ifItem.item.isInDatabase())
				{
					saveItem(ifItem.item);
				}
			}
		}
		
		// Save the item
		commitQueue.put(item, Transactional.Action.CREATE);
		
		// Save items with before=FALSE
		if (saveIfItems != null)
		{
			for (SaveIf ifItem : saveIfItems)
			{
				if (!ifItem.before && !ifItem.item.isInDatabase())
				{
					saveItem(ifItem.item);
				}
			}
		}
	}
	
	/**
		Schedule a new item to be saved in the database if another item is also saved.
		If the <code>itemIf</code> item is saved then <code>item</code> is also saved,
		otherwise it is not. The method is useful for utility methods that creates a lot
		of child items to a parent item and we wan't to wait putting them in the real
		save queue until the parent item is saved. If the parent item is already in the
		database or saveItem(itemIf) has been called this method is the same as 
		<code>saveItem(item)</code>
		
		@param itemIf The parent item that must be saved
		@param item The item to save when the parent item is saved
		@param before TRUE if the <code>item</code> should be saved before the <code>itemIf</code>,
			FALSE to save <code>itemIf</code> first
		@see #saveItem(BasicItem)
	*/
	public void saveItemIf(BasicItem itemIf, BasicItem item, boolean before)
	{
		if (itemIf == null || item == null) return;
		if (itemIf.isInDatabase() || commitQueue.containsKey(itemIf))
		{
			saveItem(item);
		}
		else
		{
			if (saveIfQueue == null) saveIfQueue = new HashMap<BasicItem, List<SaveIf>>();
			List<SaveIf> items = saveIfQueue.get(itemIf);
			if (items == null)
			{
				items = new LinkedList<SaveIf>();
				saveIfQueue.put(itemIf, items);
			}
			items.add(new SaveIf(item, before));
		}
	}
	
	/**
		Schedule an existing item to be deleted from the database.
		The item is deleted the next time the {@link #commit()} method is called.
		If you are deleting multiple items it is better to use
		{@link Trashcan#delete(SessionControl, java.util.Collection, boolean, ProgressReporter)}
		since it will automatically select the best order to delete the items.

		@param item The item to be deleted
		@throws PermissionDeniedException If the logged in user doesn't have
			delete permission
		@throws ItemNotFoundException If the item doesn't exist in the
			database
		@throws BaseException If there is another error
	*/
	public void deleteItem(BasicItem item)
		throws PermissionDeniedException, ItemNotFoundException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		if (!item.isInDatabase())
		{
			throw new ItemNotFoundException(item.toString());
		}
		item.setDbControl(this);
		item.initPermissions(0, 0);
		item.checkPermission(Permission.DELETE);
		commitQueue.put(item, Transactional.Action.DELETE);
	}

	/**
		Detach an item from this <code>DbControl</code>. The detached item 
		will no longer be managed and changes will not be saved to the database.
		For some items, certain operations are forbidden for detached items. For example
		you are not allowed to upload files to {@link File} items.
		@param item The item to be detached
		@throws BaseException If there is an error
		@see #reattachItem(BasicItem, boolean)
	*/
	public void detachItem(BasicItem item)
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		HibernateUtil.evictData(hSession, item.getData());
		commitQueue.remove(item);
		itemCache.remove(item.getData());
		item.setDbControl(null);
	}
	
	/**
		Reattach a detached item from this <code>DbControl</code>. If the
		reattached already exists in the database it will be managed and 
		changes will be saved to the database.
		
		@param item The item to be reattached
		@param updated If TRUE, the item has been updated while it was detached
			and changes will be saved to the database. If FALSE, the item will
			only be saved to the database if it is modified after the
			reattachement
		@return The reattached item
		@throws PermissionDeniedException If the logged in user doesn't have
			read/write permission
		@throws BaseException If there is another error
		@see #detachItem(BasicItem)
		@since 2.13
	*/
	public <T extends BasicItem> T reattachItem(T item, boolean updated)
		throws PermissionDeniedException, ItemNotFoundException, BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		item.setDbControl(this);
		if (item.isInDatabase())
		{
			item.initPermissions(0, 0);
			if (updated)
			{
				if (item.hasPermission(Permission.RESTRICTED_WRITE))
				{
					HibernateUtil.updateData(hSession, item.getData());
				}
				else
				{
					throw new PermissionDeniedException(Permission.WRITE, item.toString());
				}
			}
			else
			{
				if (item.hasPermission(Permission.READ))
				{
					HibernateUtil.lockData(hSession, item.getData(), org.hibernate.LockOptions.NONE);
				}
				else
				{
					throw new PermissionDeniedException(Permission.READ, item.toString());
				}
			}
			itemCache.put(item.getData(), item);
			if (item instanceof Controlled)
			{
				commitQueue.put(item, Transactional.Action.UPDATE);
			}
		}
		return item;
	}
	
	/**
		Reload the item from the database. If the item isn't attached to this
		DbControl it is automatically reattached first.
		
		@param item The item to reload
		@since 2.4
	*/
	public void refreshItem(BasicItem item)
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		if (!isAttached(item)) reattachItem(item, false);
		HibernateUtil.refresh(hSession, item.getData());
	}
	
	/**
		Check if an item is attached to this <code>DbControl</code>.
		An item is attached if it is found in either the item cache (existing items 
		only) or commit queue (new or existing items).
	*/
	boolean isAttached(BasicItem item)
	{
		assert item != null : "item == null";
		return commitQueue.containsKey(item) || (itemCache.containsKey(item.getData()) && item.isInDatabase());
	}
	
	/**
		Initialise a collection on an item. This method is useful if you detach
		an item but needs to use a collection while it is detached. Since collections
		are not loaded until they are needed, you will get an exception if you try
		to access an unitialised collection of a detached item. One example is if
		you have detach a {@link Quota} object and calls the 
		{@link Quota#getQuotaValue(QuotaType, Location)} method.
		
		@param item The item 
		@param collectionName The name of the collection
		@throws BaseException If there is an error
		@see #detachItem(BasicItem)
		@see #reattachItem(BasicItem, boolean)
	*/
	public void initCollection(BasicItem item, String collectionName)
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		sc.updateLastAccess();
		HibernateUtil.initCollection(hSession, item.getData(), collectionName);
	}

	/**
		Add a <code>Batcher</code> to the batcher queue.
		@param batcher The <code>Batcher</code> to add
	*/
	void addBatcher(Batcher batcher)
	{
		if (isClosed()) throw new ConnectionClosedException();
		if (batchers == null)
		{
			batchers = new ArrayList<Batcher>();
		}
		assert batcher != null : "batcher is null";
		batchers.add(batcher);
	}
	
	/**
		Remove a <code>Batcher</code> from the batcher queue.
		@param batcher The <code>Batcher</code> to remove
	*/
	void removeBatcher(Batcher batcher)
	{
		if (batcher != null)
		{
			batchers.remove(batcher);
		}
	}
	
	/**
		Clean up if a bad client application forgot to close the connection.
	*/
	protected void finalize()
		throws Throwable
	{
		if (!isClosed()) 
		{
			SessionControl sc = getSessionControl();
			log.warn("Found unclosed DbControl during finalize" + 
				"; isLoggedIn = " + sc.isLoggedIn() +
				"; clientId = " + sc.getClientId() + 
				"; userId = " + sc.getLoggedInUserId(),
				calledFrom
			);
			close();
		}
		super.finalize();
	}
	
	private static class SaveIf
	{
		private final BasicItem item;
		private final boolean before;
		
		private SaveIf(BasicItem item, boolean before)
		{
			this.item = item;
			this.before = before;
		}
	}
	
	
}
