/**
	$Id: HqlQueryJoiners.java 6026 2012-03-23 07:57:12Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.query.AnnotationExpression;
import net.sf.basedb.core.query.Annotations;
import net.sf.basedb.core.query.AutoJoiner;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.HqlQuery;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.ReporterListExpression;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;

/**
	This class collects <code>AutoJoiner</code>:s for <code>HqlQuery</code>
	queries. The implementations are thread-safe and can be re-used between threads.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2012-03-23 08:57:12 +0100 (Fri, 23 Mar 2012) $
*/
public final class HqlQueryJoiners
{

	/**
		This instance joins reporter lists with a left join, unless the
		{@link ReporterListExpression#getAutoJoinType()} method overrides 
		the default.
	*/
	public static final ReporterListJoiner REPORTER_LIST_LEFT = 
		new ReporterListJoiner(JoinType.LEFT);
	
	/**
		This instance joins reporter lists with an inner join, unless the
		{@link ReporterListExpression#getAutoJoinType()} method overrides 
		the default.
	*/
	public static final ReporterListJoiner REPORTER_LIST_INNER = 
		new ReporterListJoiner(JoinType.INNER);
	
	/**
		Get a joiner instance for raw data for the specified join type.
		@param joinType The type of join
		@return {@link #REPORTER_LIST_LEFT} or {@link #REPORTER_LIST_INNER}
	 */
	public static ReporterListJoiner getReporterListJoiner(JoinType joinType)
	{
		return joinType == JoinType.LEFT ? REPORTER_LIST_LEFT : REPORTER_LIST_INNER;
	}
	
	/**
		Get an annotation auto joiner. Do not use the same auto-joiner for more
		than one query, since it will only join a given annotation type once.
	*/
	public static AnnotationJoiner getAnnotationJoiner(JoinType joinType)
	{
		return new AnnotationJoiner(joinType);
	}

	/**
		Joiner for reporter lists.
		@author nicklas
		@version 2.8
		@base.modified $Date: 2012-03-23 08:57:12 +0100 (Fri, 23 Mar 2012) $
	*/
	private static class ReporterListJoiner
		implements AutoJoiner<HqlQuery, ReporterListExpression>
	{
		private final JoinType joinType;
	
		private ReporterListJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		public Class<HqlQuery> getQueryClass()
		{
			return HqlQuery.class;
		}
		public Class<ReporterListExpression> getElementClass()
		{
			return ReporterListExpression.class;
		}
		public void joinIfNeeded(HqlQuery query, ReporterListExpression element)
		{
			JoinType jt = element.getAutoJoinType();
			String tableAlias = RealTable.REPORTER_LIST_SCORES.getAlias(element.getReporterList());
			Restriction on = Restrictions.eq(
					Hql.property(tableAlias, "reporterList.id"), 
					Expressions.integer(element.getReporterList().getId()
				)
			);
			if (jt == JoinType.LEFT)
			{
				query.join(Hql.leftJoin(element.getJoinAlias(), element.getJoinProperty(), 
					tableAlias, on, false));
			}
			else
			{
				query.join(Hql.innerJoin(element.getJoinAlias(), element.getJoinProperty(), 
					tableAlias, on, false));
			}
		}
		@Override
		public void reset(HqlQuery query)
		{}
		// -------------------------------------------
	}
	
	/**
		Joiner for annotations.
		@since 3.2
	*/
	private static class AnnotationJoiner
		implements AutoJoiner<HqlQuery, AnnotationExpression>
	{
		private final JoinType joinType;
		private final Set<AnnotationType> hasJoined;
		
		private AnnotationJoiner(JoinType joinType)
		{
			this.joinType = joinType;
			this.hasJoined = new HashSet<AnnotationType>();
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<HqlQuery> getQueryClass()
		{
			return HqlQuery.class;
		}
		@Override
		public Class<AnnotationExpression> getElementClass()
		{
			return AnnotationExpression.class;
		}
		@Override
		public void joinIfNeeded(HqlQuery query, AnnotationExpression element)
		{
			AnnotationType at = element.getAnnotationType();
			if (!hasJoined.add(at)) return;
			
			if (joinType == JoinType.LEFT)
			{
				query.join(Annotations.leftJoin(element.getAlias(), at, element.getJoinAlias()));
			}
			else
			{
				query.join(Annotations.innerJoin(element.getAlias(), at, element.getJoinAlias()));
			}
		}
		@Override
		public void reset(HqlQuery query)
		{
			hasJoined.clear();
		}
		// -------------------------------------------
		
	}

}
