/*
	$Id: InternalStorageController.java 5689 2011-08-11 09:09:06Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.FileUtil;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Set;
import java.util.Map;
import java.util.Collections;

/**
	A simple implementation of a secondary storage controller which just moves
	files to another directory. In the secondary storage the files are named
	according to their ID in the database. This is neccessary because we need
	to be able to detect if a file has been deleted from the database or not.
	<p>
	The configuration
	is done in the <code>base.config</code> file with the following settings:
	
	<table>
	<tr>
		<th>Setting</th>
		<th>Default/possible value(s)</th>
		<td>Description</th>
	</tr>
	<tr>
		<td>secondary.storage.driver</td>
		<td>net.sf.basedb.core.InternalStorageController</td>
		<td>
			The class name of the secondary storage controller.
		</td>
	</tr>
	<tr>
		<td>secondary.storage.init</td>
		<td>-</td>
		<td>
			The path to the seconday storage directory.
		</td>
	</tr>
	</table>

	<p>
	This class logs information to a logger named <code>net.sf.basedb.core.storage</code>.
	Configure the settings for the logger in the <code>log4j.properties</code> file.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
*/
final class InternalStorageController
	implements SecondaryStorageController
{

	/**
		Log storage events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.storage.InternalStorageController");

	/**
		The session controller. We are logging in using our
		own keyring: {@link StorageKeyring}.
	*/
	private SessionControl sc;
	
	/**
		The path to the directory where the secondary storage is located.
	*/
	private java.io.File path;

	/**
		Create the storage controller. The one and only instance is created
		by Application.start
	*/
	InternalStorageController()
	{}

	/*
		From the SecondaryStorageController class
		-------------------------------------------
	*/
	
	public void init(String settings)
		throws BaseException 
	{
		log.info("Initialising InternalStorageController...");
		path = new java.io.File(settings);
		if (!path.exists())
		{
			throw new BaseException("Path not found: " + settings);
		}
		sc = new SessionControl(null, "net.sf.basedb.clients.internal", false, Application.getHostName());
		sc.login(new StorageKeyring());
		log.info("InternalStorageController has been initialised");
		log.info("path="+path);
	}
	
	/**
		Check the database for files for waiting to be moved.
	*/
	public void run()
	{
		log.info("Checking for files to move");
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			ItemQuery<File> fileQuery = File.getQuery();
			fileQuery.restrict(
				Restrictions.neq(
					Hql.property("action"),
					Expressions.integer(File.Action.NOTHING.getValue())
				)
			);
			fileQuery.include(Include.OTHERS);
			fileQuery.order(Orders.asc(Hql.property("owner")));
			ItemResultIterator<File> filesToMove = fileQuery.iterate(dc);
			while (filesToMove.hasNext())
			{
				File file = filesToMove.next();
				File.Action action = file.getAction();
				try
				{
					if (action == File.Action.MOVE_TO_PRIMARY)
					{
						moveToPrimary(file.getId());
						log.info("File '" + file + "' moved to primary storage");
					}
					else if (action == File.Action.MOVE_TO_SECONDARY)
					{
						moveToSecondary(file.getId());
						log.info("File '" + file + "' moved to secondary storage");
					}
					dc.detachItem(file); // Otherwise the commit will overwrite the DiskUsage information
				}
				catch (Throwable t)
				{
					// Cancel the move and send a message to send to the user
					log.error("File '" + file + "' could not be moved", t);
					file.setAction(File.Action.NOTHING);
					User owner = file.getOwner();
					Message message = Message.getNewSystemMessage(dc, owner, "InternalStorageController", null, null);
					message.setName("File '" + file + "' could not be moved.");
					StringWriter sw = new StringWriter();
					PrintWriter pw = new PrintWriter(sw);
					pw.println(t.getMessage());
					t.printStackTrace(pw);
					String description = sw.toString().replaceAll("net.sf.basedb.core.", "");
					message.setDescription(description);
					dc.saveItem(message);
				}
			}
			dc.commit();
			
			// Check if all files in the secondary directory are present in the database
			// and if they have their Location set to SECONDARY, otherwise we delete the
			// file in the secondary directory
			java.io.File[] allSecondary = path.listFiles();
			dc = sc.newDbControl();
			for (java.io.File file : allSecondary)
			{
				int fileId = 0;
				try
				{
					// The filename is the ID in the database
					fileId = Integer.parseInt(file.getName());
					boolean delete = false;
					try
					{
						File f = File.getById(dc, fileId);
						Location l = f.getLocation();
						delete = l != Location.SECONDARY;
					}
					catch (ItemNotFoundException ex)
					{
						delete = true;
					}
					if (delete) 
					{
						if (file.delete())
						{
							log.info("Deleted orphaned file with ID: " + fileId);
						}
						else
						{
							log.warn("Failed to delete orphaned file with ID: " + fileId);
						}
					}
				}
				catch (Throwable t)
				{
					log.error("Exception while checking for orphaned files", t);
				}
			}
			dc.close();
		}
		catch (Throwable ex)
		{
			log.error("Exception while checking for files to move", ex);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	/**
		Cleanup
	*/
	public void close()
	{
		log.info("Closing InternalStorageController...");
		sc.logout();
		sc = null;
		log.info("InternalStorageController has been closed");
	}
	// -------------------------------------------
	
	private void moveToPrimary(int fileId)
		throws Throwable
	{
		DbControl dc = null;
		java.io.File f = new java.io.File(path, Integer.toString(fileId));
		try
		{
			dc = sc.newDbControl();
			File file = File.getById(dc, fileId);
			OutputStream toPrimary = file.getUploadStream(true);
			InputStream fromSecondary = FileUtil.getInputStream(f);
			FileUtil.copy(fromSecondary, toPrimary);
			toPrimary.close();
			fromSecondary.close();
			file.setLocation(Location.PRIMARY);
			file.setAction(File.Action.NOTHING);
			dc.commit();
			// Remove file from secondary storage
			if (!f.delete())
			{
				log.warn("Failed to remove secondary file: " + fileId);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	private void moveToSecondary(int fileId)
		throws Throwable
	{
		DbControl dc = null;
		java.io.File f = new java.io.File(path, Integer.toString(fileId));
		try
		{
			dc = sc.newDbControl();
			File file = File.getById(dc, fileId);
			InputStream fromPrimary = file.getDownloadStream(0);
			OutputStream toSecondary = FileUtil.getOutputStream(f);
			FileUtil.copy(fromPrimary, toSecondary);
			fromPrimary.close();
			toSecondary.close();
			file.setLocation(Location.SECONDARY);
			file.setAction(File.Action.NOTHING);
			dc.commit();
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/**
		Our own keyring containing the permissions we need to move the files.
		The required permissions are: 
		<ul>
		<li>{@link Permission#WRITE} for {@link Item#FILE}
		<li>{@link Permission#READ} for {@link Item#USER}
		<li>{@link Permission#CREATE} for {@link Item#MESSAGE}
		</ul>
		All other permissions are denied.
	*/
	private static class StorageKeyring
		extends Keyring
	{
	
		private final Map<Item, Integer> permissions;
		private final Set<Integer> emptySet;
		private final Map<Integer, Integer> emptyMap;
	
		private StorageKeyring()
		{
			super(0);
			this.permissions =  new HashMap<Item, Integer>();
			this.permissions.put(Item.FILE, Permission.grant(Permission.WRITE));
			this.permissions.put(Item.USER, Permission.grant(Permission.READ));
			this.permissions.put(Item.MESSAGE, Permission.grant(Permission.CREATE));
			this.emptySet = Collections.emptySet();
			this.emptyMap = Collections.emptyMap();
		}
		/*
			From the Keyring class
			-------------------------------------------
		*/
		/**
			Never reload since everything is static.
		*/
		boolean getReload()
		{
			return false;
		}
		/**
			@see StorageKeyring
		*/
		int getRolePermissions(Item itemType)
		{
			Integer permission = permissions.get(itemType);
			return permission == null ? 0 : permission.intValue();
		}
		/**
			Return 0.
		*/
		int getSharedPermissions(ShareableData sharableData)
		{
			return 0;
		}
		/**
			Return 0.
		*/
		int getAllPermissions(Item itemType, UserData owner, ItemKeyData itemKey, ProjectKeyData projectKey)
		{
			return 0;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getRoles()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getGroups()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getUsers()
		{
			return emptySet;
		}
		/**
			Return empty map.
		*/
		Map<Integer, Integer> getProjects()
		{
			return emptyMap;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getItemKeys()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getProjectKeys()
		{
			return emptySet;
		}
		// -------------------------------------------
	
	}
}
