/**
	$Id: JobAgentSettings.java 5595 2011-03-17 12:47:15Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.JobAgentSettingsData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class contains a job agent's settings for a specific plugin.
	The settings may override the default settings for the JAR path
	where to plugin is located, maximum memory to use, and if the plugin
	is trusted or not. 

	@author nicklas
	@version 2.0
	@base.modified $Date: 2011-03-17 13:47:15 +0100 (Thu, 17 Mar 2011) $
*/
public class JobAgentSettings
	extends ChildItem<JobAgentSettingsData>
{

	/**
		The type of item represented by this class.
		@see Item#JOBAGENTSETTINGS
		@see #getType()
	*/
	public static final Item TYPE = Item.JOBAGENTSETTINGS;

	/**
	 * Create a new <code>JobAgentSettings</code> item.
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param agent The job agent these settings belong to
		@param plugin The plugin definition these settings are for
		@return The new <code>JobAgentSettings</code> item
		@throws BaseException If there is another error
	*/
	static JobAgentSettings getNew(DbControl dc, JobAgent agent, PluginDefinition plugin)
	{
		JobAgentSettings settings = dc.newItem(JobAgentSettings.class);
		settings.setJobAgent(agent);
		settings.setPluginDefinition(plugin);
		return settings;
	}
	
	/**
		Get a <code>JobAgentSettings</code> item when you know the ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param id The ID of the item to load
		@return The <code>JobAgentSettings</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the item
		@throws BaseException If there is another error
	*/
	public static JobAgentSettings getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		JobAgentSettings jas = dc.loadItem(JobAgentSettings.class, id);
		if (jas == null) throw new ItemNotFoundException("JobAgentSettings[id="+id+"]");
		return jas;
	}
	
	/**
		Get a query configured to retrieve settings for the specified agent.
		
		@param agent The job agent to retreive settings for, null is not allowed
		@return An {@link ItemQuery} object
		@see JobAgent#getSettings()
	*/
	static ItemQuery<JobAgentSettings> getQuery(JobAgent agent)
	{
		ItemQuery<JobAgentSettings> query = 
			new ItemQuery<JobAgentSettings>(JobAgentSettings.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("jobAgent"), 
				Hql.entity(agent)
			)
		);
		return query;
	}
	
	/**
		This filter will only return items if the logged in user has
		generic read permission to job agents.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.JOBAGENT);
	
	/**
		Get a query configured to retrieve settings for the specified plugin.
		This query only return items if the logged in user has generic read
		permission for all job agents.
		
		@param plugin The plugin to retreive settings for, null is not allowed
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<JobAgentSettings> getQuery(PluginDefinition plugin)
	{
		ItemQuery<JobAgentSettings> query = new ItemQuery<JobAgentSettings>(
			JobAgentSettings.class, RUNTIME_FILTER);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("pluginDefinition"), 
				Hql.entity(plugin)
			)
		);
		return query;
	}

	JobAgentSettings(JobAgentSettingsData data)
	{
		super(data);
	}
	
	/*
	 From the Identifiable interface
	 -------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
	 From the ChildItem class
	 -------------------------------------------
	*/
	SharedData getSharedParent()
	{
		return getData().getJobAgent();
	}
	// -------------------------------------------

	/**
		Get the job agent that owns these settings.
		
		@return The <code>JobAgent</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public JobAgent getJobAgent()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(JobAgent.class, getData().getJobAgent());
	}
	
	private void setJobAgent(JobAgent agent)
	{
		getData().setJobAgent(agent.getData());
	}
	
	/**
		Get the plugin that these settings are used for.
		
		@return The <code>PluginDefinition</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public PluginDefinition getPluginDefinition()
	{
		return getDbControl().getItem(PluginDefinition.class, getData().getPluginDefinition());
	}
	
	private void setPluginDefinition(PluginDefinition plugin)
	{
		getData().setPluginDefinition(plugin.getData());
	}
	
	/**
		If the plugin is a trusted plugin or not. If null the trusted flag returned by
		the plugin definition should be used.
		@return A boolean vaue or null
		@see PluginDefinition#isTrusted()
		@see #isEffectivelyTrusted()
	*/
	public Boolean isTrusted()
	{
		return getData().isTrusted();
	}
	
	/**
		If the plugin is a trusted plugin or not. If these settings hasn't configured
		a value the trusted flag returned by the plugin definition is used.
		@return TRUE if the plugin is trusted, FALSE otherwise
		@see PluginDefinition#isTrusted()
		@see #isTrusted()
	*/
	public boolean isEffectivelyTrusted()
	{
		Boolean trusted = getData().isTrusted();
		if (trusted == null)
		{
			trusted = getData().getPluginDefinition().isTrusted();
		}
		return trusted.booleanValue();
	}
	
	/**
		Set the trusted flag for this plugin.
		@param trusted A boolean value or null if the plugin definition setting should be used
		@see PluginDefinition#isTrusted()
	*/
	public void setTrusted(Boolean trusted)
	{
		checkPermission(Permission.WRITE);
		getData().setTrusted(trusted);
	}
	
	/**
		Get the maximum amount of memory the plugin is allowed to use. If null
		the max memory returned by the plugin definition should be used.
		@return The maximum amount of memory or null
		@see PluginDefinition#getMaxMemory()
		@see #getEffectiveMaxMemory()
	*/
	public Long getMaxMemory()
	{
		return getData().getMaxMemory();
	}
	
	/**
		Get the effective maximum amount of memory the plugin is allowed to use.
		If these settings hasn't configured an amount, the max memory from the
		plugin definition is returned.
		@return The maximum memory or null if no maximum has been specified
		@see PluginDefinition#getMaxMemory()
		@see #getMaxMemory()
	*/
	public Long getEffectiveMaxMemory()
	{
		Long maxMemory = getData().getMaxMemory();
		if (maxMemory == null)
		{
			maxMemory = getData().getPluginDefinition().getMaxMemory();
		}
		return maxMemory;
	}
	
	/**
		Set maximim amount of memory the plugin is allowed to use.
		@param maxMemory The maximum amount or null to use the amount specified
			by the plugin definition
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is less than zero
	*/
	public void setMaxMemory(Long maxMemory)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (maxMemory != null && maxMemory < 0)
		{
			throw new NumberOutOfRangeException("maxMemory", maxMemory, 0, false);
		}
		getData().setMaxMemory(maxMemory);
	}
	
	/**
		Get the priority boost given to jobs executed by this plugin. The priority
		boost is subtracted from the value job's priority value, resulting in a lower
		priority value for the job. Jobs with lower values are executed befor jobs
		with higher values.
		<p>
		The priority boost is useful if we, for example, want to use one server mainly for
		importing data. By giving all import plugins a priority boost they will be executed
		before all other jobs.
		
		@see Job#getPriority()
		@return The priority boost, a value between 0 and 10
	*/
	public int getPriorityBoost()
	{
		return getData().getPriorityBoost();
	}
	
	/**
		Set the priority boost given to jobs executed by this plugin.
		@param priorityBoost The priority boost, must be between 0 and 10
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is less than 0 or greater than 10
	*/
	public void setPriorityBoost(int priorityBoost)
	{
		checkPermission(Permission.WRITE);
		getData().setPriorityBoost(IntegerUtil.checkMinMax(priorityBoost, "priorityBoost", 0, 10));
	}
	
	/**
		Get the effective priority for the specified job, when the priority boost is
		taken into account.
		@param job The job to get the prioriy for
		@return The effective priority
		@throws InvalidDataException If the job is null or has a different plugin
			definition than these settings
		@see Job#getPriority()
	*/
	public int getEffectivePriority(Job job)
		throws InvalidDataException
	{
		if (job == null) throw new InvalidUseOfNullException("job");
		if (!getData().getPluginDefinition().equals(job.getData().getPluginDefinition()))
		{
			throw new InvalidDataException("The job " + job + 
				" has a different plugin than these settings.");
		}
		return job.getPriority()-getPriorityBoost();
	}
	
}
