/*
	$Id: Path.java 4951 2009-05-27 10:07:18Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.regex.Pattern;

/**
	This class is used represent the path to a {@link Directory}
	or {@link File} item. Some examples:

	<table>
	<tr><td>/</td><td>The root diretory</td></tr>
	<tr><td>/filename</td><td>A file in the root diretory</td></tr>
	<tr><td>/directory</td><td>A directory in root diretory</td></tr>
	<tr><td>/directory/filename</td><td>A file in a directory</td></tr>
	<tr><td>/directory/subdirectory</td><td>A subdirectory</td></tr>
	<tr><td>/directory/subdirectory/filename</td><td>A file in a subdirectory</td></tr>
	<tr><td>~</td><td>The currently logged in user's home directory</td></tr>
	<tr><td>~/filename</td><td>A file in the current user's home directory</td></tr>
	<tr><td>~/directory</td><td>A directory in the current user's home directory</td></tr>
	<tr><td>~login</td><td>The home directory of the user with the specified login</td></tr>
	<tr><td>~login/filename</td><td>A file in the home directory of the user with the specified login</td></tr>
	<tr><td>~login/directory</td><td>A directory in the home directory of the user with the specified login</td></tr>
	</table>


	@author enell, nicklas
	@version 2.0
	@base.modified $Date: 2009-05-27 12:07:18 +0200 (Wed, 27 May 2009) $

*/
public class Path
{
	/**
		These characters are not valid within a file or directory name:
		{@value}
	*/
	public static final String INVALID_CHARACTERS = "~\\/:;*?<>|\"";

	/**
		A regexp checking for invalid characters.
	*/
	private static final Pattern invalid = Pattern.compile("[\\~\\\\\\/\\:\\;\\*\\?\\<\\>\\|\\\"]+");

	/**
		Check if a name is valid as a partial path name, ie. it must not contain any
		of the characters in {@link #INVALID_CHARACTERS}.
		
		@param name The string to be tested.
		@return TRUE if this name is valid.
	*/
	public static boolean isValidName(String name)
	{
		return !invalid.matcher(name).find();
	}

	/**
		Convert a string to a name that is safe to use as a filename.
		It will replace all characters mathing any of the 
		{@link #INVALID_CHARACTERS} charactes with the specified string.
		
		@param originalName The original name
		@param replaceWith The string to replace all unsafe characters with
		@since 2.2
	*/
	public static String makeSafeFilename(String originalName, String replaceWith)
	{
		String safeFilename = invalid.matcher(originalName).replaceAll(replaceWith);
		return safeFilename;
	}
	
	
	/**
		The type of the path.
	*/
	private Path.Type type;

	/**
		The userlogin part of the path.
	*/
	private String userlogin;

	/**
		The directory part of the path.
	*/
	private String[] directories;

	/**
		The filename part of the path.
	*/
	private String filename;

	/**
		Create a new <code>Path</code> object by parsing the given string.
		@param path The string representation of a path
		@param type If the path represents a {@link Type#FILE} or
			{@link Type#DIRECTORY}
		@throws InvalidPathException If the path is not valid
	*/
	public Path(String path, Path.Type type)
		throws InvalidPathException
	{
		this.type = type;

		if (!path.startsWith("/") && !path.startsWith("~"))
		{
			throw new InvalidPathException("Path must begin with a / or ~: "+path);
		}
		String[] parts = path.split("/");
		if (path.startsWith("~"))
		{
			userlogin = parts[0].substring(1);
			if (userlogin == null) userlogin = "";
			parts[0] = "";
		}
		if (type == Type.FILE)
		{
			if (parts.length < 2)
			{
				throw new InvalidPathException("Path must contain a filename: "+path);
			}
			filename = parts[parts.length-1];
			if (!Path.isValidName(filename)) 
			{
				throw new InvalidPathException("Path contains invalid characters ("+Path.INVALID_CHARACTERS+"): "+path);
			}
			parts[parts.length-1] = "";
		}
		
		int i = 0;
		for (String part : parts)
		{
			if (!"".equals(part))
			{
				if (!Path.isValidName(part)) 
				{
					throw new InvalidPathException("Path contains invalid characters ("+Path.INVALID_CHARACTERS+"): "+path);
				}
				parts[i] = part;
				i++;
			}
		}
		directories = new String[i];
		System.arraycopy(parts, 0, directories, 0, i);
	}

	/**
		Create a new <code>Path</code> object.
		@param userlogin The userlogin part of the path
		@param directories The directories in the path
		@param filename The filename part of the path
	*/
	Path(String userlogin, String[] directories, String filename)
	{
		this.userlogin = userlogin;
		this.directories = directories;
		this.filename = filename;
		this.type = (filename == null) ? Path.Type.DIRECTORY : Path.Type.FILE;
	}

	/**
		Create a new <code>Path</code> object.
		
		@param directoryPath The directory path
		@param filename The filename part of the path
	*/
	Path(Path directoryPath, String filename)
	{
		this(directoryPath.userlogin, directoryPath.directories, filename);
	}

	/**
		Get the parent path.
		@return The path to the parent directory, or null if this is the
			root directory
	*/
	public Path getParent()
	{
		Path newPath = null;
		if (type == Type.FILE)
		{
			return new Path(userlogin, directories, null);
		}
		else if (directories.length > 0)
		{
			String[] parentDirectories = new String[directories.length - 1];
			System.arraycopy(directories, 0, parentDirectories, 0, parentDirectories.length);
			newPath = new Path(userlogin, parentDirectories, null);
		}
		return newPath;
	}
	
	/**
		Get the type of the path.
		@return The {@link Path.Type} of the Path
	*/
	public Path.Type getType()
	{
		return type;
	}

	/**
		Get the userlogin part of the path. An empty string
		represents the currently logged in user. Null is returned if the path
		doesn't contain a user login.
		@return The login of the user
	*/
	public String getUserLogin()
	{
		return userlogin;
	}

	/**
		Get the number of directories on the path.
		@return The number of directories
	*/
	public int getDirectoryCount()
	{
		return directories == null ? 0 : directories.length;
	}

	/**
		Get the name of the i:th directory on the path, starting with 0.
		@param i Should be between 0 and {@link #getDirectoryCount() getDirectoryCount-1}
		@return The name of the directory
	*/
	public String getDirectory(int i)
	{
		return directories[i];
	}

	/**
		Get the filename part of the path.
		@return The name of the file
	*/
	public String getFilename()
	{
		return filename;
	}

	/**
		Get the string representation of the path.
		
		@return The string representation of this path.
	*/
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		if (userlogin != null)
		{
			sb.append("~").append(userlogin);
		}
		for (String directory : directories)
		{
			sb.append("/").append(directory);
		}
		if (filename != null && getType() == Type.FILE)
		{
			sb.append("/").append(filename);
		}
		if (sb.length() == 0) sb.append("/"); // For the root directory
		return sb.toString();
	}
	
	/**
		Get this Path's URL-string encoded with a given 
		character encoding.
		
		@param enc The encoding to use. UTF-8 is used is this is null.
		@return The encoded URL-string.
	  	@throws UnsupportedEncodingException If the give encoding is not supported
	  	@since 2.11.1
	 */
	public String toURLString(String enc) 
		throws UnsupportedEncodingException
	{
		StringBuilder sb = new StringBuilder();
		enc = (enc == null || enc.length() == 0) ? "UTF-8" : enc;
		if (userlogin != null)
		{
			sb.append("~").append(URLEncoder.encode(userlogin, enc));
		}
		for (String directory : directories)
		{
			sb.append("/").append(URLEncoder.encode(directory, enc));
		}
		if (filename != null && getType() == Type.FILE)
		{
			sb.append("/").append(URLEncoder.encode(filename, enc));
		}
		if (sb.length() == 0) sb.append("/"); // For the root directory
		return sb.toString();
	}

	/**
		An enum that describes what type the path is
	 
		@author enell
		@version 2.0
	*/
	public enum Type
	{
		/**
			The path is a path to a {@link File}.
		*/
		FILE, 
		
		/**
			The path is a path to a {@link Directory}.
		*/
		DIRECTORY;
	}
}
