/**
	$Id: EntityDetails.java 6080 2012-08-07 11:28:40Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;

import org.hibernate.EntityMode;
import org.hibernate.collection.PersistentCollection;
import org.hibernate.type.Type;
import org.hibernate.type.VersionType;

/**
	Contains detailed information about the changes made to an
	entity during a transaction. This information is sent to
	an {@link EntityLogger} that is responsible for the actual
	logging.
	<p>
	NOTE! The information in this object is more or less drawn from
	what Hibernate reveals about the entity. It is recommended that
	no information is changed.

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2012-08-07 13:28:40 +0200 (Tue, 07 Aug 2012) $
*/
public class EntityDetails
{

	private final Object entity;
	private final ChangeType changeType;
	private final Object[] state;
	private final Object[] previousState;
	private final String[] properties;
	private final Type[] types;
	private final int[] dirty;
	
	public EntityDetails(Object entity, ChangeType changeType,
			Object[] state, Object[] previousState, String[] properties, Type[] types)
	{
		this.entity = entity;
		this.changeType = changeType;
		this.state = state;
		this.previousState = previousState;
		this.properties = properties;
		this.types = types;
		this.dirty = new int[types.length]; // 0 = not known, -1 = not dirty, 1 = dirty
		findDirtyCollections();
	}
		
	/**
		Get the entity that was changed.
	*/
	public Object getEntity()
	{
		return entity;
	}
	
	/**
		Was it a creation/update/deletion?
	*/
	public ChangeType getChangeType()
	{
		return changeType;
	}
	
	/**
		Get an array with the current state of the entity.
	*/
	public Object[] getState()
	{
		return state;
	}
	
	/**
	 	Get an array with the previous state of the entity. The
	 	array is null for new items and is identical to the array
	 	returned by {@link #getState()} for delete events.
	*/
	public Object[] getPreviousState()
	{
		return previousState;
	}

	/**
		Get an array with the names of all properties.
	*/
	public String[] getProperties()
	{
		return properties;
	}
	
	/**
		Get an array representing the data type for each property.
	*/
	public Type[] getTypes()
	{
		return types;
	}
	
	public int getDirty(int index)
	{
		if (changeType != ChangeType.UPDATE) return -1;
		if (dirty[index] == 0)
		{
			Object c = state[index];
			Object o = previousState[index];
			dirty[index] = types[index].isEqual(c, o, EntityMode.POJO) ? -1 : 1;
		}
		return dirty[index];
	}
	
	/**
		Get all modified properties as a string, separating each
		property name with the given separator.
		@param prefix A string that is added first in the result, or null
		@param separator A separator string (if null, comma is used)
		@param suffix A string that is added last in the result, or null
		@return A string with the modified properties, or null
			if this is a CREATE or DELETE change or if no modified 
			properties was found
	*/
	public String getModifiedProperties(String prefix, String separator, String suffix)
	{
		if (changeType != ChangeType.UPDATE) return null;
		
		StringBuilder modified = new StringBuilder();
		if (prefix != null) modified.append(prefix);
		if (separator == null) separator = ",";
		String s = "";
		int numModified = 0;
		for (int i = 0; i < types.length; ++i)
		{
			if (dirty[i] == 0)
			{
				Type t = types[i];
				if (!t.isCollectionType() && !(t instanceof VersionType))
				{
					Object c = state[i];
					Object o = previousState[i];
					dirty[i] = t.isEqual(c, o, EntityMode.POJO) ? -1 : 1;
				}
			}
			if (dirty[i] == 1)
			{
				modified.append(s).append(properties[i]);
				s = separator;
				numModified++;
			}
		}
		if (suffix != null) modified.append(suffix);
		return numModified == 0 ? null : modified.toString();
	}
	
	/**
		Create a new ChangeHistoryDetailData object from the information
		in this object assuming that the entity is a subclass of
		{@link BasicData}.
		@param detailedProperties If TRUE, the {@link ChangeHistoryDetailData#getChangeInfo()}
			is populated with a list of the names of the modified properties (only
			relevant for UPDATE changes)
		@return A new ChangeHistoryDetailData instance
	*/
	public ChangeHistoryDetailData toChangeHistoryDetailData(boolean detailedProperties)
	{
		return toChangeHistoryDetailData((BasicData)entity, changeType, detailedProperties);
	}

	/**
		Create a new ChangeHistoryDetailData object using a mix of the information
		in this object, and the given parameters.
		@param data The entity that was changed
		@param changeType The change type
		@param detailedProperties If TRUE, the {@link ChangeHistoryDetailData#getChangeInfo()}
			is populated with a list of the names of the modified properties (only
			relevant for UPDATE changes)
		@return  A new ChangeHistoryDetailData instance
	*/
	public ChangeHistoryDetailData toChangeHistoryDetailData(BasicData data, ChangeType changeType, boolean detailedProperties)
	{
		ChangeHistoryDetailData details = new ChangeHistoryDetailData();
		details.setChangeType(changeType.getValue());
		details.setItemId(data.getId());
		details.setItemType(Item.fromDataObject(data).getValue());
		if (detailedProperties)
		{
			details.setChangeInfo(getModifiedProperties("Updated: ", ", ", null));
		}
		return details;
	}
	
	/**
		We must do this from the constructor since the "dirtyness" of a
		collection is reset after flush the SQL to the database, which is 
		completed before the rest of the logging is done.
	*/
	private void findDirtyCollections()
	{
		for (int i = 0; i < types.length; ++i)
		{
			Type t = types[i];
			if (t.isCollectionType())
			{
				Object c = state[i];
				if (c instanceof PersistentCollection)
				{
					dirty[i] = ((PersistentCollection)c).isDirty() ? 1 : -1;
				}
			}
		}
	}
	
}
