/*
	$Id $

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Collection;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.util.EqualsHelper;

/**
	Join annotations values to the query. The value is available as
	the joined alias.
	<p>
	Throws an <code>UnsupportedOperationException</code> if 
	{@link Query#getQueryType()} isn't {@link QueryType#HQL}.

	@author Nicklas
	@since 3.2
	@see Annotations#innerJoin(AnnotationType, String)
	@see Annotations#leftJoin(AnnotationType, String)
	@base.modified $Date$
*/
public class AnnotationJoin
	implements Join
{
	private final JoinType joinType;
	private final String alias;
	private final AnnotationType annotationType;
	private final String joinedAlias;

	AnnotationJoin(JoinType joinType, String alias, AnnotationType annotationType, String joinedAlias)
	{
		assert joinType != null : "joinType == null";
		assert annotationType != null : "annotationType == null";
		assert joinedAlias != null : "joinedAlias == null";
		this.joinType = joinType;
		this.alias = alias;
		this.annotationType = annotationType;
		this.joinedAlias = joinedAlias;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.HQL)
		{
			String tmpAlias = "atmp" + annotationType.getId();
			String rootAlias = alias == null ? query.getRootAlias() : alias;
			String valuesProperty = StringUtil.decapitalize(annotationType.getValueType().getTableName());
			
			StringBuilder hql = new StringBuilder();
			// JOIN <root>.annotations atmp WITH atmp.annotationTypeId=<id>
			hql.append(joinType.getJoinExpression()).append(" ").append(rootAlias).append(".annotations ").append(tmpAlias);
			hql.append(" WITH ").append(tmpAlias).append(".annotationTypeId=").append(annotationType.getId());
			// JOIN atmp.fooValues <alias>
			hql.append(" ");
			hql.append(joinType.getJoinExpression()).append(" ").append(tmpAlias).append(".").append(valuesProperty).append(" ").append(joinedAlias);
			return hql.toString();
		}
		else
		{
			throw new UnsupportedOperationException("'" + joinType.getJoinExpression() + "' operation not supported by "+
				query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
	 	From the Join interface
	 	-------------------------------------------
	*/
	public String getThetaJoin(Query query, DbControl dc) 
		throws BaseException 
	{
		throw new UnsupportedOperationException("AnnotationJoin operation not supported by "+query.getQueryType()+" query.");
	}
		
	public String getThetaJoinCondition(Query query, DbControl dc) 
		throws BaseException 
	{
		throw new UnsupportedOperationException("AnnotationJoin operation not supported by "+query.getQueryType()+" query.");
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		String tmpAlias = "atmp" + annotationType.getId();
		String valuesProperty = StringUtil.decapitalize(annotationType.getValueType().getTableName());
		return joinType.getJoinExpression() + " " + 
			(alias == null ? "" : alias + ".") + "annotations " + tmpAlias +
			" " + joinType.getJoinExpression() + " " + tmpAlias + "." + valuesProperty +" " + joinedAlias;
	}
	/**
		This join is equal to another AnnotationJoin if they 
		join to the same annotation type with the same alias.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		AnnotationJoin o = (AnnotationJoin)other;
		return EqualsHelper.equals(annotationType, o.annotationType) && 
			EqualsHelper.equals(joinType, o.joinType) && 
			EqualsHelper.equals(alias, o.alias) && EqualsHelper.equals(joinedAlias, o.joinedAlias);
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(annotationType, joinType, alias, joinedAlias);
	}
	// -------------------------------------------
}

