/*
	$Id: Dynamic.java 5887 2011-11-23 13:49:03Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.MeanRawDataPropertyExpression;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.VirtualTable;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ReporterPropertyExpression;
import net.sf.basedb.core.RawDataPropertyExpression;
import net.sf.basedb.core.ExtraValueExpression;
import net.sf.basedb.core.ExtraValue;

/**
	A factory class to create expressions, joins, etc. that are only used
	by queries in the dynamic database. The elements will throw an
	<code>UnsupportedOperationException</code> if passed to
	a HQL query.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-11-23 14:49:03 +0100 (Wed, 23 Nov 2011) $
*/
public class Dynamic
{

	/**
		Select a column in the root table.
		@param column The column to select
		@return The new selection element
		@throws InvalidDataException If the column is null
		@see #select(VirtualTable, VirtualColumn)
	*/
	public static Select select(VirtualColumn column)
		throws InvalidDataException
	{
		return select(null, column);
	}
	
	/**
		Select a column in a joined or the root table.
		@param table The joined table where the column is located, 
			or null to use the root table
		@param column The column to select
		@return The new selection element
		@throws InvalidDataException If the column is null
	*/
	public static Select select(VirtualTable table, VirtualColumn column)
		throws InvalidDataException
	{
		return Selects.expression(column(table, column), column.getName());
	}

	/**
		Select a column in the reporters table. The expression uses cloned
		reporter information if available in the current experiment.
		
		@param property The name of the property
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinReporters(JoinType)
	*/
	public static Select selectReporter(String property)
		throws InvalidDataException
	{
		return selectReporter(property, null);
	}

	/**
		Select a column in the reporters table. The expression can be forced
		to use cloned or master reporter information.
		
		@param property The name of the property
		@param useCloned TRUE to use cloned reporter information (if available),
			FALSE to not use cloned information, null to automatically decide
		
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinReporters(JoinType, boolean)
		@see net.sf.basedb.core.DynamicQuery#setUseClonedReporters(boolean)
		@since 3.1
	*/
	public static Select selectReporter(String property, Boolean useCloned)
		throws InvalidDataException
	{
		return Selects.expression(reporter(property, useCloned), property);
	}
	
	/**
		Select a column in the raw data table. It is expected that
		raw data has been joined to the query.
		@param property The name of the property
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinRawData(JoinType)
	*/
	public static Select selectRawData(String property)
		throws InvalidDataException
	{
		return Selects.expression(rawData(property), property);
	}
	
	/**
		Select an extra value. It is expected that the
		extra value has been joined to the query.
		@param extraValue The extra value to select
		@param alias The alias of the extra value
		@throws InvalidDataException If the extraValue or alias is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinExtraValue(ExtraValue, JoinType)
	*/
	public static Select select(ExtraValue extraValue, String alias)
		throws InvalidDataException
	{
		return Selects.expression(extraValue(extraValue), alias);
	}

	/**
		Create an expression representing a column in the root table:
		expression = rootAlias.columnName.

		@param column The column 
		@return The new expression
		@throws InvalidDataException If the column is null
		@see #column(VirtualTable, VirtualColumn)
	*/
	public static Expression column(VirtualColumn column)
		throws InvalidDataException
	{
		return column(null, column);
	}

	/**
		Create an expression representing a column in a table:
		expression = tableAlias.columnName. It is expected that
		the specified table is the root table or is already
		joined by the query.

		@param table The table the column is located in or null if
			the column is located in the root table
		@param column The column 
		@return The new expression
		@throws InvalidDataException If the column is null
	*/
	public static Expression column(VirtualTable table, VirtualColumn column)
		throws InvalidDataException
	{
		if (column == null) throw new InvalidUseOfNullException("column");
		return new VirtualColumnExpression(table, column);
	}

	/**
		Create an expression representing a property of a reporter.
		expression = reporter.property. The expression uses cloned
		reporter information if available in the current experiment.

		@param property The name of the property
		@return The new expression
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinReporters(JoinType, boolean)
		@see net.sf.basedb.core.DynamicQuery#setUseClonedReporters(boolean)
	*/
	public static Expression reporter(String property)
		throws InvalidDataException
	{
		return reporter(property, null);
	}
	
	/**
		Create an expression representing a property of a reporter.
		expression = reporter.property. The expression can be forced
		to use cloned or master reporter information.
		
		@param property The name of the property
		@param useCloned TRUE to use cloned reporter information (if available),
			FALSE to not use cloned information, null to automatically decide
		@return The new expression
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinReporters(JoinType, boolean)
		@see net.sf.basedb.core.DynamicQuery#setUseClonedReporters(boolean)
		@since 3.1
	*/
	public static Expression reporter(String property, Boolean useCloned)
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		return new ReporterPropertyExpression(property, useCloned);
	}
	
	/**
		Create an expression representing a property of a raw data spot.
		expression = rawdata.property. It is expected that the
		raw data table is already joined by the query.

		@param property The name of the property
		@return The new expression
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinRawData(JoinType)
	*/
	public static Expression rawData(String property)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		return new RawDataPropertyExpression(property);
	}

	/**
		Create an expression representing the mean value of a property of a 
		raw data spot. expression = mean(rawdata.property). It is expected that 
		the raw data table is already joined by the query.
	
		@param property The name of the property
		@return The new expression
		@throws InvalidDataException If the property is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinRawData(JoinType)
	*/
	public static Expression meanRawData(String property)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		return new MeanRawDataPropertyExpression(property);
	}

	
	/**
		Create an expression representing the extra value of a spot
		or position. It is expected that the
		extra value is already joined by the query.

		@param extraValue The extra value
		@return The new expression
		@throws InvalidDataException If the extra value is null
		@see net.sf.basedb.core.DynamicSpotQuery#joinExtraValue(ExtraValue, JoinType)
	*/
	public static Expression extraValue(ExtraValue extraValue)
	{
		if (extraValue == null) throw new InvalidUseOfNullException("extraValue");
		return new ExtraValueExpression(extraValue);		
	}
	
	public static Restriction isPartOf(ReporterList reporterList)
	{
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		return Restrictions.neq(new ReporterListExpression(reporterList, "reporter_id", JoinType.INNER), null);
	}
	
	public static Restriction isNotPartOf(ReporterList reporterList)
	{
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		return Restrictions.eq(new ReporterListExpression(reporterList, "reporter_id", JoinType.LEFT), null);
	}
	
	public static Expression score(ReporterList reporterList)
	{
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		return new ReporterListExpression(reporterList, "score", JoinType.INNER);
	}

}
