/*
	$Id: Query.java 5881 2011-11-21 10:54:26Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.BaseException;

/**
	This defines the simplest form of a relational query. A relational
	query can be written in the form:
	
	<pre class="code">
SELECT [DISTINCT] <i>&lt;selection-list&gt;</i>
FROM <i>&lt;root-entity&gt;</i>
[JOIN <i>&lt;joined-entity&gt;</i>]
[WHERE <i>&lt;restrictions&gt;</i>]
[GROUP BY <i>&lt;groupby-list</i>]
[HAVING <i>&lt;restrictions&gt;</i>]
[ORDER BY <i>&lt;orderby-list&gt;</i>]
</pre>

	Only the <code>SELECT</code> and <code>FROM</code> query elements are needed, 
	everything else is optional. The methods in this interface are used to build 
	the query by adding {@link QueryElement}:s to the query.
	<p>
	All query elements can be either permanent or non-permanent. The difference
	is that permanent options are not cleared by the {@link #reset()} method.
	<p>
	Query implementations are not required to support all options. If that
	is the case a {@link UnsupportedOperationException} should be thrown.
	<p>
	This interface doesn't define any execution methods, because the type of 
	return object depends on the data that is returned. It is up to the 
	implementors to define appropriate execution methods. For
	example: {@link net.sf.basedb.core.ItemQuery#list(DbControl)}.
	
	@author Nicklas
	@version 2.0
	@see <a href="../../../../../../developer/api/query_api.html">Developer documentation: The Query API</a>
	@see net.sf.basedb.core.AbstractQuery
	@base.modified $Date: 2011-11-21 11:54:26 +0100 (Mon, 21 Nov 2011) $
*/
public interface Query
{

	/**
		Get the type of query. Ie. The query language.
		@return A value from the {@link QueryType} enumeration
	*/
	public QueryType getQueryType();

	public void addAutoJoiner(AutoJoiner<?, ?> joiner);
	
	/**
		Add a selection query element to the selection list.
		@param s The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #selectPermanent(Select)
	*/
	public void select(Select s);
	
	/**
		Permanently add a selection query element to the selection list. The
		query element is not cleared by the {@link #reset()} method.
		@param s The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #select(Select)
	*/
	public void selectPermanent(Select s);

	/**
		Add a join query element to the join list.
		@param j The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #joinPermanent(Join)
	*/
	public void join(Join j);
	
	/**
		Permanently add a join query element to the join list. The
		query element is not cleared by the {@link #reset()} method.
		@param j The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #join(Join)
	*/
	public void joinPermanent(Join j);

	/**
		Add a restriction query element to the restriction list.
		@param r The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #restrictPermanent(Restriction)
	*/
	public void restrict(Restriction r);
	
	/**
		Permanently add a restriction query element to the restriction list. The
		query element is not cleared by the {@link #reset()} method.
		@param r The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #restrict(Restriction)
	*/
	public void restrictPermanent(Restriction r);

	/**
		Add an expression query element to the group list.
		@param e The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #groupPermanent(Expression)
	*/
	public void group(Expression e);
	
	/**
		Permanently add an expression query element to the group list. The
		query element is not cleared by the {@link #reset()} method.
		@param e The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #group(Expression)
	*/
	public void groupPermanent(Expression e);
	
	/**
		Add a restriction query element to the having list.
		@param r The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #havingPermanent(Restriction)
	*/
	public void having(Restriction r);
	
	/**
		Permanently add a restriction query element to the having list. The
		query element is not cleared by the {@link #reset()} method.
		@param r The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #having(Restriction)
	*/
	public void havingPermanent(Restriction r);
	
	/**
		Add an ordering query element to the orderby list.
		@param o The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #orderPermanent(Order)
	*/
	public void order(Order o);
	
	/**
		Permanently add an ordering query element to the orderby list. The
		query element is not cleared by the {@link #reset()} method.
		@param o The query element to add
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #order(Order)
	*/
	public void orderPermanent(Order o);
	
	/**
		Set the value of a query parameter. If the value of this parameter
		has already been set by a call to the {@link #setPermanentParameter(String, Object, Type)}
		method, an <code>InvalidDataException</code> is thrown.
		@param name The name of the parameter
		@param value The value of the parameter
		@param valueType The type of the value, or null if not needed
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@throws InvalidDataException If the value of the parameter has already 
			been permanently set
		@see #setPermanentParameter(String, Object, Type)
	*/
	public void setParameter(String name, Object value, Type valueType);
	
	/**
		Permanently set the value of a query parameter. If the value of this parameter
		has already been set by a call to this method, an <code>InvalidDataException</code> 
		is thrown.
		@param name The name of the parameter
		@param value The value of the parameter
		@param valueType The type of the value, or null if not needed
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@throws InvalidDataException If the value of the parameter has already 
			been permanently set
		@see #setParameter(String, Object, Type)
	*/
	public void setPermanentParameter(String name, Object value, Type valueType);
	
	/**
		Check if a value for the specified parameter has been set or not.
		The method should check both permanent and non-permanent parameters.
		@return TRUE if a value (including null) has been set, FALSE otherwise
	*/
	public boolean hasParameterValue(String name);

	/**
		Get a set with the names for all parameters in this
		query.
		@return A set (empty if no parameters exists)
		@since 2.9
	*/
	public Set<String> getParameterNames();
	
	/**
		Get parameter information for the parameter with given name.
		@param name The name of the parameter
		@return A QueryParameter object or null if no parameter with
			that name exists
		@since 2.9
	*/
	public QueryParameter getQueryParameter(String name);
	
	/**
		Specify that the query should start returning rows from the
		specified row number. 0 = start returning from the first row. If the
		value is higher than the total number of rows, no rows are returned.
		This option is best combined with the {@link #setMaxResults(int)}
		option.
		
		@param firstResult The number of the first row starting at 0
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #setMaxResults(int)
	*/
	public void setFirstResult(int firstResult);
	
	/**
		The number of the row the query should start returning.
		@see #setFirstResult(int)
	*/
	public int getFirstResult();
	
	/**
		Specify that the query should at most return the specified number of
		rows. If the matching number of rows is less than this value only
		the matching rows are returned. A value of 0 disabled this option.
		This option is best combined with the {@link #setFirstResult(int)}
		option.
		
		@param maxResults The maximum number of rows to return
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #setFirstResult(int)
	*/
	public void setMaxResults(int maxResults);
	
	/**
		The maximum number of rows returned by the query. A value of 0
		indicates no limit.
		@see #setMaxResults(int)
	*/
	public int getMaxResults();

	/**
		Specify if the query should return a count for the total number
		of items that would have been returned if the {@link #setFirstResult(int)}
		and {@link #setMaxResults(int)} had been disabled, or if the results are
		loaded by an iterator where the number of rows not are known beforehand.
		@param flag TRUE if the query should return the total count, FALSE otherwise
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
	*/
	public void setReturnTotalCount(boolean flag);
	
	/**
		If this query returns the total count or not.
		@see #setReturnTotalCount(boolean)
	*/
	public boolean isReturningTotalCount();
	
	/**
		Specify if the query should only return distinct rows. Exctly what this means
		depends on what is selected by the query. For example, if we are selecting
		items no item would be returned more than once, if we are selecting
		numbers the same number wouldn't be returned more than once.
		@param flag TRUE if the query should return distinct results, FALSE otherwise
	*/
	public void setDistinct(boolean flag);
	
	/**
		If this query returns distinct results of not.
		@see #setDistinct(boolean)
	*/
	public boolean isDistinct();
	
	/**
		If this query is readonly and cannot be structurally modified. Ie. no more
		query elements can be added. It is still possible to set parameter values.
		A call to {@link #reset()} would unlock the query and allow it to be modified again.
		@return TRUE if the query is readonly, FALSE otherwise
	*/
	public boolean isReadonly();
	
	/**
		Reset the query, removing all non-permanent query elements and parameter
		values.
	*/
	public void reset();

	/**
		The alias of the <code>root-entity</code> in the <code>FROM</code> part.
		This value is used by the query elements to let a client application
		skip specifying the root alias all the time.
		@see Hql#property(String, String)
	*/
	public String getRootAlias();
	
	/**
		Count the number of items/rows that are returned by the query.
		The method ignores the {@link #getFirstResult()} and {@link #getMaxResults()}
		settings.
		@return The number of rows/items returned by the query
		@throws BaseException If there is an error
	*/
	public long count(DbControl dc)
		throws BaseException;

	/**
		Check if the query is executing the {@link #count(DbControl)} method or not
		@return TRUE if the query is currently executing the <code>count</code>
			method, FALSE otherwise
	*/
	public boolean isCounting();
	
	/**
		Get the current section of the query that is beeing built. This
		method can be used by {@link QueryElement}:s if they need
		different syntax in different sections.
		@return A QuerySection
	*/
	public QuerySection getQuerySection();
	
	/**
		Generate the query string that is going to be sent to the
		underlying query system (eg. HQL/Hibernate or SQL/JDBC).
		
		@param dc The DbControl that is going to be used for
			executing the query
		@return The query string
		@since 2.9
	*/
	public String toQl(DbControl dc);
	
	/**
		Set this flag to ensure that the transaction used to execute the query 
		is allowed to continue even after a failure at the SQL level of this query.
		Setting this flag to TRUE should allow the transaction to continue, setting 
		it to FALSE may or may not allow it to continue. The default setting is FALSE.
		
		@param failSafe TRUE to ensure that the transaction is allowed to continue,
			FALSE if it doesn't matter
		@since 3.1
	*/
	public void setFailSafe(boolean failSafe);

	/**
		Check if a failure of this query should still allow the current transaction
		to continue.
		@return TRUE if the transaction should be allowed to continue,
			FALSE if it doesn't  matter
		@since 3.1
	*/
	public boolean isFailSafe();
	
}
