/**
	$Id: AnnotationUtil.java 6120 2012-09-13 07:53:46Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.filter.Filter;

/**
	Utility class for working with annotations.

	@author Nicklas
	@version 2.9
	@base.modified $Date: 2012-09-13 09:53:46 +0200 (Thu, 13 Sep 2012) $
*/
public class AnnotationUtil
{
	/**
		Create a new cache object that can be used with
		{@link #getAllAnnotatableParentItems(DbControl, Annotatable, Filter, Cache)}
		@since 3.2
	*/
	public static Cache createCache()
	{
		return new Cache();
	}
	
	/**
		Get all annotatable parent items for a given item. A {@link Filter}
		may be used to filter the returned collection. If a filter
		is specified only parent items for which the {@link Filter#evaluate(Object)}
		method returns true are included in the returned collection. Note
		that all parent items are loaded before the filter is evaluated
		and that the evaluation order is undefined.
		
		@param dc The DbControl to use for database access
		@param item The item to find the parents for
		@param filter An optional filter that can be used to filter 
			which parents to include
		@return A set containing the parent items
		@see #getAllAnnotatableParentItems(DbControl, Annotatable, Filter, Cache)
	*/
	public static Set<Annotatable> getAllAnnotatableParentItems(DbControl dc, 
		Annotatable item, Filter<Annotatable> filter)
	{
		return getAllAnnotatableParentItems(dc, item, filter, null);
	}

	/**
		Get all annotatable parent items for a given item. A {@link Filter}
		may be used to filter the returned collection. If a filter
		is specified only parent items for which the {@link Filter#evaluate(Object)}
		method returns true are included in the returned collection. Note
		that all parent items are loaded before the filter is evaluated
		and that the evaluation order is undefined. A cache may be used
		to improve performance if this method is called multiple times
		with different items that have overlapping parent trees.
		
		@param dc The DbControl to use for database access
		@param item The item to find the parents for
		@param filter An optional filter that can be used to filter 
			which parents to include
		@param cache An optional cache object
		@return A set containing the parent items
		@see #createCache()
		@since 3.2
	*/
	public static Set<Annotatable> getAllAnnotatableParentItems(DbControl dc, 
		Annotatable item, Filter<Annotatable> filter, Cache cache)
	{
		Set<AnnotatableWrapper> processed = new LinkedHashSet<AnnotatableWrapper>();
		Queue<AnnotatableWrapper> unprocessed = new LinkedList<AnnotatableWrapper>();
		unprocessed.add(new AnnotatableWrapper(item, cache));
	
		// As long as we have unprocessed entries...
		while (!unprocessed.isEmpty())
		{
			AnnotatableWrapper toProcess = unprocessed.remove();
			// ...mark it as processed...
			if (processed.add(toProcess))
			{
				// ...and queue the parents items for processing...
				unprocessed.addAll(toProcess.getAnnotatableParentWrappers());
			}
		}
		
		// All parents are now loaded...
		Set<Annotatable> allParents = new LinkedHashSet<Annotatable>(processed.size());
		for (AnnotatableWrapper wrapper : processed)
		{
			// ...copy all parents that passes the filter
			Annotatable parent = wrapper.getAnnotatable();
			if (filter == null || filter.evaluate(parent)) allParents.add(parent);
		}
		// Do not include the root item in the result
		allParents.remove(item);
		return allParents;
	}

	
	/**
		This wrapper class is needed because we need to keep
		track of extracts linked with raw bioassays and derived bioassays
		until we have loaded the Extracts of a Physical Bioassay.
		<p>
		When we reach a {@link RawBioAssay} item or {@link DerivedBioAssay} item, we 
		remember the extract on	that as we move up towards PhysicalBioAssay.
		When the PhysicalBioAssay is reached we use the extracts to call
		{@link PhysicalBioAssay#getAnnotatableParents(int, Collection)}.
	*/
	static class AnnotatableWrapper
	{
		/**
	 		Get the extract to use. This method returns:
	 		
	 		<ul>
	 		<li>If the item is a raw bioassay, {@link RawBioAssay#getParentExtract()}
	 		<li>If the item is a derived bioassay, {@link DerivedBioAssay#getExtract()}
	 		<li>null in all other cases.
	 		</ul>
		*/
		private static Extract getExtract(Annotatable item)
		{
			Extract result = null;
			if (item instanceof RawBioAssay)
			{
				result = ((RawBioAssay)item).getParentExtract();
			}
			else if (item instanceof DerivedBioAssay)
			{
				result = ((DerivedBioAssay)item).getExtract();
			}
			return result;
		}
		
		private final Annotatable item;
		private final Set<Extract> extractsInChain;
		private final AnnotatableWrapper chain;
		private final Cache cache;
		private final Extract extract;

		AnnotatableWrapper(Annotatable item, Cache cache)
		{
			this.item = item;
			this.chain = null;
			this.cache = cache;
			this.extract = getExtract(item);
			this.extractsInChain = loadExtracts();
		}
		
		AnnotatableWrapper(Annotatable item, AnnotatableWrapper chain)
		{
			this.item = item;
			this.chain = chain;
			this.cache = chain.cache;
			this.extract = getExtract(item);
			this.extractsInChain = loadExtracts();
		}
		
		
		/**
			The annotatable that we are wrapping.
		*/
		Annotatable getAnnotatable()
		{
			return item;
		}
		
		/**
			Load extracts in parent chain. We have to consider this for rawbioassay, 
			derived bioassay and physical bioassay since the annotatable parents 
			to a physical bioassay depends on the extracts we have seen while 
			traversing the path from the raw bioassay and up. After reaching the physical
			bioassay we can forget about the extracts.
		*/
		private Set<Extract> loadExtracts()
		{
			Set<Extract> extractsInChain;
			switch (item.getType())
			{
				case PHYSICALBIOASSAY:
				case DERIVEDBIOASSAY:
				case RAWBIOASSAY:
				{
					extractsInChain = new HashSet<Extract>();
					if (chain != null) extractsInChain.addAll(chain.extractsInChain);
					if (extract != null) extractsInChain.add(extract);
					break;
				}
				default:
				{
					extractsInChain = Collections.emptySet();
				}
			}
			return extractsInChain;
		}
		
		/**
			Get all annotatable parents wrapped inside AnnotatableWrapper:s
		*/
		Set<AnnotatableWrapper> getAnnotatableParentWrappers() 
			throws BaseException
		{
			Set<Annotatable> parents = null;
			Set<AnnotatableWrapper> wrappedParents = Collections.emptySet();
			
			if (cache != null && cache.contains(this))
			{
				wrappedParents = cache.get(this);
			}
			else
			{
				if (item instanceof PhysicalBioAssay)
				{
					parents = ((PhysicalBioAssay)item).getAnnotatableParents(0, extractsInChain);
				}
				else
				{
					parents = item.getAnnotatableParents();
				}
				if (parents != null)
				{
					wrappedParents = new HashSet<AnnotatableWrapper>(parents.size());
					for (Annotatable parent : parents)
					{
						wrappedParents.add(new AnnotatableWrapper(parent, this));
					}
				}
				if (cache != null) cache.put(this, wrappedParents);
			}
			return wrappedParents;
		}
		
		/**
			A wrapper is equal to another if they references the same item
			and chain of extracts.
		*/
		@Override
		public boolean equals(Object o)
		{
			if (o == this) return true;
			if (o == null || o.getClass() != this.getClass()) return false;
			AnnotatableWrapper other = (AnnotatableWrapper)o;
			return this.item.equals(other.item) && this.extractsInChain.equals(other.extractsInChain);
		}

		@Override
		public int hashCode()
		{
			return item.hashCode() + extractsInChain.hashCode();
		}
		
		@Override
		public String toString()
		{
			return item.toString() + ":" + extractsInChain;
		}
	}
	
	public static class Cache
	{
		private final Map<AnnotatableWrapper, Set<AnnotatableWrapper>> cmap;
		
		Cache()
		{
			this.cmap = new HashMap<AnnotatableWrapper, Set<AnnotatableWrapper>>();
		}
		
		boolean contains(AnnotatableWrapper wrapper)
		{
			return cmap.containsKey(wrapper);
		}
		
		Set<AnnotatableWrapper> get(AnnotatableWrapper wrapper)
		{
			return cmap.get(wrapper);
		}
		
		void put(AnnotatableWrapper wrapper, Set<AnnotatableWrapper> parents)
		{
			cmap.put(wrapper, parents);
		}
		
	}
}
