/**
	$Id:Registry.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemNotFoundException;

import net.sf.basedb.core.plugin.About;
import net.sf.basedb.util.extensions.events.EventFilter;
import net.sf.basedb.util.extensions.events.EventHandler;
import net.sf.basedb.util.extensions.events.EventType;
import net.sf.basedb.util.extensions.xml.XmlLoader;

/**
	A registry for keeping track of extensions points and 
	installed extensions. Extension points and extensions can, for example,
	be created programmatically using the {@link ExtensionPointBean}
	and {@link ExtensionBean} implementations, or loaded from
	an XML extension definition file with the {@link XmlLoader}
	class. Applications may provide other means of creating and
	registering extensions and extensions points.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class Registry
{

	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.util.extensions.Registry");

	/**
		Comparator that sort extension points according to their names.
		If two names are equal, the sort is also done on the id.
		@since 3.0
	*/
	public static final Comparator<ExtensionPoint> EXTENSIONPOINT_COMPARATOR = 
		new Comparator<ExtensionPoint>()
		{
			@Override
			public int compare(ExtensionPoint o1, ExtensionPoint o2)
			{
				int result = o1.getName().compareTo(o2.getName());
				if (result == 0) result = o1.getId().compareTo(o2.getId());
				return result;
			}
		};

	/**
		Comparator that sort extensions according to their names.
		If two names are equal, the sort is also done on the id.
		@since 3.0
	*/
	public static final Comparator<Extension> EXTENSION_COMPARATOR = 
		new Comparator<Extension>()
		{
			@Override
			public int compare(Extension o1, Extension o2)
			{
				String name1 = o1.getAbout() == null ? null : o1.getAbout().getName();
				String name2 = o2.getAbout() == null ? null : o2.getAbout().getName();
				int result = name1 != null && name2 != null ? name1.compareTo(name2) : 0;
				if (result == 0) result = o1.getId().compareTo(o2.getId());
				return result;
			}
		};

		
	private final Map<String, RegisteredExtensionPoint<?>> extensionPoints;
	private final Map<String, RegisteredExtension<?>> extensions;
	private Map<AttributeKey, Object> attributes;
	private Map<EventHandler, EventFilter> eventHandlers;
	private ErrorHandlerFactory<Action> defaultErrorHandlerFactory;

	/**
		Create a new empty registry.
	*/
	public Registry()
	{
		extensionPoints = new HashMap<String, RegisteredExtensionPoint<?>>();
		extensions = new HashMap<String, RegisteredExtension<?>>();
		defaultErrorHandlerFactory = new LoggingErrorHandlerFactory();
	}
	
	/**
		Set the default error handler factory.
		@param defaultErrorHandlerFactory Null is not allowed
		@since 2.17
	*/
	public void setDefaultErrorHandlerFactory(ErrorHandlerFactory<Action> defaultErrorHandlerFactory)
	{
		if (defaultErrorHandlerFactory == null) 
		{
			throw new InvalidUseOfNullException("defaultErrorHandlerFactory");
		}
		this.defaultErrorHandlerFactory = defaultErrorHandlerFactory;
	}
	
	/**
		Get the default error handler factory.
		@since 2.17
	*/
	public ErrorHandlerFactory<Action> getDefaultErrorHandlerFactory()
	{
		return defaultErrorHandlerFactory;
	}
	
	/**
		Get the value of a extra attribute for a registered extension
		or extension point.
		@param id The ID of a registered extension or extension point
		@param name The name of the attribute
		@return The attribute value or null if no value has been set
	*/
	public Object getAttribute(String id, String name)
	{
		if (id == null || name == null || attributes == null) return null;
		AttributeKey key = new AttributeKey(id, name);
		return attributes.get(key);
	}
	
	/**
		Set an extra attribute on a registered extension or extension point.
		@param id The ID of a registered extension or extension point
		@param name The name of the attribute
		@param value The attribute value. A null value removes the attribute.
	*/
	public void setAttribute(String id, String name, Object value)
	{
		if (id == null || name == null) return;
		AttributeKey key = new AttributeKey(id, name);
		if (attributes == null) attributes = new HashMap<AttributeKey, Object>();
		if (value == null)
		{
			attributes.remove(key);
		}
		else
		{
			attributes.put(key, value);
		}
	}
	
	/**
		Register an event handler for responding to events happening to 
		extension points and extensions in the registry. The registry itself
		sends the events listed below. Other events may be sent externally
		by calling {@link #handleEvent(EventType, ExtensionPoint, Extension)}.
		
		<ul>
		<li>{@link EventType#AFTER_REGISTRATION}
		<li>{@link EventType#BEFORE_UPDATE}
		<li>{@link EventType#AFTER_UPDATE}
		<li>{@link EventType#BEFORE_UNREGISTRATION}
		
		@param eventHandler The handler that should be called
		@param filter A filter that determines which events that should be
			sent to the handler, or null to send all events
		@since 2.8
	*/
	public void registerEventHandler(EventHandler eventHandler, EventFilter filter)
	{
		if (eventHandlers == null) eventHandlers = new HashMap<EventHandler, EventFilter>();
		eventHandlers.put(eventHandler, filter);
	}
	
	/**
		Unregisters an event handler.
		@param eventHandler The event handler
		@since 2.8
	*/
	public void unregisterEventHandler(EventHandler eventHandler)
	{
		if (eventHandlers == null) return;
		eventHandlers.remove(eventHandler);
	}
	
	/**
		Send an event to all event handlers.
		
		@param eventType The event type
		@param extensionPoint The extension point the event is about
		@param extension The extension the event is about, or null
		@since 2.8
	*/
	public void handleEvent(EventType eventType, ExtensionPoint<?> extensionPoint, Extension<?> extension)
	{
		if (eventHandlers == null) return;
		for (Map.Entry<EventHandler, EventFilter> entry : eventHandlers.entrySet())
		{
			EventHandler handler = entry.getKey();
			EventFilter filter = entry.getValue();
			if (filter == null || filter.shouldSend(eventType, extensionPoint, extension))
			{
				try
				{
					handler.handleEvent(eventType, extensionPoint, extension);
				}
				catch (Throwable t)
				{
					log.error("Error when seningd event '" + eventType + "' to extension point '" + 
						extensionPoint + "' and extension '" + extension + "'", t);
				}
			}
		}
	}
	
	/**
		Checks if an extension point is registered or not.
		@param id The ID of the extension point
		@return TRUE if an extension point with the given ID is registered,
			FALSE otherwise
	*/
	public boolean extensionPointIsRegistered(String id)
	{
		return extensionPoints.containsKey(id);
	}
	
	/**
		@deprecated In 3.2, use {@link #registerExtensionPoint(ExtensionPoint, ClassLoader)} instead
	*/
	@Deprecated
	public synchronized <A extends Action> void 
		registerExtensionPoint(ExtensionPoint<A> extensionPoint)
	{
		registerExtensionPoint(extensionPoint, null);
	}
	
	/**
		Register an extension point. If the extension point already exists,
		the description and renderer factory is updated. The action class
		can't be changed for a registered extension point. The registry keeps
		it's own internal representation of the extension point. Changes to
		the object passed as a parameter to this method will not be reflected
		in the registry.
		
		@param extensionPoint The extension point to register
		@param classLoader The class loader that is responsible for loading classes
			for the extension point, or null if the default class loader should be used
		@throws InvalidUseOfNullException If the extension point is null, or if the ID 
			or actionClass is null
		@throws ClassCastException If the action class doesn't implement the
			{@link Action} interface
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public synchronized <A extends Action> void 
		registerExtensionPoint(ExtensionPoint<A> extensionPoint, ClassLoader classLoader)
	{
		if (extensionPoint == null) throw new InvalidUseOfNullException("extensionPoint");

		String id = extensionPoint.getId();
		if (id == null) 
		{
			throw new InvalidUseOfNullException("extensionPoint.id of " + extensionPoint);
		}
		Class<A> actionClass = extensionPoint.getActionClass();
		if (actionClass == null) 
		{
			throw new InvalidUseOfNullException("Action class of extensionPoint[id=" + id + "]");
		}
		if (!Action.class.isAssignableFrom(actionClass))
		{
			throw new ClassCastException("Action class '" + actionClass.getName() + 
				"' of extensionPoint[id=" + id + "] doesn't implement the 'Action' class");
		}
		
		RegisteredExtensionPoint<A> rep = 
			(RegisteredExtensionPoint<A>)extensionPoints.get(extensionPoint.getId());
		if (rep == null)
		{
			// Create a new registered entry
			rep = new RegisteredExtensionPoint<A>(extensionPoint, classLoader);
			extensionPoints.put(extensionPoint.getId(), rep);
			handleEvent(EventType.AFTER_REGISTRATION, rep, null);
		}
		else
		{
			// Update the existing registration
			handleEvent(EventType.BEFORE_UPDATE, rep, null);
			rep.update(extensionPoint, classLoader);
			handleEvent(EventType.AFTER_UPDATE, rep, null);
		}
	}
	
	/**
		Unregister an extension point. If the extension point doesn't exists,
		nothing is done. This method will also unregister all extensions
		that extends the given extension point.
		
		@param id The ID of the extension point to unregister
		@throws InvalidUseOfNullException If the ID is null
	*/
	public synchronized void unregisterExtensionPoint(String id)
	{
		if (id == null) throw new InvalidUseOfNullException("id");
		RegisteredExtensionPoint<?> rep = extensionPoints.get(id);
		if (rep != null)
		{
			handleEvent(EventType.BEFORE_UNREGISTRATION, rep, null);
			for (RegisteredExtension<?> rext : rep.getExtensions())
			{
				handleEvent(EventType.BEFORE_UNREGISTRATION, rep, rext);
				extensions.remove(rext.getId());
			}
			extensionPoints.remove(id);
		}
	}
	
	/**
		Get the extension point with a given ID.
		@param id The ID of the extension point
		@return An {@link ExtensionPoint} object or null if that 
			extension point doesn't exists
	*/
	public ExtensionPoint<?> getExtensionPoint(String id)
	{
		return extensionPoints.get(id);
	}
	
	/**
		Get a list with all registered extension points.
		The returned list is a copy of the registered extension points
		at the time of this call and will not reflect changes
		made to the registry after this call has returned.
		
		@return A list
		@since 3.0 (Returned an Iterator in BASE 2.x)
	*/
	public synchronized List<ExtensionPoint<?>> getExtensionPoints()
	{
		List<ExtensionPoint<?>> copy = 
			new ArrayList<ExtensionPoint<?>>(extensionPoints.values());
		return copy;
	}
	
	/**
		Checks if an extension is registered or not.
		@param id The ID of the extension
		@return TRUE if an extension with the given ID is registered,
			FALSE otherwise
	*/
	public boolean extensionIsRegistered(String id)
	{
		return extensions.containsKey(id);
	}
	
	/**
		Get information about the last error that happened
		when rendering an extension point.
		@param id The id of the extension point
		@return The error, or null if no error information is available
		@since 2.9
	*/
	public Throwable getLastExtensionPointError(String id)
	{
		RegisteredExtensionPoint<?> rep = extensionPoints.get(id);
		return rep == null ? null : rep.getLastError();
	}
	
	/**
		@deprecated In 3.2, use {@link #registerExtension(Extension, ClassLoader)} instead
	*/
	@Deprecated
	public synchronized <A extends Action> void registerExtension(Extension<A> extension)
	{
		registerExtension(extension, null);
	}

	/**
		Register an extension. If the extension is already registered,
		the description, position, action factory and renderer factory
		is updated. The registry keeps it's own internal representation of 
		the extension. Changes made to the object passed as a parameter to this 
		method will not be reflected in the registry.
		
		@param extension The extension to register
		@param classLoader The class loader that is responsible for loading classes
			for the extension, or null if the default class loader should be used
		@throws InvalidUseOfNullException If the extension is null, or if
			the ID or action factory is null
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public synchronized <A extends Action> void registerExtension(Extension<A> extension, ClassLoader classLoader)
	{
		if (extension == null) 
		{
			throw new InvalidUseOfNullException("extension");
		}
		String id = extension.getId();
		if (id == null) 
		{
			throw new InvalidUseOfNullException("extension.id of " + extension);
		}
		ActionFactory<?> actionFactory = extension.getActionFactory();
		if (actionFactory == null) 
		{
			throw new InvalidUseOfNullException("Action factory for extension '" + id + "' is null.");
		}

		String extensionPointId = extension.getExtends();
		RegisteredExtensionPoint<A> rep = 
			(RegisteredExtensionPoint<A>)extensionPoints.get(extensionPointId);
		if (rep == null)
		{
			throw new ItemNotFoundException("ExtensionPoint[id=" + extensionPointId + "]");
		}
		
		RegisteredExtension<A> rext = (RegisteredExtension<A>)extensions.get(extension.getId());
		if (rext != null && !rext.getId().equals(extensionPointId))
		{
			// This extension was previously registered under another extension point
			// Unregister it and re-register it as a new extension
			unregisterExtension(id);
			rext = null;
		}
		if (rext == null)
		{
			// Create a new registry entry
			rext = new RegisteredExtension<A>(extension, rep, classLoader);
			rep.addExtension(rext);
			extensions.put(extension.getId(), rext);
			handleEvent(EventType.AFTER_REGISTRATION, rep, rext);
		}
		else
		{
			// Update the existing entry
			handleEvent(EventType.BEFORE_UPDATE, rep, rext);
			rext.update(extension, classLoader);
			handleEvent(EventType.AFTER_UPDATE, rep, rext);
		}
	}
	
	/**
		Unregister an extension. If the extension point doesn't exists,
		nothing is done.
		
		@param id The ID of the extension to unregister
		@throws InvalidUseOfNullException If the ID is null
	*/
	public synchronized void unregisterExtension(String id)
	{
		if (id == null) throw new InvalidUseOfNullException("id");
		RegisteredExtension<?> rext = extensions.get(id);
		if (rext != null)
		{
			RegisteredExtensionPoint<?> rep = extensionPoints.get(rext.getExtends());
			handleEvent(EventType.BEFORE_UNREGISTRATION, rep, rext);
			extensions.remove(id);
			if (rep != null) rep.removeExtension(id);
		}
	}
	
	/**
		Get the extension with a given ID.
		@param id The ID of the extension point
		@return An {@link Extension} object or null if that 
			extension doesn't exists
	*/
	public Extension<?> getExtension(String id)
	{
		return extensions.get(id);
	}
	
	/**
		Get information about the last error that happened
		when rendering an extension.
		@param id The id of the extension
		@return The error, or null if no error information is available
		@since 2.9
	*/
	public Throwable getLastExtensionError(String id)
	{
		RegisteredExtension<?> ext = extensions.get(id);
		return ext == null ? null : ext.getLastError();
	}

	
	/**
		Get a list with all registered extensions.
		The returned list is a copy of the registered extensions
		at the time of this call and will not reflect changes
		made to the registry after this call has returned.

		@return A list
		@since 3.0 (Returned an Iterator in BASE 2.x)
	*/
	public synchronized List<Extension<?>> getExtensions()
	{
		List<Extension<?>> copy = 
			new ArrayList<Extension<?>>(extensions.values());
		return copy;
	}
	
	/**
		Get a list with all registered extensions
		for a given extension point. If the extension point is not found
		an empty list is returned. 
		The returned list is a copy of the registered extension points
		at the time of this call and will not reflect changes
		made to the registry after this call has returned.
		
		@param id The ID of the extension point
		@return A list
		@since 3.0 (Returned an Iterator in BASE 2.x)
	*/
	public synchronized List<Extension<?>> getExtensions(String id)
	{
		RegisteredExtensionPoint<?> rep = extensionPoints.get(id);
		if (rep == null)
		{
			return Collections.emptyList();
		}
		else
		{
			// Create a copy, or the iterator will fail if new
			// extension points are registered by another thread
			List<Extension<?>> copy = 
				new ArrayList<Extension<?>>(rep.getExtensions());
			return copy;
		}
	}
	
	/**
		Use extensions from one or more extension points. This method will 
		return an invoker object that allows the use and rendering of extensions
		for one or more extension points. The context object should have been
		initialised with information about the currently logged in user and,
		depending on the application, some information about what it about to 
		happen. For example, that we are going to display a list of samples.
		
		<p>
		Each extension that has been registered with the extension points will
		get a chance to check the context object and to further initialise it
		with extension-dependent information. This is a task for the 
		{@link ActionFactory#prepareContext(InvokationContext)} method. If this method 
		returns false the extension will not be included in the invoker. Extensions
		may for example use this to check if the logged in user has enough permissions
		to be able to use the extension.
		
		<p>
		If the extension points and/or extensions also provide renderer factories,
		the {@link RendererFactory#prepareContext(InvokationContext)} is called.
		For renderer factories that are attached to extension points the method
		is only called if at least one enabled extension exists. If the method is
		called, it is only called once with a null argument for the extension.
		Renderer factories that are attached to extensions are called once for each
		enabled extension, but only if the extension point allows renderer overriding.
		See {@link ExtensionPoint#allowRendererOverride()}.
		
		<p>
		The returned invoker will not reflect changes made to the registry by the 
		same or other threads after this call has returned. Each use/invokation of
		an extension point should call this method to get a new invoker object.
		
		@param clientContext Information about the current context
	 	@param filter A filter object that can be used to filter out
			disabled extensions and sort the extensions in a particular order.
			If this value is null, no extensions are filtered out and they 
			are sorted according to their {@link Extension#getIndex()} value
	 	@param extensionPointIds An array of extension point ID:s to get
			extension for. In most cases the extension points should
			require the same action class but this is not required
			if the extension point renderer can handle more than one action
			type. Extension points that are not found are ignored.
	*/
	@SuppressWarnings("unchecked")
	public ExtensionsInvoker<?> useExtensions(ClientContext clientContext, ExtensionsFilter filter, 
		String... extensionPointIds)
	{
		if (filter == null) filter = DEFAULT_FILTER;

		List<ExtensionContext<Action>> contexts = new LinkedList<ExtensionContext<Action>>();
		for (String id : extensionPointIds)
		{
			if (id == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}
			
			// Find a registered extension point for each ID
			RegisteredExtensionPoint<Action> rep = 
				(RegisteredExtensionPoint<Action>)extensionPoints.get(id);
			
			// Is there an enabled extension point?
			if (rep == null || !filter.isEnabled(rep))
			{
				if (rep == null)
				{
					// Trying to use an unedefined extension point should log a warning
					log.warn("Undefined extension point: " + id, new RuntimeException());
				}
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}

			// YES! ...
			ErrorHandlerFactory ehf = rep.getErrorHandlerFactory();
			ExtensionPointContext<Action> mainContext = new ExtensionPointContext<Action>(this, 
					clientContext, rep, ehf == null ? defaultErrorHandlerFactory : ehf);
			boolean allowRenderOverride = rep.allowRendererOverride();
			int enabledExtensions = 0;
			
			// ... find the extensions for it
			for (RegisteredExtension<Action> ext : rep.getExtensions())
			{
				// Check with the filter if the extension is enabled
				if (!filter.isEnabled(ext))
				{
					/* #### CONTINUE-STATEMENT #### */
					continue;
				}

				// Create invokation context for the extension
				ExtensionContext<Action> context = 
					new ExtensionContext<Action>(mainContext, ext);
				
				// Check with the action factory as well
				if (!context.prepareActionFactory()) 
				{
					/* #### CONTINUE-STATEMENT #### */
					continue;
				}

				// Yippi, this extension is enabled!!!
				if (allowRenderOverride) context.prepareRendererFactory();
				contexts.add(context);
				enabledExtensions++;
			}
			
			// Prepare the renderer factory for the extension point
			// if there is at least one enabled extension
			if (enabledExtensions > 0) mainContext.prepareRendererFactory();
		}
		
		// Sort the extensions
		filter.sort(contexts);
		return new ExtensionsInvoker<Action>(contexts);
	}
	
	/**
		Default filter implementation for 
		{@link #useExtensions(ClientContext, ExtensionsFilter, String...)}
	*/
	private static final ExtensionsFilter DEFAULT_FILTER = new DefaultFilter();
	
	/**
		Internal representation of an extension point.
	*/
	static class RegisteredExtensionPoint<A extends Action>
		implements ExtensionPoint<A>
	{

		private final String id;
		private final Class<A> actionClass;
		private final Map<String, RegisteredExtension<A>> extensions;

		private String name;
		private String description;
		private RendererFactory<? super A> rendererFactory;
		private boolean allowRendererOverride;
		private ErrorHandlerFactory<? super A> errorHandlerFactory;
		private ClassLoader classLoader;
		
		private Throwable lastError;
		private A lastErrorAction;
		
		/**
			Create a new registered extension point by copying the
			information from the parameter.
		*/
		RegisteredExtensionPoint(ExtensionPoint<A> ep, ClassLoader classLoader)
		{
			this.id = ep.getId();
			this.actionClass = ep.getActionClass();
			this.extensions = new HashMap<String, RegisteredExtension<A>>();
			update(ep, classLoader);
		}
		
		/*
			From the ExtensionPoint interface
			---------------------------------
		*/
		@Override
		public String getId() 
		{
			return id;
		}		

		@Override
		public String getName() 
		{
			return name;
		}

		@Override
		public String getDescription() 
		{
			return description;
		}
			
		@Override
		public Class<A> getActionClass()
		{
			return actionClass;
		}
		
		@Override
		public RendererFactory<? super A> getRendererFactory()
		{
			return rendererFactory;
		}
		
		@Override
		public boolean allowRendererOverride() 
		{
			return allowRendererOverride;
		}
		
		@Override
		public ErrorHandlerFactory<? super A> getErrorHandlerFactory()
		{
			return errorHandlerFactory;
		}		
		// ---------------------------------
		
		@Override
		public String toString()
		{
			return "ExtensionPoint[id="+ id +"; name=" + name + "]";
		}
		
		/**
			Update the registered information.
		*/
		void update(ExtensionPoint<A> ep, ClassLoader classLoader)
		{
			this.classLoader = classLoader;
			this.name = ep.getName();
			this.description = ep.getDescription();
			this.rendererFactory = ep.getRendererFactory();
			this.allowRendererOverride = ep.allowRendererOverride();
			this.errorHandlerFactory = ep.getErrorHandlerFactory();
		}
		
		/**
	 		Add an extension to this extension point.
		*/
		void addExtension(RegisteredExtension<A> extension)
		{
			this.extensions.put(extension.getId(), extension);
		}
		
		/**
			Remove the extension with a given ID.
		*/
		void removeExtension(String id)
		{
			this.extensions.remove(id);
		}
		
		/**
			Get all extensions registered with this extension point.
		*/
		Collection<RegisteredExtension<A>> getExtensions()
		{
			return this.extensions.values();
		}
		
		/**
			@since 3.2
		*/
		ClassLoader getClassLoader()
		{
			return classLoader;
		}
		
		synchronized void setError(A action, Throwable t)
		{
			this.lastErrorAction = action;
			this.lastError = t;
		}
		
		synchronized void clearError()
		{
			this.lastError = null;
			this.lastErrorAction = null;
		}
		
		Throwable getLastError()
		{
			return lastError;
		}
	}
		
	/**
		Internal representation of an extension.
	*/
	static class RegisteredExtension<A extends Action>
		implements Extension<A>
	{

		private final RegisteredExtensionPoint<? super A> rep;
		private final String id;
		private final String extensionPoint;
		private RegisteredAbout about;
		private float index;
		private ActionFactory<? extends A> actionFactory;
		private RendererFactory<? super A> rendererFactory;
		private ClassLoader classLoader;

		private Throwable lastError;
		private A lastErrorAction;
		
		/**
			Create a new registered extension by copying the
			information from the parameter.
		*/
		RegisteredExtension(Extension<A> extension, RegisteredExtensionPoint<? super A> rep, ClassLoader classLoader)
		{
			this.rep = rep;
			this.id = extension.getId();
			this.extensionPoint = extension.getExtends();
			update(extension, classLoader);
		}

		/*
			From the Extension interface
			---------------------------------
		*/
		@Override
		public String getId() 
		{
			return id;
		}

		@Override
		public String getExtends() 
		{
			return extensionPoint;
		}
		
		@Override
		public float getIndex() 
		{
			return index;
		}

		@Override
		public ActionFactory<? extends A> getActionFactory() 
		{
			return actionFactory;
		}
		
		@Override
		public RendererFactory<? super A> getRendererFactory() 
		{
			return rendererFactory;
		}

		@Override
		public About getAbout() 
		{
			return about;
		}
		// ---------------------------------

		@Override
		public String toString()
		{
			return "Extension[id="+ id +"]";
		}
		
		/**
			Update the registered information.
		*/
		void update(Extension<A> extension, ClassLoader classLoader)
		{
			this.classLoader = classLoader;
			this.index = extension.getIndex();
			this.actionFactory = extension.getActionFactory();
			this.rendererFactory = extension.getRendererFactory();
			if (extension.getAbout() == null)
			{
				this.about = null;
			}
			else
			{
				if (this.about == null)
				{
					this.about = new RegisteredAbout(extension.getAbout());
				}
				else
				{
					this.about.update(extension.getAbout());
				}
			}
		}
		
		/**
			Get the extension point this extension is registered with.
		*/
		RegisteredExtensionPoint<? super A> getExtensionPoint()
		{
			return rep;
		}

		/**
			@since 3.2
		*/
		ClassLoader getClassLoader()
		{
			return classLoader;
		}
		
		synchronized void setError(A action, Throwable t)
		{
			this.lastErrorAction = action;
			this.lastError = t;
		}
		
		synchronized void clearError()
		{
			this.lastError = null;
			this.lastErrorAction = null;
		}
		
		Throwable getLastError()
		{
			return lastError;
		}
	}		
		
	static class RegisteredAbout
		implements About
	{

		private String name;
		private String description;
		private String version;
		private String minBaseVersion;
		private String maxBaseVersion;
		private String copyright;
		private String contact;
		private String email;
		private String url;
		
		/**
			Create a new registered about by copying the
			information from the parameter.
		*/
		RegisteredAbout(About about)
		{
			update(about);
		}
		
		/*
			From the About interface
			---------------------------------
		*/
		@Override
		public String getContact()
		{
			return contact;
		}
		
		@Override
		public String getCopyright() 
		{
			return copyright;
		}
		
		@Override
		public String getDescription() 
		{
			return description;
		}
		
		@Override
		public String getEmail() 
		{
			return email;
		}

		@Override
		public String getName() 
		{
			return name;
		}

		@Override
		public String getUrl() 
		{
			return url;
		}

		@Override
		public String getVersion() 
		{
			return version;
		}
		
		@Override
		public String getMinBaseVersion() 
		{
			return minBaseVersion;
		}

		@Override
		public String getMaxBaseVersion() 
		{
			return maxBaseVersion;
		}
		// ---------------------------------

		/**
			Update the registered information.
		*/
		void update(About about)
		{
			this.name = about.getName();
			this.description = about.getDescription();
			this.version = about.getVersion();
			this.minBaseVersion = about.getMinBaseVersion();
			this.maxBaseVersion = about.getMaxBaseVersion();
			this.contact = about.getContact();
			this.copyright = about.getCopyright();
			this.email = about.getEmail();
			this.url = about.getUrl();
		}
	}
	
	/**
		Identifies attribute. The key is: 
		<code>
		id of extension + name of attribute
		</code>
	
		@author nicklas
		@version 2.7
		@base.modified $Date: 2012-03-28 12:38:04 +0200 (Wed, 28 Mar 2012) $
	 */
	static class AttributeKey
	{
		private final String id;
		private final String name;
		AttributeKey(String id, String name)
		{
			this.id = id;
			this.name = name;
		}
		
		@Override
		public int hashCode()
		{
			return 13 * id.hashCode() + 5 * name.hashCode();
		}
		
		@Override
		public boolean equals(Object other)
		{
			if (this == other) return true;
			if (other == null || other.getClass() != this.getClass()) return false;
			AttributeKey ak = (AttributeKey)other;
			return this.id.equals(ak.id) && this.name.equals(ak.name);
		}
		
	}
	
}
