/**
	$Id: Validator.java 6098 2012-09-03 13:25:00Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import net.sf.basedb.util.overview.extensions.ValidationRuleAction;
import net.sf.basedb.util.overview.validator.NodeValidator;

/**
	Placeholder for a validation rule. An object instance contains a 
	name and description of the validation and a default severity level.
	<p>
	The actual validation takes place in the overview generating code.
	See implementations of {@link NodeValidator} interface.
	
	@author Nicklas
	@version 2.2
	@base.modified $Date: 2012-09-03 15:25:00 +0200 (Mon, 03 Sep 2012) $
*/
public class Validator
	implements ValidationRuleAction
{

	/**
		Validator rule that checks for missing protocols. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_PROTOCOL = new Validator(
			"missing.protocol", "Missing protocol", 
			"Checks if a protocol has been specified for items that can have a protocol",
			"Missing protocol", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to protocols. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_PROTOCOL = new Validator(
			"denied.protocol", "Access denied to protocol", 
			"Checks if the user is denied access to a protocol",
			"Access denied to protocol", Severity.WARNING
		);

	/**
		Validator rule that checks for missing parent derived bioassay. 
		Default severity level is WARNING.
		@since 3.0
	*/
	public static final Validator MISSING_DERIVEDBIOASSAY = new Validator(
			"missing.derivedbioassay", "Missing derived bioassay", 
			"Checks if a parent bioassay has been specified for a raw bioassay",
			"Missing parent bioassay", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to derived bioassay.
		Default severity level is WARNING.
		@since 3.0
	*/
	public static final Validator DENIED_DERIVEDBIOASSAY = new Validator(
			"denied.derivedbioassay", "Access denied to derived bioassay", 
			"Checks if the user is denied access to a derived bioassay",
			"Access denied to derived bioassay", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to platform. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_PLATFORM = new Validator(
			"denied.platform", "Access denied to platform", 
			"Checks if the user is denied access to a platform",
			"Access denied to platform", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to (data) files. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_FILE = new Validator(
			"denied.file", "Access denied to file", 
			"Checks if the user is denied access to a file",
			"Access denied to file", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to data file types. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_DATAFILETYPE = new Validator(
			"denied.datafiletype", "Access denied to data file type", 
			"Checks if the user is denied access to a data file type",
			"Access denied to data file type", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing software. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_SOFTWARE = new Validator(
			"missing.software", "Missing software", 
			"Checks if a software has been specified for a raw bioassay or derived bioassay",
			"Missing software", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to software. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_SOFTWARE = new Validator(
			"denied.software", "Access denied to software", 
			"Checks if the user is denied access to a software",
			"Access denied to software", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing hardware. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_HARDWARE = new Validator(
			"missing.hardware", "Missing hardware", 
			"Checks if a hardware (scanner, print robot, etc.) has been specified for " +
			"item that can have a hardware.",
			"Missing hardware", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to hardware. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_HARDWARE = new Validator(
			"denied.hardware", "Access denied to hardware", 
			"Checks if the user is denied access to a hardware",
			"Access denied to hardware", Severity.WARNING
		);

	/**
		Validator rule that checks for missing physical bioassay. Default severity level
		is WARNING.
		@since 3.0
	*/
	public static final Validator MISSING_PHYSICALBIOASSAY = new Validator(
			"missing.physicalbioassay", "Missing physical bioassay", 
			"Checks if a physical bioassay has been specified for a derived bioassay or array slide",
			"Missing bioassay", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to physical bioassay. Default severity level
		is WARNING.
		@since 3.0
	*/
	public static final Validator DENIED_PHYSICALBIOASSAY = new Validator(
			"denied.physicalbioassay", "Access denied to physical bioassay", 
			"Checks if the user is denied access to a physical bioassay",
			"Access denied to physical bioassay", Severity.WARNING
		);

	/**
		Validator rule that checks for missing array slide. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_ARRAYSLIDE = new Validator(
			"missing.slide", "Missing array slide", 
			"Checks if an array slide has been specified for a physical bioassay",
			"Missing array slide", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to array slides. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_ARRAYSLIDE = new Validator(
			"denied.slide", "Access denied to array slide", 
			"Checks if the user is denied access to an array slide",
			"Access denied to array slide", Severity.WARNING
		);

	/**
		Validator rule that checks for missing extract. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_EXTRACT = new Validator(
			"missing.extract", "Missing extract", 
			"Checks if at least one extract has been specified for a " +
			"physical bioassay",
			"Missing extract", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to extracts. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_EXTRACT = new Validator(
			"denied.extract", "Access denied to extract", 
			"Checks if the user is denied access to a extract",
			"Access denied to extract", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing sample. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_SAMPLE = new Validator(
			"missing.sample", "Missing sample", 
			"Checks if a sample has been specified for an extract or pooled sample",
			"Missing sample", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to samples. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_SAMPLE = new Validator(
			"denied.sample", "Access denied to sample", 
			"Checks if the user is denied access to a sample",
			"Access denied to sample", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing biosource. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_BIOSOURCE = new Validator(
			"missing.biosource", "Missing biosource", 
			"Checks if a biosource has been specified for a non-pooled sample",
			"Missing biosource", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to biosource. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_BIOSOURCE = new Validator(
			"denied.biosource", "Access denied to biosource", 
			"Checks if the user is denied access to a biosource",
			"Access denied to biosource", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing tag. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_TAG = new Validator(
			"missing.tag", "Missing tag", 
			"Checks if a tag has been specified for an extract",
			"Missing tag", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to tags. Default severity level
		is WARNING.
		@since 3.0
	*/
	public static final Validator DENIED_TAG = new Validator(
			"denied.tag", "Access denied to tag", 
			"Checks if the user is denied access to a tag.",
			"Access denied to tag", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing array design. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_ARRAYDESIGN = new Validator(
			"missing.arraydesign", "Missing array design", 
			"Checks if an array design has been specified for a raw bioassay",
			"Missing array design", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to an array design. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_ARRAYDESIGN = new Validator(
			"denied.arraydesign", "Access denied to array design", 
			"Checks if the user is denied access to an array design",
			"Access denied to array design", Severity.WARNING
		);

	/**
		Validator rule that checks for missing array batch. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_ARRAYBATCH = new Validator(
			"missing.arraybatch", "Missing array batch", 
			"Checks if an array batch has been specified for an array slide",
			"Missing array batch", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to an array batch. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_ARRAYBATCH = new Validator(
			"denied.arraybatch", "Access denied to array batch", 
			"Checks if the user is denied access to an array batch",
			"Access denied to array batch", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing experimental factor value. 
		Default severity level is ERROR.
	*/
	public static final Validator MISSING_FACTORVALUE = new Validator(
			"missing.factorvalue", "Missing factor value", 
			"Checks if a value has been specified for all experimental factors for a raw bioassay",
			"Missing factor value", Severity.ERROR
		);

	/**
		Validator rule that checks for missing protocol parameter values. 
		Default severity level is ERROR.
	*/
	public static final Validator MISSING_PARAMETER = new Validator(
			"missing.parameter", "Missing parameter value", 
			"Checks if a parameter value has been specified for all protocol parameters",
			"Missing parameter value", Severity.ERROR
		);

	/**
		Validator rule that checks for missing "Required for MIAME" annotation. 
		Default severity level is ERROR.
	*/
	public static final Validator MISSING_MIAME_ANNOTATION = new Validator(
			"missing.miameannotation", "Missing MIAME annotation value", 
			"Checks if an annotation value has been specified for a 'Required for MIAME' annotation type",
			"Missing MIAME annotation value", Severity.ERROR
		);

	/**
		Validator rule that checks for denied access to an annotation type. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_ANNOTATIONTYPE = new Validator(
			"denied.annotationtype", "Access denied to annotation type", 
			"Checks if the user is denied access to an annotation type",
			"Access denied to annotation type", Severity.WARNING
		);
	
	/**
		Validator rule that checks for a missing reference in an any-to-any link. 
		Eg. due to the non-relational nature an any-to-any entry may point to an
		item that has been deleted. Default severity level is WARNING.
		@since 2.17
	*/
	public static final Validator MISSING_ANYTOANY = new Validator(
			"missing.anytoany", "Missing any-to-any target", 
			"Checks if the item referenced to by an any-to-any link actually exists or not.",
			"Missing any-to-any target", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access the referenced item in an
		any-to-any link. Default severity level is WARNING.
		@since 2.17
	*/
	public static final Validator DENIED_ANYTOANY = new Validator(
			"denied.anytoany", "Access denied to any-to-any", 
			"Checks if the user is denied access to the referenced item in any-to-any link.",
			"Access denied to any-to-any", Severity.WARNING
		);

	/**
		Validator rule that checks for missing bioplate location.
		Default severity level is IGNORE.
		@since 3.2
	*/
	public static final Validator MISSING_BIOPLATE = new Validator(
			"missing.bioplate", "Missing bioplate", 
			"Checks if bioplate location has been specified for a biomaterial",
			"Missing bioplate location", Severity.IGNORE
		);

	
	/**
		Validator rule that checks for denied access to a bioplate. 
		Default severity level is WARNING.
		@since 3.2
	*/
	public static final Validator DENIED_BIOPLATE = new Validator(
			"denioed.bioplate", "Access denied to bioplate",
			"Checks if the user is defined access to a bioplate.",
			"Access denied to bioplate", Severity.WARNING);
	
	
	/**
		Validator rule that checks if an item uses a different platform / variant than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_PLATFORM = new Validator(
			"nondefault.platform", "Use of non-default platform",
			"Checks if an item uses a different platform than the default for the current project",
			"Use of non-default platform", Severity.WARNING
		);

	/**
		Validator rule that checks if an item uses a different protocol than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_PROTOCOL = new Validator(
			"nondefault.protocol", "Use of non-default protocol",
			"Checks if an item uses a different protocol than the default for the current project",
			"Use of non-default protocol", Severity.WARNING
		);

	/**
		Validator rule that checks if an item uses a different hardware than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_HARDWARE = new Validator(
			"nondefault.hardware", "Use of non-default hardware",
			"Checks if an item uses a different hardware (scanner, print robot, etc.) " +
			"than the default for the current project",
			"Use of non-default hardware", Severity.WARNING
		);
	
	/**
		Validator rule that checks if an item uses a different software than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_SOFTWARE = new Validator(
			"nondefault.software", "Use of non-default software",
			"Checks if an item uses a different software (feature extraction) " +
			"than the default for the current project",
			"Use of non-default software", Severity.WARNING
		);

	/**
		Validator rule that checks if an item uses a different array design than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_ARRAYDESIGN = new Validator(
			"nondefault.arraydesign", "Use of non-default array design",
			"Checks if an item uses a different array design " +
			"than the default for the current project",
			"Use of non-default array design", Severity.WARNING
		);
	
	/**
		Validator rule that checks if an item uses a different raw data type than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_RAWDATATYPE = new Validator(
			"nondefault.rawdatatype", "Use of non-default raw data type",
			"Checks if an experiment uses the same raw data type as specified by the " +
			"default for the current project",
			"Use of non-default raw data type", Severity.WARNING
		);
	
	/**
		Validator rule that checks if all raw bioassays in an experiment
		uses the same array design. Default severity level is WARNING.
	*/
	public static final Validator MULTIPLE_ARRAYDESIGNS = new Validator(
			"misc.multiplearraydesigns", "Multiple array designs",
			"Checks if all raw bioassays in an experiment uses the same " +
			"array design.",
			"Multiple array designs", Severity.WARNING
		);

	/**
		Validator rule that checks if the array design for a raw bioassay
		matches the array design of the associated physical bioassay.
		Default severity level is WARNING.
	*/
	public static final Validator NONMATHCING_ARRAYDESIGN = new Validator(
			"misc.nonmatchingarraydesigns", "Array design mismatch",
			"Checks if the array design for a raw bioassay is the same " +
			"array design as of the associated physical bioassay.",
			"Array design mismatch", Severity.WARNING
		);

	/**
		Validator rule that checks if the number of raw data spots 
		are the same as the number of features (given a certain threshold).
		Default severity level is WARNING.
	*/
	public static final Validator NONMATCHING_SPOTCOUNT = new Validator(
			"misc.nonmatchingspotcount", "Raw spots <> features",
			"Checks if the number of raw data spots matches the number of " +
			"array design features. A threshold can be given as either an " +
			"absolute number or as a percentage.",
			"Raw spots <> features", Severity.WARNING
		);

	/**
		Validator rule that checks if the tag attached to an extract
		has a correct subtype. Default severity level is WARNING.
		@since 3.0
	*/
	public static final Validator INCORRECT_TAGTYPE = new Validator(
			"subtype.tag", "Incorrect tag subtype",
			"Checks if the tag attached to an extract has a correct subtype.",
			"Incorrect tag subtype", Severity.WARNING
		);

	
	/**
		Validator rule that checks if the protocol attached to an item
		has a correct protocol type. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_PROTOCOLTYPE = new Validator(
			"subtype.protocol", "Incorrect protocol subtype",
			"Checks if the protocol attached to an item has a correct protocol type.",
			"Incorrect protocol subtype", Severity.WARNING
		);

	/**
		Validator rule that checks if the software attached to an item
		has a correct software type. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_SOFTWARETYPE = new Validator(
			"subtype.software", "Incorrect software subtype",
			"Checks if the software attached to an item has a correct software type.",
			"Incorrect software subtype", Severity.WARNING
		);

	/**
		Validator rule that checks if the hardware attached to an item
		has a correct hardware type. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_HARDWARETYPE = new Validator(
			"subtype.hardware", "Incorrect hardware subtype",
			"Checks if the hardware attached to an item has a correct hardware type.",
			"Incorrect hardware subtype", Severity.WARNING
		);
	
	/**
		Validator that checks if the parent biosource of a sample
		has the expected subtype. Default severity level is WARNING.
		@since 3.0
	 */
	public static final Validator INCORRECT_BIOSOURCE_TYPE = new Validator(
			"subtype.biosource", "Incorrect biosource subtype",
			"Checks if the parent biosource of a sample has the expected subtype.",
			"Incorrect biosource subtype", Severity.WARNING
	);

	/**
		Validator that checks if the parent sample of a sample/extract
		has the expected subtype. Default severity level is WARNING.
		@since 3.0
	 */
	public static final Validator INCORRECT_SAMPLE_TYPE = new Validator(
			"subtype.sample", "Incorrect sample subtype",
			"Checks if the parent sample of a sample/extract has the expected subtype.",
			"Incorrect sample subtype", Severity.WARNING
	);

	/**
		Validator that checks if the parent extract of a extract/bioassay
		has the expected subtype. Default severity level is WARNING.
		@since 3.0
	 */
	public static final Validator INCORRECT_EXTRACT_TYPE = new Validator(
			"subtype.extract", "Incorrect extract subtype",
			"Checks if the parent extract of an extract/physical bioassay has the expected subtype.",
			"Incorrect extract subtype", Severity.WARNING
	);

	/**
		Validator that checks if the biomaterial on a bioplate is
		of the correct item type. Default severity level is ERROR.
		@since 3.2
	 */
	public static final Validator INCORRECT_BIOMATERIAL_TYPE = new Validator(
			"subtype.biomaterial", "Incorrect biomaterial type",
			"Checks if the biomaterial on a bioplate is of the correct item type.",
			"Incorrect biomaterial type", Severity.ERROR
	);
	
	/**
		Validator that checks if a physical bioassay 
		has the expected subtype. Default severity level is WARNING.
		@since 3.0
	 */
	public static final Validator INCORRECT_PHYSICALBIOASSAY_TYPE = new Validator(
			"subtype.physicalbioassay", "Incorrect physical bioassay subtype",
			"Checks if the physical bioassay has the expected subtype.",
			"Incorrect physical bioassay subtype", Severity.WARNING
	);

	/**
		Validator that checks if a derived bioassay 
		has the expected subtype. Default severity level is WARNING.
		@since 3.0
	 */
	public static final Validator INCORRECT_DERIVEDBIOASSAY_TYPE = new Validator(
			"subtype.derivedbioassay", "Incorrect derived bioassay subtype",
			"Checks if the derived bioassay has the expected subtype.",
			"Incorrect derived bioassay subtype", Severity.WARNING
	);

	
	/**
		Validator rule that checks if the position of an extract
		on a physical bioassay is between 1 and the size of the
		bioassay. Default severity level is WARNING.
		@since 3.0
	*/
	public static final Validator INCORRECT_BIOASSAY_POSITION = new Validator(
			"misc.incorrectbioassayposition", "Incorrect bioassay position",
			"Checks if the position of an extract on a physical bioassay " +
			"is between 1 and the size of the bioassay.",
			"Incorrect bioassay position", Severity.WARNING
		);

	/**
		Validator rule that checks if a physical bioassay has more than one
		tagged extract with the same tag and position. Default severity level is ERROR.
		@since 3.0
	*/
	public static final Validator NONUNIQUE_TAG = new Validator(
			"misc.nonuniquetag", "Multiple extracts with same tag+position",
			"Checks if a physical bioassay has more than one tagged extract with the " +
			"same tag and position",
			"Multiple tagged extracts with same tag+position", Severity.ERROR
		);

	/**
		Validator rule that checks if the parent to a tagged extract has 
		a different tag. Default severity level is WARNING.
		@since 3.0
	*/
	public static final Validator DIFFERENT_TAG = new Validator(
			"misc.differenttag", "Parent extract has different tag",
			"Checks if the parent to a tagged extract has a " +
			"different tag",
			"Parent extract has different tag", Severity.WARNING
		);
	
	/**
		Validator rule that checks if the parent to a non-tagged extract 
		has a tag. Default severity level is WARNING.
		@since 3.0
	*/
	public static final Validator TAGGED_PARENT = new Validator(
			"misc.taggedparent", "Un-tagged extract has tagged parent",
			"Checks if the parent to an un-tagged extract has " +
			"a tag.",
			"Un-tagged extract has tagged parent", Severity.WARNING
		);
	
	/**
		Validator rule that checks if a child a raw- or derived-
		bioassay has a different extract than the parent bioassay.
		@since 3.0
	 */
	public static final Validator PARENT_EXTRACT_MISMATCH = new Validator(
			"misc.extractmismatch", "Extract mismatch",
			"Checks if a child derived- or raw- bioassay has an " +
			"extract that doesn't match the parent bioassay extract. It doesn't have to " +
			"be an exact match, but the extracts should belong to the same subtree.",
			"Extract mismatch", Severity.WARNING
		);
	
	/**
		Validator rule that checks if two items of the same type has the same
		name. Default severity level is ERROR.
	*/
	public static final Validator NONUNIQUE_NAME_FOR_TYPE = new Validator(
			"misc.nonuniquenamefortype", "Non-unique name",
			"Checks if more than one item of the same type has the same name.",
			"Non-unique name", Severity.ERROR
		);

	/**
		Validator rule that checks if two items of different type has the same
		name. Default severity level is IGNORE.
	*/
	public static final Validator NONUNIQUE_NAME_GLOBAL = new Validator(
			"misc.nonuniquenameglobal", "Global non-unique name",
			"Checks if more than one item (of any type) has the same name.",
			"Global non-unique name", Severity.IGNORE
		);

	/**
		Validator rule that checks for circular references among pooled items. 
		Default severity level is ERROR.
	*/
	public static final Validator CIRCULAR_REFERENCE = new Validator(
			"misc.circular", "Circular reference to pooled item",
			"Checks if there are circular references among pooled items.",
			"Circular reference to pooled item", Severity.ERROR
		);

	
	/**
		Validator rule that checks if an annotation is a protocol parameter.
		Default severity level is WARNING.
	*/
	public static final Validator ANNOTATION_IS_PARAMETER = new Validator(
			"annotation.isparameter", "Annotation is protocol parameter",
			"Checks if an item has annotations that are protocol parameters.",
			"Annotation is protocol parameter", Severity.WARNING
		);

	/**
		Validator rule that checks if an annotation is not valid for the
		current item type. Default severity level is WARNING.
	*/
	public static final Validator ANNOTATION_INVALID_ITEM = new Validator(
			"annotation.invaliditem", "Annotation invalid for item",
			"Checks if an annotation is invalid for the type of item.",
			"Annotation is invalid for item", Severity.WARNING
		);

	/**
		Validator rule that checks if an annotation value is invalid
		according to annotation type specification. Default severity level is ERROR.
	*/
	public static final Validator ANNOTATION_INVALID_VALUE = new Validator(
			"annotation.invalidvalue", "Annotation has invalid value",
			"Checks if an annotation has an invalid value.",
			"Annotation has invalid value", Severity.ERROR
		);

	/**
		Validator rule that checks if an annotation have more values than
		what the 'multiplicity' setting for the annotation type allow. 
		Default severity level is WANRING.
		@since 3.2
	*/
	public static final Validator ANNOTATION_TOO_MANY_VALUES = new Validator(
			"annotation.toomanyvalues", "Annotation has too many values",
			"Checks if an annotation has more values than the 'multiplicty' setting allow.",
			"Annotation has too many values", Severity.WARNING
		);

	
	/**
		Validator rule that checks if an inherited annotation value is inherited
		from a parent item. Default severity level is ERROR.
	*/
	public static final Validator ANNOTATION_INHERIT_FROM_NONPARENT = new Validator(
			"annotation.inhertitfromnonparent", "Inheriting annotation from non-parent",
			"Checks if an item inherits an annotation that is is not from a parent item.",
			"Inherting annotation from non-parent", Severity.ERROR
		);
	
	/**
		Validator rule that checks if the logged in user has access to
		the parent item of an inherited annotation value. Default severity 
		level is WARNING.
		@since 2.10
	*/
	public static final Validator ANNOTATION_INHERIT_FROM_DENIEDPARENT = new Validator(
			"annotation.inhertitfromdeniedparent", "Inheriting annotation from denied parent",
			"Checks if an item inherits an annotation from a parent item to which the " +
			"logged in user doesn't have access.",
			"Inherting annotation from denied parent", Severity.WARNING
		);

	
	/**
		Validator rule that checks if two or more inherited annotations have the
		same annotation type. Default severity level is ERROR.
	*/
	public static final Validator ANNOTATION_INHERIT_MULTIPLE = new Validator(
			"annotation.inhertitmultiple", "Inheriting multiple annotation of same type",
			"Checks if an item inherits more than one annotation of the same annotation type.",
			"Inheriting multiple annotation of same type", Severity.ERROR
		);
	
	/**
		Validator rule that checks if a data file can be used on the
		current item type. Default severity level is WARNING.
	*/
	public static final Validator DATAFILE_INVALID_ITEM = new Validator(
			"datafile.invaliditem", "Data file can't be used on item",
			"Checks if a data file can be used for a given item type.",
			"Data file can't be used on item", Severity.WARNING
		);

	/**
		Validator rule that checks if a data file has been validated
		or not. Default severity level is WARNING, but this rule is not 
		checked if the data file type doesn't have a validator.
	*/
	public static final Validator DATAFILE_NOTVALIDATED = new Validator(
			"datafile.notvalidated", "Data file has not been validated",
			"Checks if a data file has been validated or not.",
			"Data file has not been validated", Severity.WARNING
		);
	
	/**
		Validator rule that checks if a data file is valid.
		Default severity level is ERROR, but this rule is not checked if
		the data file hasn't been validated.
	*/
	public static final Validator DATAFILE_INVALID = new Validator(
			"datafile.invalid", "Data file is invalid",
			"Checks if a data file is valid or not.",
			"Data file is invalid", Severity.ERROR
		);

	/**
		Validator rule that checks if all files required by the platform
		are present or not. Default severity level is ERROR.
	*/
	public static final Validator MISSING_DATAFILE = new Validator(
			"datafile.missing", "Missing a required data file",
			"Checks if all files required by a platform are present.",
			"Missing a required data file", Severity.ERROR
		);

	/**
		Validator rule that checks if a file is registered with
		a platform/item subtype or not. Default severity level is WARNING.
	*/
	public static final Validator DATAFILE_NOTPLATFORM = new Validator(
			"datafile.notplatform", "File is not used by platform/subtype",
			"Checks that all data files are used by the current platform/subtype.",
			"File is not used by platform/subtype", Severity.WARNING
		);
	
	/**
		Validator rule that checks if there are more than one file
		of a file type that doesn't allow multiple files. Default 
		severity level is WARNING.
		@since 3.0
	*/
	public static final Validator DATAFILE_TOOMANYFILES = new Validator(
			"datafile.toomanyfiles", "Too many files",
			"Checks if an item has too many files of a type that doesn't " +
			"allow more than one file.",
			"Too many files", Severity.WARNING
		);
	
	private String id;
	private String title;
	private String description;
	private String failureSummary;
	private Severity defaultSeverity;
	
	/**
		Create a new validator.
		@param id A unique ID among all validators
		@param title The title for the validation rule 
		@param description A lengthier description of the validation that takes place
		@param failureSummary A short summary that is displayed when the validation
			fails
		@param defaultSeverity The default severity level of the validation
	*/
	public Validator(String id, String title, String description, String failureSummary, Severity defaultSeverity)
	{
		this.id = id;
		this.title = title;
		this.description = description;
		this.failureSummary = failureSummary;
		this.defaultSeverity = defaultSeverity;
	}
	
	/**
		Get the unique ID of this validator.
	*/
	@Override
	public String getId()
	{
		return id;
	}
	
	/**
		Get the title of this validator.
	*/
	@Override
	public String getTitle()
	{
		return title;
	}
	
	/**
		Get a lengthier description of the validation that takes place
	*/
	@Override
	public String getDescription()
	{
		return description;
	}
	
	/**
		Get a short summary that is displayed when the validation fails
	*/
	@Override
	public String getFailureSummary()
	{
		return failureSummary;
	}
	
	/**
		Get the default severity level.
	*/
	@Override
	public Severity getDefaultSeverity()
	{
		return defaultSeverity;
	}
}
