/**
	$Id$
	
	Copyright (C) 2012 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.BioPlateNameGenerator;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NodeFactory;


/**
	Node loader implementation for bioplates. Bioplates are
	loaded as root nodes, or as property nodes of a biomaterial.
	The forward-loading direction goes to the biomaterial that is 
	located on the plate. The reverse direction is not loading 
	anything.
	
	@author nicklas
	@since 3.2
*/
public class BioPlateLoader 
	extends BasicItemNodeLoader<BioPlate> 
{
	public BioPlateLoader()
	{
		super(Item.BIOPLATE, ALLOW_ROOT_NODE, 
				new BioPlateNameGenerator("bioplate", "Bioplate"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/

	/**
		Create a bioplate property node from a biomaterial that is located on a plate.
		@return A bioplate node, or null if the biomaterial is not located on a plate
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentNode.getItem() instanceof MeasuredBioMaterial)
		{
			returnNode = createPropertyNode((MeasuredBioMaterial)parentNode.getItem(dc), dc, context, parentNode);
		}
		return returnNode;
	}	
	// -------------------------------------
	
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	
	/**
		Loads the non-empty biowells on the bioplate.
		@see BioWellLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node plateNode)
	{
		getNodeLoader(context, Item.BIOWELL).createForwardNode(dc, context, plateNode);
	}
	
	/**
		Load the freezer node.
		@see HardwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node plateNode) 
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, plateNode);
		getNodeLoader(context, Item.HARDWARE).createPropertyNode(dc, context, plateNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, plateNode);
	}
	// ---------------------------------------


	/**
		Create a propert node when the parent item is a biomaterial.
	*/
	private Node createPropertyNode(MeasuredBioMaterial bioMaterial, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<BioPlate> nf = getNodeFactory(dc, context);
		BioWell well = null;
		BioPlate plate = null;
		boolean denied = false;
		try
		{
			well = bioMaterial.getBioWell();
			if (well != null) plate = well.getPlate();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node plateNode = createItemNode(nf, plate, well, denied, parentNode, ChildNodeDirection.PROPERTY);
		return plateNode;
	}

}
