/**
	$Id: NodeFactory.java 5818 2011-10-21 11:15:41Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.node;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.cache.DirectionalCacheKey;
import net.sf.basedb.util.overview.cache.NodeCache;
import net.sf.basedb.util.overview.validator.NodeValidator;

/**
	A node factory is responsible for creating item-type nodes.
	The primary function of a node factory is to interface with
	a {@link NodeValidator} to actually decide if a node should 
	be created or not for a given item (which may be missing or
	the current user may be denied access).
	<p>
	Another function is to interface with the current 
	{@link OverviewContext} and it's optional {@link NodeCache}
	so that we don't have to use resources for re-loading a deep
	sub-tree starting with an item that has already been loaded
	in another branch of the tree.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-21 13:15:41 +0200 (Fri, 21 Oct 2011) $
*/
public class NodeFactory<I extends BasicItem>
{
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.util.overview.node.NodeFactory");
	private static final boolean debug = log.isDebugEnabled();

	private DbControl dc;
	private OverviewContext context;
	private NodeValidator<? super I> validator;
	private NodeNameGenerator<? super I> nameGenerator;
	
	/**
		Create a node factory. If a validator is specified it will be
		used to determine if a node should be created or not. Without
		a validator nodes are not created for missing items, only for
		existing item (including if access is denied)
		
		@param dc A DbControl to use if database access is needed
		@param context The current overview context
		@param validator An optional validator
		@param nameGenerator A name generator to use for node name and title
			generation
	*/
	public NodeFactory(DbControl dc, OverviewContext context, 
			NodeValidator<? super I> validator, NodeNameGenerator<? super I> nameGenerator)
	{
		this.dc = dc;
		this.context = context;
		this.validator = validator;
		this.nameGenerator = nameGenerator;
	}
	
	/**
		Get the node validator that is used by this factory.
		@return A node validator or null if no validator is used
	*/
	public NodeValidator<? super I> getNodeValidator()
	{
		return validator;
	}
	
	/**
		Create a node representing a missing item. A node is only created
		if a validator is present and it's {@link 
		NodeValidator#preMissingItem(DbControl, OverviewContext, Node)}
		method returns TRUE.

		@param parentNode The parent node
		@return The new node, or null if no node was created
	*/
	public Node createMissingNode(Node parentNode)
	{
		Node node = null;
		if (validator != null && validator.preMissingItem(dc, context, parentNode))
		{
			String nodeName = nameGenerator.getMissingNodeName(parentNode);
			String nodeTitle = nameGenerator.getMissingNodeTitle(parentNode);
			node = createNode(null, nodeName, nodeTitle, null, parentNode, ChildNodeDirection.NONE);
			validator.postMissingItem(dc, context, node, parentNode);
		}
		return node;
	}
	
	/**
		Create a node representing an item that the logged in user is denied to
		access. A node is created if no validator is present or if the
		validator's {@link 
		NodeValidator#preDeniedItem(DbControl, OverviewContext, Node)}
		method returns TRUE.
	
		@param parentNode The parent node
		@return The new node, or null if no node was created
	*/
	public Node createDeniedNode(Node parentNode)
	{
		Node node = null;
		if (validator == null || validator.preDeniedItem(dc, context, parentNode))
		{
			String nodeName = nameGenerator.getDeniedNodeName(parentNode);
			String nodeTitle = nameGenerator.getDeniedNodeTitle(parentNode);
			node = createNode(null, nodeName, nodeTitle, null, parentNode, ChildNodeDirection.NONE);
			if (validator != null) validator.postDeniedItem(dc, context, node, parentNode);
		}
		return node;
	}
	
	/**
		Create a node representing an existing item. A node is created if no 
		validator is present or if the validator's {@link 
		NodeValidator#preValidate(DbControl, OverviewContext, Object, Node)}
		method returns TRUE.
		
		@param item The item to attach to the node
		@param cacheKey An optional cache key
		@param parentNode The parent node
		@param direction The direction to use when loading the node's children
		@return The new node, or null if no node was created
	*/
	public Node createNode(I item, Object cacheKey, Node parentNode, ChildNodeDirection direction)
	{
		Node node = null;
		if (validator == null || validator.preValidate(dc, context, item, parentNode))
		{
			String nodeName = nameGenerator.getNodeName(item, parentNode);
			String nodeTitle = nameGenerator.getNodeTitle(item, parentNode);
			node = createNode(item, nodeName, nodeTitle, cacheKey, parentNode, direction);
			if (validator != null) validator.postValidate(dc, context, node, parentNode);
		}
		return node;
	}

	private Node createNode(I item, String name, String title, Object cacheKey, Node parentNode, ChildNodeDirection direction)
	{
		Node node = null;
		if (cacheKey != null)
		{
			node = cloneCachedNode(cacheKey, parentNode, direction);
		}
		if (node == null)
		{
			node = new Node(name, title, parentNode, item, direction);
			if (cacheKey != null) cacheNewNode(cacheKey, node, direction);
		}
		return node;
	}
	
	/**
		Clones a cached node if the current context supports
		caching and another node with the given cache key 
		already exists. This method wraps the given cacheKey 
		with a {@link DirectionalCacheKey} since the loading
		direction will affect the children of a node.
		
		@param cacheKey The cache key to use for node lookup
		@param parentNode The node that should be the parent of the
			cloned node
		@param direction The direction we are loading child nodes in
		@return The new (cloned) node, or null if no cached node was found
		@see OverviewContext#getNodeCache()
	*/
	protected Node cloneCachedNode(Object cacheKey, Node parentNode, ChildNodeDirection direction)
	{
		NodeCache<Object> cache = context.getNodeCache();
		if (cache == null) return null;
		Node node = cache.getNode(new DirectionalCacheKey(cacheKey, direction));
		if (node != null) 
		{
			if (debug) log.debug("Cloning node: " + node + "; parent=" + parentNode);
			node = node.deepCopy(parentNode);
		}
		return node;
	}
	
	/**
		Store a node in the node cache if the current overview context
		supports caching.
		
		@param cacheKey The cache key
		@param node The node to store
		@param direction The direction we are loading child nodes in
		@see OverviewContext#getNodeCache()
	*/
	protected void cacheNewNode(Object cacheKey, Node node, ChildNodeDirection direction)
	{
		NodeCache<Object> cache = context.getNodeCache();
		if (cache == null) return;
		cache.putNode(new DirectionalCacheKey(cacheKey, direction), node);
	}

	public void postValidateFolder(Node folderNode, Node parentNode)
	{
		if (validator != null)
		{
			validator.postValidateFolder(dc, context, folderNode, parentNode);
		}
	}
}
