/**
	$Id: ExtractValidator.java 6100 2012-09-04 08:43:21Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.BioMaterialEventSource;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.SpecialQuery;
import net.sf.basedb.core.Tag;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.biomaterial.ParentsTransformer;
import net.sf.basedb.util.overview.Failure;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.ChildNodeDirection;

/**
	Validator implementation for extracts. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_EXTRACT}
	<li>Access denied: {@link Validator#DENIED_EXTRACT}
	<li>Subtype of parent extract: {@link Validator#INCORRECT_EXTRACT_TYPE}
	<li>Subtype of parent sample: {@link Validator#INCORRECT_SAMPLE_TYPE}
	<li>Different tag than parent: {@link Validator#DIFFERENT_TAG}
	<li>Parent is tagged, but child is not: {@link Validator#TAGGED_PARENT}
	<li>Duplicates of same tag+position on bioassay: {@link Validator#NONUNIQUE_TAG}
	<li>Incorrect position on bioassay: {@link Validator#INCORRECT_BIOASSAY_POSITION}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2012-09-04 10:43:21 +0200 (Tue, 04 Sep 2012) $
*/
public class ExtractValidator
	extends NameableNodeValidator<Extract>
{
	
	public ExtractValidator()
	{
		super(Validator.MISSING_EXTRACT, Validator.DENIED_EXTRACT);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	/**
		Do not report missing items if the parent node is a derived or raw bioassay.
	 */
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		Item parentType = parentNode.getItemType();
		if (parentType == Item.DERIVEDBIOASSAY || parentType == Item.RAWBIOASSAY) return false;
		return super.preMissingItem(dc, context, parentNode);
	}

	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		super.postValidate(dc, context, node, parentNode);

		if (parentNode != null)
		{
			Node parentItemNode = parentNode.getNodeType() == Node.Type.ITEM ? parentNode : parentNode.getParent();
			Item parentType = parentItemNode.getItemType();

			if (parentType == Item.SAMPLE)
			{
				checkParentSampleSubtype(dc, context, node, parentItemNode);
			}
			else if (parentType == Item.EXTRACT)
			{
				if (node.getChildNodeDirection() == ChildNodeDirection.FORWARD)
				{
					checkParentExtractSubtype(dc, context, node, parentItemNode);
					checkExtractTag(dc, context, node, parentItemNode);
				}
				else
				{
					checkParentExtractSubtype(dc, context, parentItemNode, node);
					checkExtractTag(dc, context, parentItemNode, node);
				}
			}
			else if (parentType == Item.PHYSICALBIOASSAY)
			{
				PhysicalBioAssayValidator.checkExtractSubtype(dc, context, parentItemNode, node);
				checkExtractOnBioAssay(dc, context, node, parentItemNode);
			}
			else if (parentType == Item.DERIVEDBIOASSAY)
			{
				DerivedBioAssayValidator.checkExtractSubtype(dc, context, parentItemNode, node);
			}
			else if (parentType == Item.BIOPLATE)
			{
				BioPlateValidator.checkBioMaterialInWell(dc, context, parentItemNode, node);
			}
		}
	}

	
	/**
		The suggested fix is to add parent extracts to the physical
		bioassay or one or more extracts to the pooled extract.
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		Fix fix = null;
		Item parentType = parentNode.getItemType();
		BasicItem parentItem = parentNode.getItem();
		if (parentType == Item.PHYSICALBIOASSAY)
		{
			fix =  new Fix("Add extracts to '" + ((PhysicalBioAssay)parentItem).getName() + "'", parentItem);
		}
		else if (parentType == Item.EXTRACT)
		{
			fix = new Fix("Add parent extracts to '" + ((Extract)parentItem).getName() + "'", parentItem);
		}
		return fix;
	}
	
	/**
		If the parent is a physical bioassay, check that extracts on the
		same position has different tags.
	*/
	@SuppressWarnings("unchecked")
	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, Node parentNode)
	{
		super.postValidateFolder(dc, context, folderNode, parentNode);
		if (parentNode.getItemType() != Item.PHYSICALBIOASSAY) return;

		PhysicalBioAssay pba = (PhysicalBioAssay)parentNode.getItem(dc);
		BioMaterialEvent creationEvent = pba.getCreationEvent();
		
		// Get all positions that are used by the child nodes
		if (folderNode != null)
		{
			// Create a map of tag+position => extract(s)
			Map<TagPosition, Set<Extract>> tagPositions = 
					(Map<TagPosition, Set<Extract>>)context.getCachedObject("TAGPOSITIONS:" + pba.getId());
			if (tagPositions == null)
			{
				tagPositions = new HashMap<ExtractValidator.TagPosition, Set<Extract>>();
				context.setCachedObject("TAGPOSITIONS:" + pba.getId(), tagPositions);

				SpecialQuery<BioMaterialEventSource> query = creationEvent.getEventSources();
				for (BioMaterialEventSource src : query.list(dc))
				{
					try
					{
						Extract extract = (Extract)src.getBioMaterial();
						Tag tag = extract.getTag();
						int position = src.getPosition();
						TagPosition tp = new TagPosition(tag, position);
						Set<Extract> extracts = tagPositions.get(tp);
						if (extracts == null)
						{
							// First extract with this tag+position
							extracts = new HashSet<Extract>();
							tagPositions.put(tp, extracts);
						}
						extracts.add(extract);
					}
					catch (PermissionDeniedException ex)
					{}
				}
			}
			
			// Get the positions for the extracts in the current folder
			Set<Integer> positions = new HashSet<Integer>();
			for (Node extractNode : folderNode.getChildren())
			{
				Extract extract = (Extract)extractNode.getItem(dc);
				BioMaterialEventSource src = creationEvent.getEventSource(extract);
				positions.add(src.getPosition());
			}

			// Check in the map if there are any tag+positions with more than one extract
			for (Map.Entry<TagPosition, Set<Extract>> entry : tagPositions.entrySet())
			{
				Set<Extract> extracts = entry.getValue();
				TagPosition tp = entry.getKey();
				if (extracts.size() > 1 && positions.contains(tp.position))
				{
					// There are duplicates
					Failure failure = context.createFailure(Validator.NONUNIQUE_TAG, parentNode, 
						"Multiple extracts on position " + tp.position + " with tag: " + 
						(tp.tag == null ? "none" : tp.tag.getName()));
					if (failure != null)
					{
						failure.setDescriptor(tp);
						failure.addFix(new Fix("Change extracts on physical bioassay", pba));
						for (Extract e : extracts)
						{
							failure.addFix(new Fix("Change tag of '" + e.getName() + "'", e));
						}
					}
				}
			}
		}
		
	}
	// ----------------------------
	
	
	/**
		Check the subtype of the sample that is the parent to an extract. Given that
		the extract has a non-null subtype, which has a related SAMPLE subtype, this
		method checks if the parent sample has the same subtype. If there is no related
		subtype for the extract, any subtype for the parent is accepted.
		@since 3.0
	*/
	public static void checkParentSampleSubtype(DbControl dc, OverviewContext context, Node extractNode, Node sampleNode)
	{
		Extract extract = (Extract)extractNode.getItem(dc);
		Sample sample = (Sample)sampleNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, extract, Item.SAMPLE, Item.EXTRACT);
			ItemSubtype subtype = sample.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = extract.getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related sample subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);
				if (expectedSubtype.getMainItemType() == Item.SAMPLE)
				{
					context.createFailure(Validator.INCORRECT_SAMPLE_TYPE, extractNode, 
						"Expected a sample with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype of '" + sample.getName() + "' to '" + expectedSubtype.getName() + "'", sample),
						new Fix("Change subtype or parent sample of '" + extractNode.getTitle() + "'", extract),
						subtypeFix
					);
				}
				else
				{
					context.createFailure(Validator.INCORRECT_SAMPLE_TYPE, extractNode, 
						"Expected an extract with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype or parent of '" + extractNode.getTitle() + "'", extract),
						subtypeFix
					);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	/**
		Check the subtype of the extract that is the parent to another extract. Given that
		the child extract has a non-null subtype, which has a related EXTRACT subtype, this
		method checks if the parent extract has the same subtype. If there is no related
		subtype for the child extract, any subtype for the parent is accepted.
		@since 3.0
	*/
	public static void checkParentExtractSubtype(DbControl dc, OverviewContext context, Node childNode, Node parentNode)
	{
		Extract child = (Extract)childNode.getItem(dc);
		Extract parent = (Extract)parentNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, child, Item.EXTRACT, Item.SAMPLE);
			ItemSubtype subtype = parent.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = child.getItemSubtype();
				Fix subtypeFix = subtype == null || subtype.equals(mySubtype) ? null : 
					new Fix("Change related extract subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);
				if (expectedSubtype.getMainItemType() == Item.EXTRACT)
				{
					context.createFailure(Validator.INCORRECT_EXTRACT_TYPE, childNode, 
						"Expected an extract with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype of '" + parent.getName() + "' to '" + expectedSubtype.getName() + "'", parent),
						new Fix("Change subtype or parent extract of '" + childNode.getTitle() + "'", child),
						subtypeFix
					);
				}
				else
				{
					context.createFailure(Validator.INCORRECT_EXTRACT_TYPE, childNode, 
						"Expected a sample with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype or parent of '" + childNode.getTitle() + "'", child),
						subtypeFix
					);
					
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	/**
		Check the tag of the child extract in relation to a parent extract. If the parent
		extract has a tag, the child must have the same tag. If the parent is not
		tagged, then the child may have any tag or no tag.
		@since 3.0
	*/
	public static void checkExtractTag(DbControl dc, OverviewContext context, Node childNode, Node parentNode)
	{
		Extract child = (Extract)childNode.getItem(dc);
		Extract parent = (Extract)parentNode.getItem(dc);
		
		try
		{
			Tag parentTag = parent.getTag();
			if (parentTag != null)
			{
				Tag childTag = child.getTag();
				if (!parentTag.equals(childTag))
				{
					if (childTag != null)
					{
						context.createFailure(Validator.DIFFERENT_TAG, childNode,
							"Extract '" + child.getName() + "' has tag mismatch: " + childTag.getName(),
							new Fix("Change tag of '" + child.getName() + "' to '" + parentTag.getName() + "'", child),
							new Fix("Change tag of '" + parent.getName() + "' to '" + childTag.getName() + "'", parent)
						);
					}
					else
					{
						context.createFailure(Validator.TAGGED_PARENT, childNode,
							"Extract '" + child.getName() + "' has no tag but parent has",
							new Fix("Change tag of '" + child.getName() + "' to '" + parentTag.getName() + "'", child),
							new Fix("Remove tag from '" + parent.getName() + "'", parent)
						);
					}
					
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	/**
		Check if two given extracts have a parent-child relationship.
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public static boolean isRelated(DbControl dc, OverviewContext context, Extract parent, Extract child)
	{
		// Short-cut if both extracts are the same
		if (child.equals(parent)) return true;
		
		String cacheKey = "extract." + child.getId() + ".parents";
		Set<Extract> allParents = (Set<Extract>)context.getCachedObject(cacheKey);
		
		if (allParents == null)
		{
			if (child.getParentType() == Item.EXTRACT)
			{
				ItemQuery<Extract> query = (ItemQuery<Extract>)child.getCreationEvent().getSources();
				query.include(Include.ALL);
				allParents = new HashSet<Extract>();
				new ParentsTransformer<Extract>(dc, true, Extract.getQuery()).transform(query.list(dc), allParents);
			}
			else
			{
				allParents = Collections.emptySet();
			}
			context.setCachedObject(cacheKey, allParents);
		}
		return allParents.contains(parent);
	}

	private void checkExtractOnBioAssay(DbControl dc, OverviewContext context, Node extractNode, Node bioAssayNode)
	{
		// Check that the extract's position is within the expected range
		PhysicalBioAssay bioAssay = (PhysicalBioAssay)bioAssayNode.getItem(dc);
		int bioAssaySize = bioAssay.getSize();
		Extract extract = (Extract)extractNode.getItem(dc);
		BioMaterialEventSource src = bioAssay.getCreationEvent().getEventSource(extract);
		int position = src.getPosition();
		if (position > bioAssaySize || position < 1)
		{
			context.createFailure(Validator.INCORRECT_BIOASSAY_POSITION, bioAssayNode,
					"Incorrect position for '" + extract.getName() + "': " + 
						position + " (expected 1.." + bioAssaySize + ")",
					new Fix("Change position for extract: " + extract.getName(), bioAssay), 
					position <= 0 ? null : new Fix("Increase size of physical bioassay", bioAssay));
		}
	}
	
	
	static class TagPosition
	{
		final Tag tag;
		final int position;
		
		TagPosition(Tag tag, int position)
		{
			this.tag = tag;
			this.position = position;
		}
		
		@Override
		public int hashCode()
		{
			return EqualsHelper.hashCode(tag) + position;
		}
		
		@Override
		public boolean equals(Object o)
		{
			TagPosition other = (TagPosition)o;
			return EqualsHelper.equals(this.tag, other.tag) && this.position == other.position;
		}
		
	}
	
}
