/**
	$Id: ProtocolValidator.java 5814 2011-10-18 11:46:59Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for protocols. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_PROTOCOL}
	<li>Access denied: {@link Validator#DENIED_PROTOCOL}
	<li>Incorrect protocol type: {@link Validator#INCORRECT_PROTOCOLTYPE} 
	<li>Non-default protocol: {@link Validator#NONDEFAULT_PROTOCOL}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-18 13:46:59 +0200 (Tue, 18 Oct 2011) $
*/
public class ProtocolValidator
	extends NameableNodeValidator<Protocol>
{
	
	public ProtocolValidator()
	{
		super(null, Validator.DENIED_PROTOCOL);
	}

	
	/* 
		From BasicNodeValidator class
		-----------------------------
	*/
	/**
		If the parent item has a subtype that is related to a PROTOCOL subtype, report
		the missing item as a {@link Validator#MISSING_PROTOCOL} failure. Otherwise,
		ignore the missing protocol.
		@return Always false
	*/
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedProtocolSubtype(dc, parentItem);
		if (expectedSubtype != null)
		{
			context.createFailure(Validator.MISSING_PROTOCOL, parentNode, 
					null, getMissingItemFix(dc, parentNode));
		}
		return false;
	}
	
	/**
		Checks if the protocol is of the correct protocol type
		and, if a project is active, the protocol is the same as
		the default for the project.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		Protocol protocol = (Protocol)node.getItem(dc);
		
		// Validate the subtype of the protocol
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedProtocolSubtype(dc, parentItem);
		try
		{
			ItemSubtype subtype = protocol.getItemSubtype();
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = ((Subtypable)parentItem).getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related protocol subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);

				context.createFailure(Validator.INCORRECT_PROTOCOLTYPE, parentNode, 
					"Expected a protocol with subtype: " + expectedSubtype.getName(),
					new Fix("Change protocol of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Change protocol type of '" + protocol.getName() + "' to '" + expectedSubtype.getName() + "'", protocol),
					subtypeFix
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}

		// Check if using project default protocol
		Project project = context.getProject();
		if (project != null)
		{
			if (!project.isDefaultItem(protocol) && project.hasDefaultForItem(protocol))
			{
				context.createFailure(Validator.NONDEFAULT_PROTOCOL, parentNode, null,
					new Fix("Change protocol of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Add '" + protocol.getName() + "' as default protocol to project", project)
				);
			}
		}
	}
	
	/**
		@return Suggested fix is to add a protocol to the parent item
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem();
		return new Fix("Add protocol to '" + parentNode.getTitle() + "'", parentItem);
	}
	// --------------------------------
	
	/**
		Get the subtype of the associated protocol that we expect for the given
		parent item.
	*/
	protected ItemSubtype getExpectedProtocolSubtype(DbControl dc, BasicItem parentItem)
	{
		ItemSubtype expectedSubtype = null;
		int defaultSubtypeForItemType = SystemItems.getId(Protocol.getDefaultSystemId(parentItem.getType()));
		try
		{
			if (parentItem instanceof Subtypable)
			{
				expectedSubtype = ItemSubtype.getRelatedSubtype(dc, 
						(Subtypable)parentItem, Item.PROTOCOL, defaultSubtypeForItemType, 0);
			}
			else if (defaultSubtypeForItemType != 0)
			{
				expectedSubtype = ItemSubtype.getById(dc, defaultSubtypeForItemType);
			}
		}
		catch (PermissionDeniedException ex)
		{}
		return expectedSubtype;
	}

	
}
