/**
	$Id: ZipUnpacker.java 5758 2011-09-23 07:05:23Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.zip;

import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;

/**
 	This class can be used to unpack zip files into the BASE directory structure.
 	The class supports unpacking multiple zip files in the same transaction.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2011-09-23 09:05:23 +0200 (Fri, 23 Sep 2011) $
*/
public class ZipUnpacker
{

	private final DbControl dc;
	
	/**
		Cache of directories.
	*/
	private Map<String, Directory> paths;
	
	/**
		Create a new zip unpacker.
		@param dc The DbControl to use for database access
	*/
	public ZipUnpacker(DbControl dc)
	{
		this.dc = dc;
		this.paths = new HashMap<String, Directory>();
	}
	
	/**
		Unpack a zip file.

		@param rootDir The directory where the contents of the zip file should be placed
		@param zip The contents of the zip file
		@param overwrite If existing files should be overwritten or ignored
		@param progress An optional progress reporter for displaying progress
		@return The number of files unpacked
		@throws IOException If there is an error when unpacking the zip file
		@throws BaseException If there is a BASE error
		@since 3.0
	*/
	public int unpackZipFile(Directory rootDir, ZipInputStream zip, File sourceFile, boolean overwrite, 
		AbsoluteProgressReporter progress)
		throws IOException, BaseException
	{
		ZipEntry entry = null;
		String rootPath = rootDir.getPath().toString();
		paths.put(rootPath, rootDir);
		
		long totalUnpacked = 0;
		int numFiles = 0;
		
		while ((entry = zip.getNextEntry()) != null)
		{
			ThreadSignalHandler.checkInterrupted();
			String subPath = entry.getName();
			String completePath = rootPath + "/" + subPath;
			
			boolean isDirectory = entry.isDirectory();
			if (progress != null)
			{
				progress.displayAbsolute(totalUnpacked, 
					"Unpacking " + (isDirectory ? "directory: " : "file: ") + completePath + 
					" (" +Values.formatBytes(entry.getSize()) + ")");
			}
			if (!isDirectory)
			{
				Path filePath = new Path(completePath, Path.Type.FILE);

				// Get the parent directory
				Directory parentDir = getDirectory(filePath);
				
				// Get the file or create a new one
				File file = File.getFile(dc, parentDir, filePath.getFilename(), true);
				if (!file.isInDatabase() || file.getLocation() == Location.OFFLINE || overwrite)
				{
					if (!file.isInDatabase()) dc.saveItem(file);
					file.setMimeTypeAuto(null, null);
					OutputStream upload = file.getUploadStream(false);
					PackUtil.copyProperties(sourceFile, file);
					try
					{
						FileUtil.copy(zip, upload);
						numFiles++;
					}
					finally
					{
						upload.close();
					}
				}
			}
			totalUnpacked += entry.getCompressedSize();
			zip.closeEntry();
		}
		if (progress != null) progress.display(100, "Done");
		return numFiles;
	}
	
	/**
		Get the directory for the specified path. We first check the internal cache.
		If not found we try to load it from the database. If still not found we create
		a new directory and save it to the database.
		@param path The path
		@return The directory
	*/
	private Directory getDirectory(Path path)
	{
		// Make sure it is a path to the directory
		if (path.getType() == Path.Type.FILE) path = path.getParent();
		String thePath = path.toString();
		
		// Load directory from cache
		Directory dir = paths.get(thePath);
		
		if (dir == null)
		{
			// Load directory from database
			try
			{
				dir = Directory.getByPath(dc, path);
				dir.setRemoved(false);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		if (dir == null)
		{
			// Directory is not in the cache and not in the database --> create a new one
			// Get the parent recursively
			Directory parent = getDirectory(path.getParent());
			dir = Directory.getNew(dc, parent);
			dc.saveItem(dir);
			dir.setName(path.getDirectory(path.getDirectoryCount()-1));
		}
		paths.put(thePath, dir);
		return dir;	
	}
}
