/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.Set;

import org.apache.tools.bzip2.CBZip2InputStream;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Location;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.zip.AbstractFileUnpacker;
import net.sf.basedb.util.zip.PackUtil;

/**
	A plugin for unpacking a bzip2 compressed file to
	the BASE file system.
	
	@author nicklas
	@since 3.2
	@base.modified $Date$
*/
public class Bzip2FileUnpacker
	extends AbstractFileUnpacker
{

	private static final Set<String> extensions =  
		new LinkedHashSet<String>(Arrays.asList("bz2"));

	private static final Set<String> mimeTypes =  
		new LinkedHashSet<String>(Arrays.asList("application/x-bzip2"));
	
	/*
		From the FileUnpacker interface
		-------------------------------------------
	*/
	public String getFormatName()
	{
		return "BZip2 compressed files";
	}
	/**
		Return a set containing the strings [ bz2 ].
	*/
	public Set<String> getExtensions()
	{
		return extensions;
	}
	/**
		Return a set containing the strings [ application/x-bzip2 ].
	*/
	public Set<String> getMimeTypes()
	{
		return mimeTypes;
	}
	@Override
	public int unpack(DbControl dc, Directory rootDir, InputStream in,
		File sourceFile, boolean overwrite, AbsoluteProgressReporter progress)
		throws IOException
	{
		// Get the file or create a new one
		File file = File.getFile(dc, rootDir, sourceFile.getName().replace(".bz2", ""), true);
		
		if (progress != null)
		{
			progress.displayAbsolute(0, 
				"Unpacking file: " + sourceFile.getName() + " › " + file.getName());
		}
		
		// Strip first two bytes 'BZ', since they are not part of the file format
		in.read();
		in.read();
		
		CBZip2InputStream bzipStream = new CBZip2InputStream(new BufferedInputStream(in));
		
		if (!file.isInDatabase() || file.getLocation() == Location.OFFLINE || file.isRemoved() || overwrite)
		{
			if (!file.isInDatabase()) dc.saveItem(file);
			file.setMimeTypeAuto(null, null);
			OutputStream upload = file.getUploadStream(false);
			PackUtil.copyProperties(sourceFile, file);
			try
			{
				FileUtil.copy(bzipStream, upload);
			}
			finally
			{
				upload.close();
			}
		}

		if (progress != null) progress.display(100, "Done");
		return 1;
	}
	// -------------------------------------------


}
