/*
	$Id: LowessNormalization.java 5689 2011-08-11 09:09:06Z nicklas $
	
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007, 2008 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.FloatParameterType;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.query.WhenStatement;
import net.sf.basedb.core.signal.EnhancedThreadSignalHandler;
import net.sf.basedb.core.signal.Signal;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;

/**
	 @author enell, Nicklas
	 @version 2.0
	 @base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
*/

public class LowessNormalization
	extends AbstractAnalysisPlugin
	implements InteractivePlugin, SignalTarget
{

	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	private static final PluginParameter<Float> fitFractionParameter = new PluginParameter<Float>
	(
		"fitFraction",
		"Window size (fraction of points)",
		"Smoothing parameter. The higher it is the smoother the funtion will be",
		new FloatParameterType(0F, 1F, 0.33F, true)
	);
	
	private static final PluginParameter<Float> deltaParameter = new PluginParameter<Float>
	(
		"delta",
		"Minimum log10(intensity) step",
		"",
		new FloatParameterType(0F, null, 0.043F, true)
	);
	
	private static final PluginParameter<Integer> iterParameter = new PluginParameter<Integer>
	(
		"iter",
		"Iterations",
		"",
		new IntegerParameterType(1, null, 4, true)
	);

	private static final PluginParameter<Integer> blockGroupParameter = new PluginParameter<Integer>
	(
		"blockGroup",
		"Blockgroup size",
		"The number of arraydesign blocks in each group.\n0 = use all blocks as a single group.",
		new IntegerParameterType(0, null, 0, true)
	);
	
	/**
		Section for lowess options
	*/
	protected static final PluginParameter<String> lowessSection = new PluginParameter<String>(
		"lowessSection",
		"Lowess options",
		"", 
		null
	);

	private static double ywRangeFactor = 6;
	
	private RequestInformation configureJob;

	private EnhancedThreadSignalHandler signalHandler;
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request use access to Experiment:s and read access to Job:s.
		@since 2.0.2
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.JOB, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.REPORTERLIST, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}

	public void run(Request request, Response response, ProgressReporter progress)
	{
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			if (signalHandler != null) signalHandler.setWorkerThread();
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				BioAssaySet source = getSourceBioAssaySet(dc);
				String childName = Values.getString((String)job.getValue(CHILD_NAME), source.getName());
				String childDescription = (String)job.getValue(CHILD_DESCRIPTION);
				int blockGroupSize = (Integer) job.getValue(blockGroupParameter.getName());
				float delta = (Float) job.getValue(deltaParameter.getName());
				float fitFraction = (Float) job.getValue(fitFractionParameter.getName());
				int iterations = (Integer) job.getValue(iterParameter.getName());
				ReporterList excludeReporters = (ReporterList)job.getValue("excludeReporters");
				Job thisJob = getCurrentJob(dc);
				
				BioAssaySet child = normalize(dc, source, thisJob, fitFraction, delta, iterations, 
					blockGroupSize, excludeReporters, progress);
				child.setName(childName);
				child.setDescription(childDescription);
				dc.commit();
				int normalizedSpots = child.getNumSpots();
				if (progress != null) progress.display(100, normalizedSpots + " spots normalized\n");
				response.setDone(normalizedSpots + " spots normalized, " + (source.getNumSpots() - normalizedSpots) + " spots removed");
			}
			catch (SignalException ex)
			{
				if (signalHandler != null && signalHandler.hasReceived(Signal.SHUTDOWN))
				{
					response.setContinue(command);
				}
				else
				{
					response.setError("Aborted by user", Arrays.asList(ex));
				}
			}
			catch (Throwable ex)
			{
				response.setError(ex.getMessage(), Arrays.asList(ex));
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		else
		{
			response.setError("Unknown command: " + command, null);
		}
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);
		if (message == null)
		{
			BioAssaySet bas = (BioAssaySet)item;
			int channels = bas.getRawDataType().getChannels();
			if (channels != 2)
			{
				message = "Lowess normalization requires 2-channel data, not " + channels + "-channel.";
			}
			else if (bas.getNumSpots() <= 0)
			{
				message = "Lowess normalization requires spot data in the database";
			}
			else if (bas.getMaxRawMappingsForSpot() != 1)
			{
				message = "Lowess normalization can't be done on merged data.";
			}
		}
		return message;
	}

	public RequestInformation getRequestInformation(GuiContext context, String command) throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters();
		}
		return requestInformation;
	}

	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				// Validate job parameters
				RequestInformation ri = getConfigureJobParameters();
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}

				// Source bioassay set
				storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
								
				// Child name, description and transformation
				storeValue(job, request, ri.getParameter(CHILD_NAME));
				storeValue(job, request, ri.getParameter(CHILD_DESCRIPTION));

				// Lowess options
				storeValue(job, request, fitFractionParameter);
				storeValue(job, request, deltaParameter);
				storeValue(job, request, iterParameter);
				storeValue(job, request, blockGroupParameter);
				storeValue(job, request, ri.getParameter("excludeReporters"));
				
				BioAssaySet bioAssaySet = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
				response.setSuggestedJobName("Lowess normalization on bioassay set '" + 
						bioAssaySet.getName() + "'");
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------

	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		if (signalHandler == null) 
		{
			signalHandler = new EnhancedThreadSignalHandler(Signal.ABORT, Signal.SHUTDOWN);
		}
		return signalHandler;
	}
	// -------------------------------------------

	/**
		Normalise the source bioassay set using LOWESS normalization.
		All reporters are included.
		@since 2.5
		@see #normalize(DbControl, BioAssaySet, Job, float, float, int, int, ReporterList, ProgressReporter)
	*/
	public BioAssaySet normalize(DbControl dc, BioAssaySet source, Job job, float fitFraction, float delta, int iterations, int blockGroupSize, ProgressReporter progress)
	{
		return normalize(dc, source, job, fitFraction, delta, iterations, blockGroupSize, null, progress);
	}
	
	
	/**
		Normalise the source bioassay set using LOWESS normalization.
		@param dc The DbControl to use for database access
		@param source The source bioassay set that is going to be normalized
		@param job The job that is doing the normalization, or null
		@param fitFraction
		@param delta
		@param iterations
		@param blockGroupSize The number of blocks to group and normalise at the same time,
			if <=0, all blocks on a bioassay are grouped to a single dataset
		@param excludeReporters A reporter list containing reporters that should not be
			used in when normalizing, or null to use all spots
	 	@param progress Progress reporter used by the caller to keep track of the progress. Null is allowed   
		@return The normalized bioassayset
		@since 2.6
	*/
	public BioAssaySet normalize(DbControl dc, BioAssaySet source, Job job, float fitFraction, float delta, int iterations, int blockGroupSize, ReporterList excludeReporters, ProgressReporter progress)
	{
		if (progress != null) progress.display(0, "Preparing to normalize...");
		
		// Create Transformation
		Transformation t = source.newTransformation(job);
		t.setName("Normalization: Lowess");
		dc.saveItem(t);
		
		// Create the normalized bioassay set
		BioAssaySet child = t.newProduct(null, "new", true);
		dc.saveItem(child);
		
		// Batcher for inserting normalized data
		SpotBatcher batcher = child.getSpotBatcher();

		// Expressions used to get data
		Expression block = Dynamic.rawData("block");
		Expression ch1 = Dynamic.column(VirtualColumn.channelIntensity(1));
		Expression ch2 = Dynamic.column(VirtualColumn.channelIntensity(2));

		// Create restriction: ch1 > 0 and ch2 > 0
		Restriction intensityRestriction = Restrictions.and(
				Restrictions.gt(ch1, Expressions.aFloat(0)), 
				Restrictions.gt(ch2, Expressions.aFloat(0))
			);

		// Create restriction: column = :bioAssayColumn
		Restriction bioAssayRestriction = Restrictions.eq(
				Dynamic.column(VirtualColumn.COLUMN), 
				Expressions.parameter("bioAssayColumn")
			);
			
		// Build query for loading spots
		DynamicSpotQuery query = source.getSpotData();
		query.restrictPermanent(intensityRestriction);
		
		// If some reporters should be excluded, create an expression that:
		// = 0 if spot's reporter is not present in list, = 1 if it is
		Expression isPresent = Expressions.integer(0);
		if (excludeReporters != null)
		{
			query.joinReporterList(excludeReporters, JoinType.LEFT);
			isPresent = Expressions.caseWhen(isPresent, 
				new WhenStatement(Dynamic.isPartOf(excludeReporters), Expressions.integer(1)));
		}
		
		// Create query to retrieve spot data: COLUMN, POSITION, ch1, ch2, block, present
		// We use a parameter to restrict the query to return data for one bioassay at a time
		// The 'present' value is 0 to not exclude and 1 to exclude a spot
		query.select(Dynamic.select(VirtualColumn.POSITION));
		query.select(Selects.expression(ch1, "ch1"));
		query.select(Selects.expression(ch2, "ch2"));
		query.select(Selects.expression(block, "block"));
		query.select(Selects.expression(isPresent, "present"));
		query.restrict(bioAssayRestriction);
		query.order(Orders.asc(block));
		
		// Normalize one bioassay at a time
		List<BioAssay> assays = source.getBioAssays().list(dc);
		int numAssays = assays.size();
		int nextAssayIndex = 0;
		int maxThreads = Math.min(4, Runtime.getRuntime().availableProcessors());
		int maxQueueSize = 2 * maxThreads;
		int numDone = 0;
		int numWorking = 0;
		
		// Set up execution service for thread management
		final ThreadGroup workers = new ThreadGroup("Lowess-threads");
		ThreadFactory threadFactory = new ThreadFactory()
		{
			private int i=0;
			@Override
			public Thread newThread(Runnable r) 
			{
				++i;
				return new Thread(workers, r, "Lowess-worker-"+i);
			}
		};
		final ExecutorService service = Executors.newFixedThreadPool(maxThreads, threadFactory);
		final CompletionService<List<SpotData>> cService = 
			new ExecutorCompletionService<List<SpotData>>(service);
		try
		{
			// Loop until all bioassays has been processed
			while (numDone < numAssays)
			{
				ThreadSignalHandler.checkInterrupted();

				// Get data for the next bioassay unless the working queue
				// is full
				if (nextAssayIndex < numAssays && numWorking < maxQueueSize)
				{
					List<SpotData> data = readData(dc, query, assays.get(nextAssayIndex));
					nextAssayIndex++;
					if (data.size() > 0) 
					{
						CallableWorker worker = new CallableWorker(data, fitFraction, delta, iterations, blockGroupSize);
						cService.submit(worker);
						numWorking++;
					}
					else
					{
						numDone++;
					}
				}
				
				// If the working queue is full we use take() to wait until a result is
				// ready, otherwise we use poll() which let's us continue if no results
				// are completed
				Future<List<SpotData>> result = numWorking >= maxQueueSize ? 
					cService.take() : cService.poll();
				if (result != null)
				{
					numDone++;
					numWorking--;
					List<SpotData> data = result.get();
					for (SpotData spot : data)
					{
						batcher.insert(spot.column, spot.position, spot.newCh1, spot.newCh2);
					}
					batcher.flush();
					if (progress != null)
					{
						int percent = (100 * numDone) / numAssays;
						progress.display(percent, numDone + " of " + numAssays + 
							" assays done.");
					}
				}
			}
			batcher.close();
		}
		catch (SQLException e)
		{
			throw new BaseException(e);
		}
		catch (ExecutionException ex)
		{
			throw new BaseException(ex.getCause());
		}
		catch (InterruptedException ex)
		{
			if (signalHandler != null)
			{
				signalHandler.checkForSignals();
			}
			throw new SignalException("Aborted by user", ex);
		}
		finally
		{
			service.shutdownNow();
		}
		return child;
	}
	
	private List<SpotData> readData(DbControl dc, DynamicSpotQuery query, BioAssay assay)
		throws SQLException
	{
		// Prepare list for holding data
		int assaySpots = assay.getNumSpots();
		List<SpotData> data = new ArrayList<SpotData>(assaySpots);
		
		// Load spot data for this bioassay
		short bioassayColumn = assay.getDataCubeColumnNo();
		query.setParameter("bioAssayColumn", (int)bioassayColumn, Type.INT);
		
		DynamicResultIterator it = query.iterate(dc);
		int positionIndex = it.getIndex(VirtualColumn.POSITION.getName());
		int ch1Index = it.getIndex("ch1");
		int ch2Index = it.getIndex("ch2");
		int blockIndex = it.getIndex("block");
		int presentIndex = it.getIndex("present");
		
		// Copy bioassay data to SpotData objects
		while (it.hasNext())
		{
			ThreadSignalHandler.checkInterrupted();
			SqlResult r = it.next();
			SpotData spot = new SpotData(bioassayColumn, r.getInt(positionIndex),
				r.getFloat(ch1Index), r.getFloat(ch2Index), r.getInt(blockIndex), 
				r.getInt(presentIndex) == 1);
			data.add(spot);
		}
		it.close();
		return data;
	}

	private RequestInformation getConfigureJobParameters()
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				dc = sc.newDbControl();
				BioAssaySet bas = getCurrentBioAssaySet(dc);
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>(3);
				
				// Source and child bioassay set parameters
				parameters.add(getSourceBioAssaySetParameter(null, null));
				parameters.add(getChildNameParameter(null, null, bas.getName()));
				parameters.add(getChildDescriptionParameter(null, null, null));				
				
				// Lowess options
				parameters.add(lowessSection);
				parameters.add(fitFractionParameter);
				parameters.add(deltaParameter);
				parameters.add(iterParameter);
				parameters.add(blockGroupParameter);
				PluginParameter<ReporterList> excludeReporters = 
					getExcludeReportersParameters(dc);
				if (excludeReporters != null) parameters.add(excludeReporters);
	
				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Configure job",
					"Set window size, minimum step and number of iterations",
					parameters
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}

	private PluginParameter<ReporterList> getExcludeReportersParameters(DbControl dc)
	{
		// Load reporter types and initialise the reporterTypeParameter
		ItemQuery<ReporterList> query = ReporterList.getQuery();
		query.order(Orders.asc(Hql.property("name")));
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT);
		List<ReporterList> reporterLists = new ArrayList<ReporterList>(query.list(dc));
		if (reporterLists.size() == 0) return null;
		
		ItemParameterType<ReporterList> reporterListType = 
			new ItemParameterType<ReporterList>(ReporterList.class, null, false, 1, reporterLists);
		PluginParameter<ReporterList> reporterListParameter = new PluginParameter<ReporterList>(
			"excludeReporters",
			"Exclude reporters",
			"Spots with reporters in the selected list will not be used in the lowess calculations. " +
			"If no list is selected all spots are used.",
			reporterListType);
		return reporterListParameter;
	}
	
	private static List<Double> lowess(List<SpotData> data, double f, int iter, double delta)
	{
		int dataSize = data.size();
		Double[] smoothCurve = new Double[dataSize];
		int windowSize = Math.min(dataSize, (int) (dataSize * f + 0.5));

		List<Double> wFit = new ArrayList<Double>();
		wFit.addAll(Collections.nCopies(dataSize, 1D));
		for (int iteration = 0; iteration < iter; ++iteration)
		{
			if (Thread.interrupted()) return null;
			int windowStart = 0;
			int i = 0;
			int prevI = -1;
			while (prevI < dataSize - 1)
			{
				double Ai = data.get(i).A;
				// center window around the i:th value in A
				// while distance from windowStart to i is greater then distance from i to windowEnd: move window
				while ((windowStart + windowSize < dataSize) && 
						(Ai - data.get(windowStart).A > data.get(windowStart + windowSize-1).A - Ai))
				{
					windowStart++;
				}
				
				List<SpotData> window = data.subList(windowStart, windowStart + windowSize);
				List<Double> wFitWindow = wFit.subList(windowStart, windowStart + windowSize);
				
				List<Double> w = calculateWeights(window, Ai, wFitWindow);
				double[] km = weightedLeastSquaresRegression(window, w);
				smoothCurve[i] = km[0] * Ai + km[1];

				// Interpolate skipped points due to delta
				if (prevI + 1 < i)
				{
					double d = Ai - data.get(prevI).A;
					if (d == 0)
					{
						for (int j = prevI + 1; j < i; j++)
						{
							smoothCurve[j] = smoothCurve[i];
						}
					}
					else
					{
						for (int j = prevI + 1; j < i; j++)
						{
							double t = (data.get(j).A - data.get(prevI).A) / d;
							smoothCurve[j] = t * smoothCurve[i] + (1D - t) * smoothCurve[prevI];
						}
					}
				}

				// increase i, next x value must be at least delta greater then current
				prevI = i;
				double cut = Ai + delta;
				while (i < dataSize-1 && data.get(i).A <= cut)
				{
					i++;
				}
			}
			
			double invYWRange = 1D/(ywRangeFactor  * medianCorrection(data, smoothCurve));
			for (int j = 0; j < wFit.size(); j++)
			{
				double w = Math.abs((smoothCurve[j] - data.get(j).M) * invYWRange);
				wFit.set(j, w < 1 ? Math.pow(1D - w*w, 2) : 0);
			}
		}
		return Arrays.asList(smoothCurve);
	}

	private static double medianCorrection(List<SpotData> data, Double[] smoothCurve)
	{
		List<Double> temp = new ArrayList<Double>(data.size());
		for (int i = 0; i < data.size(); i++)
		{
			temp.add(Math.abs(data.get(i).M - smoothCurve[i]));
		}
		Collections.sort(temp);
		if (temp.size() % 2 == 0)
		{
			double v1 = temp.get((temp.size() / 2) - 1);
			double v2 = temp.get((temp.size() / 2));
			return 0.5 * (v1 + v2);
		}
		return temp.get(temp.size() / 2);
	}

	
	private static double[] weightedLeastSquaresRegression(List<SpotData> data, List<Double> w)
	{
		double k;
		double m;
		double sumA = 0;
		double sumM = 0;
		double sumAA = 0;
		double sumAM = 0;
		double sumW = 0;
		for (int j = 0; j < data.size(); j++)
		{
			SpotData spot = data.get(j);
			if (!spot.exclude)
			{
				double localW = w.get(j);
				double localA = spot.A;
				double localM = spot.M;
				sumA += localA * localW;
				sumM += localM * localW;
				sumAA += localA * localA * localW;
				sumAM += localA * localM * localW;
				sumW += localW;
			}
		}
		if (sumW <= 0)
		{
			throw new BaseException("Sum of weigths in line_fit is not positive. " +
					"This may happen if there are too few data points in a block group. Please " +
					"increase the 'block group' parameter, or set it to 0 to include all blocks " +
					"in a single group");
		}
		double denom = sumW * sumAA - sumA * sumA;
		if (denom != 0.)
		{
			k = (sumW * sumAM - sumA * sumM) / denom;
			m = (sumM - k * sumA) / sumW;
		}
		else
		{
			k = 0.;
			m = sumM / sumW;
		}

		return new double[] {k, m};
	}

	private static List<Double> calculateWeights(List<SpotData> data, double A1, List<Double> wFit)
	{
		List<Double> w = new ArrayList<Double>(data.size());
		double invRadius = 0;
		for (SpotData spot : data)
		{
			double abs = Math.abs(A1 - spot.A);
			if (abs > invRadius) invRadius = abs;
		}
		invRadius = 1 / invRadius;
		for (int i = 0; i < data.size(); i++)
		{
			if (!data.get(i).exclude)
			{
				double A2 = data.get(i).A;
				double distance = Math.abs(A1 - A2) * invRadius;
				w.add((distance < 1 ? Math.pow(1D - Math.pow(Math.abs(distance), 3), 3) : 0) * wFit.get(i));
			}
			else
			{
				w.add(0.0);
			}
		}
		return w;
	}

	private static class SpotData
		implements Comparable<SpotData>
	{
		final short column;
		final int position;
		final float ch1;
		final float ch2;
		final double M;
		final double A;
		final int block;
		final boolean exclude;
		float newCh1;
		float newCh2;
		static final double INV_LN2 = 1 / Math.log(2);

		public SpotData(short column, int position, float ch1, float ch2, int block, boolean exclude)
		{
			this.column = column;
			this.position = position;
			this.ch1 = ch1;
			this.ch2 = ch2;
			this.block = block;
			this.exclude = exclude;
			this.M = Math.log(ch1 / ch2) * INV_LN2;
			this.A = Math.log10(Math.sqrt(ch1 * ch2));
		}

		/**
			Sort by A in ascending order.
		*/
		public int compareTo(SpotData o)
		{
			return this.A < o.A ? -1 : 1;
		}
		
	}
	
	static class CallableWorker
		implements Callable<List<SpotData>>
	{
	
		private final List<SpotData> in;
		private final float fitFraction;
		private final float delta;
		private final int iterations;
		private int blockGroupSize;
		
		CallableWorker(List<SpotData> in,  
			float fitFraction, float delta, int iterations, int blockGroupSize)
		{
			this.in = in;
			this.fitFraction = fitFraction;
			this.delta = delta;
			this.iterations = iterations;
			this.blockGroupSize = blockGroupSize;
		}
	
		@Override
		public List<SpotData> call()
		{
			normalize(in);
			return in;
		}
		
		/**
			Normalize the spot data.
			@return TRUE if all was ok, FALSE if worker has been interrupted and should
				stop executing
		*/
		private boolean normalize(List<SpotData> data)
		{
			int dataSize = data.size();
			
			// Get range of block numbers - NOTE! query must return spots sorted in block order
			int minBlock = data.get(0).block;
			int maxBlock = data.get(data.size()-1).block;
			// If 0, we should use all blocks as a single data set
			if (blockGroupSize <= 0) blockGroupSize = maxBlock - minBlock + 1;
			
			int fromIndex = 0;
			int toIndex = 0;
			int fromBlock = minBlock;
			// Normalize each block range independently: fromBlock + blockGroupSize --> toBlock
			while (fromBlock <= maxBlock)
			{
				if (Thread.interrupted()) return false;
				
				// Find start and end index for current block range
				int toBlock = fromBlock + blockGroupSize - 1;
				if (toBlock > maxBlock) toBlock = maxBlock;
				fromIndex = toIndex;
				// Data is sorted by block. Find index of last spot with: block <= toBlock
				// spot given by toIndex should not be included
				while (toIndex < dataSize && data.get(toIndex).block <= toBlock)
				{
					++toIndex;
				}
				
				if (toIndex > fromIndex)
				{
					List <SpotData> toNormalize = data.subList(fromIndex, toIndex);
					Collections.sort(toNormalize);
					List<Double> smoothCurve = lowess(toNormalize, fitFraction, iterations, delta);
					if (smoothCurve == null) return false;
					for (int j = 0; j < smoothCurve.size(); ++j)
					{
						if (Thread.interrupted()) return false;
						SpotData spot = toNormalize.get(j);
						double factor = Math.pow(2,smoothCurve.get(j) * 0.5);
						spot.newCh1 = (float)(spot.ch1/factor);
						spot.newCh2 = (float)(spot.ch2*factor);
					}
				}
				fromBlock = toBlock + 1;
			}
			return true;
		}
	}
		
}
