/**
	$Id: ExtractImporter.java 5899 2011-12-06 08:17:37Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BioMaterial;
import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.BioMaterialEventSource;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Tag;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.DefaultItemFunction;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing extract items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2011-12-06 09:17:37 +0100 (Tue, 06 Dec 2011) $
*/
public class ExtractImporter
	extends AbstractItemImporter<Extract>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.EXTRACT, GuiContext.Type.LIST));

	protected static final PluginParameter<String> tagColumnMapping = new PluginParameter<String>(
			"tagColumnMapping",
			"Tag",
			"Mapping that picks the name or ID of a tag from the data columns. The plug-in will first " +
			"try to find a tag with the given name. If none is found and the value is numeric it will " +
			"try to load by internal ID. " +
			"Example: \\Tag\\",
			optionalColumnMapping
			);

	public ExtractImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper externalIdMapper;
	private Mapper tagMapper;
	private Mapper originalQuantityMapper;
	private Mapper createdMapper;
	private Mapper protocolMapper;
	private DefaultItemFunction<Protocol> defaultPrototolFunction;
	private Mapper parentTypeMapper;
	private Mapper parentMapper;
	private Mapper usedQuantityMapper;
	private Mapper bioPlateMapper;
	private Mapper bioWellRowMapper;
	private Mapper bioWellColMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link PropertyIdMethod#EXTERNAL_ID}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME, PropertyIdMethod.EXTERNAL_ID, 
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	@Override
	protected Item getItemForSubtypes()
	{
		return Item.EXTRACT;
	}

	/**
		Adds column mappings for name, externalId, description,
		protocol, creation date, original quantity, pooled, parent
		and used quantity from parent.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(subtypeColumnMapping);
		parameters.add(externalIdColumnMapping);
		parameters.add(tagColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(protocolColumnMapping);
		
		parameters.add(SampleImporter.createdColumnMapping);
		parameters.add(SampleImporter.originalQuantityColumnMapping);
		parameters.add(SampleImporter.parentTypeColumnMapping);
		parameters.add(SampleImporter.parentColumnMapping);
		parameters.add(SampleImporter.usedQuantityColumnMapping);
		parameters.add(SampleImporter.bioPlateColumnMapping);
		parameters.add(SampleImporter.bioWellRowColumnMapping);
		parameters.add(SampleImporter.bioWellColColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Calls {@link Extract#getQuery()}.
	*/
	@Override
	protected ItemQuery<Extract> createItemQuery()
	{
		return Extract.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? Extract.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? Extract.MAX_DESCRIPTION_LENGTH : null, null);
		externalIdMapper = getMapper(ffp, (String)job.getValue("externalIdColumnMapping"), 
				cropStrings ? Extract.MAX_EXTERNAL_ID_LENGTH : null, null);
		tagMapper = getMapper(ffp, (String)job.getValue("tagColumnMapping"), null, null);
		originalQuantityMapper = getMapper(ffp, (String)job.getValue("originalQuantityColumnMapping"), null, null);
		createdMapper = getMapper(ffp, (String)job.getValue("createdColumnMapping"), null, null);
		defaultPrototolFunction = getDefaultItemFunction(Item.PROTOCOL);
		protocolMapper = getMapper(ffp, (String)job.getValue("protocolColumnMapping"), null, null, defaultPrototolFunction);
		parentTypeMapper = getMapper(ffp, (String)job.getValue("parentTypeColumnMapping"), null, null);
		parentMapper = getMapper(ffp, (String)job.getValue("parentColumnMapping"), null, null);
		usedQuantityMapper = getMapper(ffp, (String)job.getValue("usedQuantityColumnMapping"), null, null);
		bioPlateMapper = getMapper(ffp, (String)job.getValue("bioPlateColumnMapping"), null, null);
		bioWellRowMapper = getMapper(ffp, (String)job.getValue("bioWellRowColumnMapping"), null, null);
		bioWellColMapper = getMapper(ffp, (String)job.getValue("bioWellColColumnMapping"), null, null);
	}
	
	@Override
	protected Extract createItem(DbControl dc, FlatFileParser.Data data)
	{
		Extract e = Extract.getNew(dc);
		updateItem(dc, e, data);
		return e;
	}
	
	@Override
	protected void updateItem(DbControl dc, Extract extract, FlatFileParser.Data data)
	{
		if (nameMapper != null) extract.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) extract.setDescription(descriptionMapper.getValue(data));
		if (externalIdMapper != null) extract.setExternalId(externalIdMapper.getValue(data));
		if (originalQuantityMapper != null) extract.setOriginalQuantity(originalQuantityMapper.getFloat(data));
		BioMaterialEvent creationEvent = extract.getCreationEvent();
		if (createdMapper != null) creationEvent.setEventDate(parseDate(createdMapper.getValue(data)));
		updateItemSubtype(dc, extract, data);

		if (tagMapper != null) 
		{
			String nameOrId = tagMapper.getValue(data);
			ItemSubtype type = ItemSubtype.getRelatedSubtype(dc, extract, Item.TAG, 0);
			Tag tag = findTag(dc, FallbackIdMethod.NAME_OR_ID, nameOrId, type);
			if (nameOrId == null || tag != null) extract.setTag(tag);
		}
		
		if (protocolMapper != null) 
		{
			ItemSubtype type = ItemSubtype.getRelatedSubtype(dc, extract, Item.PROTOCOL, SystemItems.getId(Protocol.EXTRACTION));
			if (defaultPrototolFunction != null) defaultPrototolFunction.setCurrentSubtype(type);
			String nameOrId = protocolMapper.getValue(data);
			Protocol protocol = findProtocol(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (nameOrId == null || protocol != null) creationEvent.setProtocol(protocol);
		}
		if (bioPlateMapper != null && bioWellRowMapper != null && bioWellColMapper != null)
		{
			String plateNameOrId = bioPlateMapper.getValue(data);
			String row = bioWellRowMapper.getValue(data);
			String column = bioWellColMapper.getValue(data);
			BioPlate plate = findBioPlate(dc, FallbackIdMethod.NAME_OR_ID, plateNameOrId);
			BioWell well = findBioWell(dc, plate, row, column);
			if (plateNameOrId == null || row == null || column == null || well != null)
			{
				extract.setBioWell(well);
			}
		}

		
		if (parentMapper != null)
		{
			// Try to find a parent type
			Item parentType = extract.getParentType();
			// Is it specified in the imported file?
			if (parentTypeMapper != null)
			{
				String type = parentTypeMapper.getValue(data);
				if (type != null) parentType = Item.valueOf(type);
			}
			
			// See if we can find a related subtype
			ItemSubtype type = null;
			if (parentType == null)
			{
				// If we don't know the parent type, check for both biosurce and sample types
				type = ItemSubtype.getRelatedSubtype(dc, extract, Item.SAMPLE, Item.BIOSOURCE);
				if (type != null) parentType = type.getMainItemType();
			}
			else
			{
				type = ItemSubtype.getRelatedSubtype(dc, extract, parentType);
			}

			// Try to find either a sample or a biosource
			String nameOrId = parentMapper.getValue(data);
			BioMaterial parent = null;
			if (parentType == Item.EXTRACT)
			{
				parent = findExtract(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			}
			else
			{
				parent = findSample(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			}
			if (nameOrId == null || parent != null) 
			{
				BioMaterialEventSource evtSrc = extract.getCreationEvent().setSource(parent);
				if (evtSrc != null) 
				{
					Float usedQuantity = usedQuantityMapper == null ? null : usedQuantityMapper.getFloat(data);
					evtSrc.setUsedQuantity(usedQuantity);
				}
			}
		}
	}
	
	/**
		If the sample is pooled, add the parent sample.
	*/
	@Override
	protected void updateMultiLineItem(DbControl dc, Extract extract, 
		FlatFileParser.Data data, int multiLineNum)
	{
		if (parentMapper != null && extract.getParentType() == Item.EXTRACT)
		{
			String nameOrId = parentMapper.getValue(data);
			ItemSubtype type = ItemSubtype.getRelatedSubtype(dc, extract, Item.EXTRACT, 0);
			Extract parent = findExtract(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (parent != null)
			{
				Float usedQuantity = usedQuantityMapper == null ?
						null : usedQuantityMapper.getFloat(data);
				extract.getCreationEvent().addSource(parent).setUsedQuantity(usedQuantity);
				// Use same tag as parent if not yet set and not mapped in the file
				if (tagMapper == null && parent.getTag() != null && extract.getTag() == null)
				{
					extract.setTag(parent.getTag());
				}
			}
		}
	}
	// ------------------------------------
}
