/**
	$Id: RawBioAssayImporter.java 5899 2011-12-06 08:17:37Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.DefaultItemFunction;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing raw bioassay items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2011-12-06 09:17:37 +0100 (Tue, 06 Dec 2011) $
*/
public class RawBioAssayImporter
	extends AbstractItemImporter<RawBioAssay>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.RAWBIOASSAY, GuiContext.Type.LIST));

	protected static final PluginParameter<String> platformColumnMapping = new PluginParameter<String>(
			"platformColumnMapping",
			"Platform",
			"Mapping that picks the name or ID of the experimental platform from the data columns. " +
			"The plug-in will first try to find a platform with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Platform\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> variantColumnMapping = new PluginParameter<String>(
			"variantColumnMapping",
			"Variant",
			"Mapping that picks the name or ID of the experimental platform variant from the data columns. " +
			"The plug-in will first try to find a variant with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Variant\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> rawDataTypeColumnMapping = new PluginParameter<String>(
			"rawDataTypeColumnMapping",
			"Raw data type",
			"Mapping that picks the name of the raw data type from the data columns. " +
			"The plug-in can only find a raw data type by it's internal ID. " +
			"Example: \\Raw data type\\",
			optionalColumnMapping
			);

	
	protected static final PluginParameter<String> softwareColumnMapping = new PluginParameter<String>(
			"softwareColumnMapping",
			"Software",
			"Mapping that picks the name or ID of the feature extraction software from the data columns. " +
			"The plug-in will first try to find a software with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Software\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> bioAssayColumnMapping = new PluginParameter<String>(
			"bioAssayColumnMapping",
			"Parent bioassay",
			"Mapping that picks the name or ID of the parent derived bioassay from the data columns. " +
			"The plug-in will first try to find a bioassay with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Parent bioassay\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> extractColumnMapping = new PluginParameter<String>(
			"extractColumnMapping",
			"Parent extract",
			"Mapping that picks the name or ID of the parent extract from the data columns. " +
			"The plug-in will first try to find an extract with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Parent extract\\\n\n" +
			"NOTE! This column may not be needed when creating new items if the parent bioassay " +
			"already contains a link to an extract (which will be copied)",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> arrayDesignColumnMapping = new PluginParameter<String>(
			"arrayDesignColumnMapping",
			"Array design",
			"Mapping that picks the name or ID of the array design from the data columns. " +
			"The plug-in will first try to find an array design with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Array design\\\n\n" +
			"NOTE! This column may not be needed when creating new items if the parent bioassay " +
			"already contains a link to an array design (which will be copied)",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> fileTypeColumnMapping = 
		new PluginParameter<String>(
				"fileTypeColumnMapping",
				"Data file type",
				"Mapping that picks the name or ID of a data file type from the data columns. " +
				"The plug-in will first try to find a file type with the given name. If none is " +
				"found and the value is numeric it will try to load by internal ID. " +
				"Example: \\File type\\",
				optionalColumnMapping
		);
	
	protected static final PluginParameter<String> fileColumnMapping = 
		new PluginParameter<String>(
				"fileColumnMapping",
				"File",
				"Mapping that picks the path of a data file from the columns. " +
				"The path can be an absolute path (starting with /) or relative " +
				"to the 'Data directory' parameter." +
				"Example: \\File\\",
				optionalColumnMapping
		);

	
	public RawBioAssayImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper protocolMapper;
	private DefaultItemFunction<Protocol> defaultProtocolFunction;
	private Mapper softwareMapper;
	private DefaultItemFunction<Software> defaultSoftwareFunction;
	private Mapper bioAssayMapper;
	private Mapper extractMapper;
	private Mapper arrayDesignMapper;
	private DefaultItemFunction<ArrayDesign> defaultArrayDesignFunction;
	private Mapper platformMapper;
	private Mapper variantMapper;
	private Mapper rawDataTypeMapper;
	private Mapper fileTypeMapper;
	private Mapper fileMapper;
	private String dataDirectory;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Add parameter for "Data directory" directory where data files may be located. 
	*/
	@Override
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(new PluginParameter<String>(
				"dataDirectory",
				"Data directory",
				"A base directory were data files are located. Relative file paths are " +
				"resolved against this directory. If no directory is given, only absolute " +
				"paths (=paths starting with /) can be used.",
				new PathParameterType(Path.Type.DIRECTORY))
			);
	}

	
	/**
		Adds column mappings for name, externalId, description,
		protocol, creation date, original quantity, pooled, parent
		and used quantity from parent.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(platformColumnMapping);
		parameters.add(variantColumnMapping);
		parameters.add(rawDataTypeColumnMapping);
		parameters.add(bioAssayColumnMapping);
		parameters.add(extractColumnMapping);
		parameters.add(arrayDesignColumnMapping);
		parameters.add(protocolColumnMapping);
		parameters.add(softwareColumnMapping);
		parameters.add(fileTypeColumnMapping);
		parameters.add(fileColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Calls {@link RawBioAssay#getQuery()}.
	*/
	@Override
	protected ItemQuery<RawBioAssay> createItemQuery()
	{
		return RawBioAssay.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? RawBioAssay.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? RawBioAssay.MAX_DESCRIPTION_LENGTH : null, null);
		defaultProtocolFunction = getDefaultItemFunction(Item.PROTOCOL);
		protocolMapper = getMapper(ffp, (String)job.getValue("protocolColumnMapping"), null, null, defaultProtocolFunction);
		defaultSoftwareFunction = getDefaultItemFunction(Item.SOFTWARE);
		softwareMapper = getMapper(ffp, (String)job.getValue("softwareColumnMapping"), null, null, defaultSoftwareFunction);
		bioAssayMapper = getMapper(ffp, (String)job.getValue("bioAssayColumnMapping"), null, null);
		extractMapper = getMapper(ffp, (String)job.getValue("extractColumnMapping"), null, null);
		defaultArrayDesignFunction = getDefaultItemFunction(Item.ARRAYDESIGN);
		arrayDesignMapper = getMapper(ffp, (String)job.getValue("arrayDesignColumnMapping"), null, null, defaultArrayDesignFunction);
		platformMapper = getMapper(ffp, (String)job.getValue("platformColumnMapping"), null, null);
		variantMapper = getMapper(ffp, (String)job.getValue("variantColumnMapping"), null, null);
		rawDataTypeMapper = getMapper(ffp, (String)job.getValue("rawDataTypeColumnMapping"), null, null);
		fileTypeMapper = getMapper(ffp, (String)job.getValue("fileTypeColumnMapping"), null, null);
		fileMapper = getMapper(ffp, (String)job.getValue("fileColumnMapping"), null, null);
		dataDirectory = (String)job.getValue("dataDirectory");
		if (dataDirectory != null && !dataDirectory.endsWith("/"))
		{
			dataDirectory += "/";
		}
	}
	
	@Override
	protected RawBioAssay createItem(DbControl dc, FlatFileParser.Data data)
	{
		Platform platform = null;
		PlatformVariant variant = null;
		String rawDataTypeId = null;
		
		if (platformMapper != null) 
		{
			platform = findPlatform(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, platformMapper.getValue(data));
		}
		else 
		{
			platform = getProjectDefaultPlatform(dc);
		}
		if (variantMapper != null) 
		{
			variant = findVariant(dc, platform, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, variantMapper.getValue(data));
		}
		else
		{
			variant = getProjectDefaultVariant(dc);
		}
		if (rawDataTypeMapper != null) rawDataTypeId = rawDataTypeMapper.getValue(data);
		RawDataType rawDataType = findRawDataType(dc, platform, variant, rawDataTypeId);
		
		RawBioAssay rba = null;
		if (variant == null && rawDataType != null) variant = rawDataType.getVariant(dc);
		if (variant != null)
		{
			rba = RawBioAssay.getNew(dc, variant, rawDataType);
		}
		else
		{
			if (platform == null && rawDataType != null) platform = rawDataType.getPlatform(dc);
			if (platform == null) platform = Platform.getByExternalId(dc, Platform.GENERIC);
			rba = RawBioAssay.getNew(dc, platform, rawDataType);
		}
		updateItem(dc, rba, data);
		return rba;
	}
	
	@Override
	protected void updateItem(DbControl dc, RawBioAssay rba, FlatFileParser.Data data)
	{
		if (nameMapper != null) rba.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) rba.setDescription(descriptionMapper.getValue(data));
		ArrayDesign designFromBioAssay = null;
		Extract extractFromBioAssay = null;
		if (bioAssayMapper != null) 
		{
			String nameOrId = bioAssayMapper.getValue(data);
			DerivedBioAssay bioAssay = findDerivedBioAssay(dc, FallbackIdMethod.NAME_OR_ID, nameOrId, null);
			if (nameOrId == null || bioAssay != null) rba.setParentBioAssay(bioAssay);
			if (bioAssay != null) 
			{
				try
				{
					designFromBioAssay = bioAssay.getArrayDesign();
				}
				catch (PermissionDeniedException ex)
				{}
				try
				{
					extractFromBioAssay = bioAssay.getExtract();
				}
				catch (PermissionDeniedException ex)
				{}
			}
		}
		
		if (extractMapper != null)
		{
			String nameOrId = extractMapper.getValue(data);
			Extract extract = nameOrId == null ? 
					extractFromBioAssay : findExtract(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
			if (nameOrId == null || extract != null) rba.setParentExtract(extract);
		}
		else if (extractFromBioAssay != null)
		{
			rba.setParentExtract(extractFromBioAssay);
		}
		
		if (!rba.hasData())
		{
			if (arrayDesignMapper != null) 
			{
				String nameOrId = arrayDesignMapper.getValue(data);
				ArrayDesign design = nameOrId == null ? 
						designFromBioAssay : findArrayDesign(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
				if (nameOrId == null || design != null) rba.setArrayDesign(design);
			}
			else if (designFromBioAssay != null)
			{
				rba.setArrayDesign(designFromBioAssay);
			}
			// Only do this when updating since the same thing is done in 'createItem'
			if (rba.isInDatabase())
			{
				Platform platform = null;
				PlatformVariant variant = null;
				String rawDataTypeId = null;
				if (platformMapper != null) 
				{
					platform = findPlatform(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, platformMapper.getValue(data));
				}
				else
				{
					platform = rba.getPlatform();
				}
				if (variantMapper != null) 
				{
					variant = findVariant(dc, platform, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, variantMapper.getValue(data));
				}
				else
				{
					variant = rba.getVariant();
				}
				if (rawDataTypeMapper != null) 
				{
					rawDataTypeId = rawDataTypeMapper.getValue(data);
				}
				else
				{
					rawDataTypeId = rba.getRawDataType().getId();
				}
				RawDataType rawDataType = findRawDataType(dc, platform, variant, rawDataTypeId);
				if (variant != null)
				{
					rba.setVariant(variant, rawDataType);
				}
				else if (platform != null)
				{
					rba.setPlatform(platform, rawDataType);
				}
			}
		}
		if (protocolMapper != null) 
		{
			ItemSubtype type = ItemSubtype.getById(dc, SystemItems.getId(Protocol.FEATURE_EXTRACTION));
			defaultProtocolFunction.setCurrentSubtype(type);
			String nameOrId = protocolMapper.getValue(data);
			Protocol protocol = findProtocol(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (nameOrId == null || protocol != null) rba.setProtocol(protocol);
		}
		if (softwareMapper != null)
		{
			ItemSubtype type = ItemSubtype.getById(dc, SystemItems.getId(Software.FEATURE_EXTRACTION));
			defaultSoftwareFunction.setCurrentSubtype(type);
			String nameOrId = softwareMapper.getValue(data);
			Software software = findSoftware(dc, FallbackIdMethod.NAME_OR_ID, nameOrId, type);
			if (nameOrId == null || software != null) rba.setSoftware(software);
		}
		updateMultiLineItem(dc, rba, data, 0);
	}
	/**
		Add data files
	*/
	@Override
	protected void updateMultiLineItem(DbControl dc, RawBioAssay rba, 
		FlatFileParser.Data data, int multiLineNum)
	{
		if (fileTypeMapper != null && fileMapper != null)
		{
			String fileTypeNameOrId = fileTypeMapper.getValue(data);
			String filePath = fileMapper.getValue(data);
			DataFileType fileType = findDataFileType(dc, 
					FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, fileTypeNameOrId);
			if (fileType != null)
			{
				File file = null;
				if (filePath != null)
				{
					if (!filePath.startsWith("/") && dataDirectory != null) 
					{
						filePath = dataDirectory + filePath;
					}
					file = File.getByPath(dc, new Path(filePath, Path.Type.FILE), false);
				}
				setOrAddFile(rba, fileType, file);
			}
		}
	}
	
	/**
		Validate data files if they have been set.
	*/
	@Override
	protected void doneWithItem(DbControl dc, RawBioAssay rba, int numMultiLines)
	{
		if (rba.hasFileSet()) rba.getFileSet().validate(dc);
	}
	// ------------------------------------
}
