/**
 $Id: ExternalProgramExecutor.java 5689 2011-08-11 09:09:06Z nicklas $

 Copyright (C) 2010 Martin Svensson
 
 This file is part of BASE - BioArray Software Environment.
 Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins.executor;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterException;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.Job.ExecutionTime;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.NamespaceParameterValuesWrapper;
import net.sf.basedb.core.plugin.NamespacePluginParameter;
import net.sf.basedb.core.plugin.NamespaceRequestWrapper;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.ChainedProgressReporter;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.FileCopyRunnable;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.formatter.DateFormatter;

import java.io.ByteArrayOutputStream;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 	 
  	@author martin, nicklas
  	@since 2.15
  	@base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
*/
public class ExternalProgramExecutor 
	extends AbstractAnalysisPlugin 
	implements InteractivePlugin, SignalTarget
{
	
	/**
		Parameter name for the path to the external program. 
		This parameter is a String parameter stored in the 
		plug-in configuration.
	*/
	public static final String PARAMETER_EXTERNAL_PATH = "externalProgramPath";
	
	/**
		Parameter name for the command line options to the external program.
		This parameter is a String parameter stored in the plug-in
		configuration.
	 */
	public static final String PARAMETER_EXTERNAL_CMDLINE = "externalProgramCmdLine";
	
	/**
		Parameter name for the xml holding the parameter definitions to
		the external program. This parameter is an XML String stored 
		in the plug-in configuration.
	*/
	public static final String PARAMETER_EXTERNAL_PARAMETERS = "externalParameters";
	
	/**
		Parameter name for the class that has been selected as the importer.
		This parameter is a String parameter stored in the plug-in
		configuration.
	*/
	public static final String PARAMETER_IMPORTER_CLASS = "importerClass";

	/**
		Parameter name for the class that has been selected as the exporter.
		This parameter is a String parameter stored in the plug-in
		configuration.
	*/
	public static final String PARAMETER_EXPORTER_CLASS = "exporterClass";

	/**
		Parameter name for debug mode flag.
	*/
	public static final String PARAMETER_DEBUG_MODE = "debugMode";
	
	/**
		Parameter name for the required raw data type.
		This parameter is a String parameter stored in the plug-in
		configuration.
	*/
	public static final String PARAMETER_RAW_DATA_TYPE = "rawDataType";
	
	/**
		Parameter name for the required number of channels.
		This parameter is an Integer parameter stored in the plug-in
		configuration.
	*/
	public static final String PARAMETER_NUM_CHANNELS = "numChannels";
	
	/**
		Parameter name for the required intensity transform of the source bioassay set.
		This parameter is a String parameter (corresponding to
		{@link IntensityTransform#name()}) stored in the plug-in
		configuration.
	*/
	public static final String PARAMETER_SOURCE_TRANSFORM = "sourceTransform";
	
	/**
		Parameter name for the source bioassay set.
		This parameter is a BioAssaySet parameter stored in the job
		configuration.
	*/
	public static final String PARAMETER_SOURCE_BIOASSSAYSET = SOURCE_BIOASSAYSET;
		
	/**
		The namespace for the exporter plug-in parameters.
	*/
	public static final String NAMESPACE_EXPORTER = "exporter";

	/**
		The namespace for the importer plug-in parameters.
	*/
	public static final String NAMESPACE_IMPORTER = "importer";

	/**
		The namespace for the external program parameters.
	*/
	public static final String NAMESPACE_EXTERNAL = "external";

	// Command to create parameters for external plugin
	private static final String COMMAND_CONFIGURE_EXTERNAL_PARAMETERS = "CREATE_EXTERNAL_PARAMETERS";
	// Command to configure job for the import-plugin and the export-plugin
	
	private ThreadSignalHandler signalHandler;
	
	/* Configuration for the ExternalProgramExecutor.
	  	Holds which external plug-in to use and
	  	which export and import plugin that should be used.
	 */
	private RequestInformation configurePlugin = null;
	
	// Configuration for which parameters the external plugin needs.
	private RequestInformation configureExternalParameters = null;
	
	/* 	Configuration for a job of the ExternalProgramExcecutor. 
		Holds parameters needed by the external plug-in and 
		error handling parameters.		
	*/ 
	private RequestInformation configureJob = null;
	
	
	// Configuration for the export-plugin and import-plugin.
	private RequestInformation configureExportImportPluginParameters = null;
	
	/**
		Temporary working directory at runtime.
	*/
	private java.io.File workDir;
	
	/**
		Files that are known to exists in the working directory.
		Unknown files can be assumed to have been created by the 
		external program.
	*/
	private Set<java.io.File> knownFiles;
	
	/**
	
	 */
	public ExternalProgramExecutor()
	{}
	
	/*
		From the Plugin interface
		-------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return true;
	}

	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);

		DbControl dc = null;
		boolean debugMode = Boolean.TRUE.equals(job.getValue(PARAMETER_DEBUG_MODE));
		try
		{
			ChainedProgressReporter chainedProgress = 
				progress == null ? null : new ChainedProgressReporter(progress);
			if (progress != null) progress.display(0, "Preparing...");
			
			// Create a working directory
			workDir = createTemporaryWorkingDirectory();
			
			// Setup wrapper for the request and response
			ResponseWrapper responseWrapper = new ResponseWrapper();
			RequestWrapper requestWrapper = new RequestWrapper(IOSupport.COMMAND_EXECUTE_IO_JOB, request);
			
			// Run the exporter plug-in
			if (chainedProgress != null) chainedProgress.setRange(5, 40);
			ThreadSignalHandler.checkInterrupted();
			runExporter(requestWrapper, responseWrapper, chainedProgress);
			if (responseWrapper.hasError())
			{
				String message = responseWrapper.getMessage("There was an error when exporting data");
				response.setError(message, responseWrapper.getErrors());
				return;
			}
			inspectWorkingDirectory();
			String externalProgram = (String)configuration.getValue(PARAMETER_EXTERNAL_PATH);
			String cmdLine = (String)configuration.getValue(PARAMETER_EXTERNAL_CMDLINE);
			if (progress != null) progress.display(50, "Executing external program: " + externalProgram);
			ThreadSignalHandler.checkInterrupted();
			runExternalProgram(externalProgram, cmdLine, workDir, "stdin.txt", "stdout.txt", progress);
			
			// Run the importer plug-in
			if (chainedProgress != null) chainedProgress.setRange(60, 95);
			responseWrapper = new ResponseWrapper();
			ThreadSignalHandler.checkInterrupted();
			runImporter(requestWrapper, responseWrapper, chainedProgress);
			if (responseWrapper.hasError())
			{
				String message = responseWrapper.getMessage("There was an error when importing data");
				response.setError(message, responseWrapper.getErrors());
				return;
			}
			
			String msg = responseWrapper.getMessage("External program ended successfully");
			if (progress != null) progress.display(100, msg);
			response.setDone(msg);
		}
		catch (Throwable t)
		{
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			if (dc != null) dc.close();
			if (!debugMode && workDir != null) FileUtil.deleteTempDirectory(workDir);
		}
		
	}
	// --------------------------
	
	/*
		From the SignalTarget interface
		-------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// ---------------------------------

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String msg = super.isInContext(context, item);
		if (msg != null) return msg;
		
		BioAssaySet bas = (BioAssaySet)item;
		RawDataType actualRdt = bas.getRawDataType();
		RawDataType requiredRdt = RawDataTypes.getRawDataType((String)configuration.getValue(PARAMETER_RAW_DATA_TYPE));
		Integer requiredChannels = (Integer)configuration.getValue(PARAMETER_NUM_CHANNELS);
		if (requiredRdt != null)
		{
			if (!requiredRdt.equals(actualRdt))
			{
				msg = "Can only be used with " + requiredRdt + 
					" data, not with " + actualRdt.getName();
			}
		}
		else if (requiredChannels != null)
		{
			if (requiredChannels != actualRdt.getChannels())
			{
				msg = "Can only be used with " + requiredChannels + 
					"-channel data, not with " + actualRdt.getChannels() + "-channel";
			}
		}
		if (msg != null) return msg;
		
		String requiredIntensityTransform = (String)configuration.getValue(PARAMETER_SOURCE_TRANSFORM);
		if (requiredIntensityTransform != null)
		{
			if (!requiredIntensityTransform.equals(bas.getIntensityTransform().name()))
			{
				msg = "Can only be used with " + requiredIntensityTransform + 
					"-transformed data, not with " + bas.getIntensityTransform();
			}
		}
		return msg;
	}
	
	@Override
	public RequestInformation getRequestInformation(GuiContext context,	String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (Request.COMMAND_CONFIGURE_PLUGIN.equals(command))
		{
			// Config 1. select external binary and export+import plug-in
			requestInformation = getConfigurePluginParameters();		
		}
		else if (IOSupport.COMMAND_CONFIGURE_IO_PLUGIN.equals(command))
		{
			// Config 2. configure exporter and importer plug-ins
			requestInformation = getConfigureExportImportPluginParameters();
		}
		else if (ExternalProgramExecutor.COMMAND_CONFIGURE_EXTERNAL_PARAMETERS.equals(command))
		{
			// Config 3. parameters to the external binary
			requestInformation = getConfigureExternalParameters();
		}
		else if (Request.COMMAND_CONFIGURE_JOB.equals(command)	)
		{
			// Job 1. 
			requestInformation = getConfigureJobParameters();
		}
		else 
		{
			throw new BaseException("Unknown command: "+command);
		}
		return requestInformation;
	}

	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();		
		if (Request.COMMAND_CONFIGURE_PLUGIN.equals(command))
		{
			// Store the this plugin's configuration
			RequestInformation ri = getConfigurePluginParameters();
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;
			}
			storeValue(configuration, request, ri.getParameter(PARAMETER_EXTERNAL_PATH));
			storeValue(configuration, request, ri.getParameter(PARAMETER_EXTERNAL_CMDLINE));
			storeValue(configuration, request, ri.getParameter(PARAMETER_EXPORTER_CLASS));
			storeValue(configuration, request, ri.getParameter(PARAMETER_IMPORTER_CLASS));
			storeValue(configuration, request, ri.getParameter(PARAMETER_DEBUG_MODE));
			storeValue(configuration, request, ri.getParameter(PARAMETER_RAW_DATA_TYPE));
			storeValue(configuration, request, ri.getParameter(PARAMETER_NUM_CHANNELS));
			storeValue(configuration, request, ri.getParameter(PARAMETER_SOURCE_TRANSFORM));
			
			response.setContinue(IOSupport.COMMAND_CONFIGURE_IO_PLUGIN);
		}
		else if (IOSupport.COMMAND_CONFIGURE_IO_PLUGIN.equals(command))
		{
			// Store the plugin-configuration for the exporter and importer. 
			DbControl dc = sc.newDbControl();
			ResponseWrapper responseWrapper = new ResponseWrapper();
			try
			{
				IOSupport exporter = getExporter(dc, false, false);
				exporter.configure(null, new NamespaceRequestWrapper(request, NAMESPACE_EXPORTER), responseWrapper);
				IOSupport importer = getImporter(dc, false, false);
				importer.configure(null, new NamespaceRequestWrapper(request, NAMESPACE_IMPORTER), responseWrapper);
			}
			finally
			{
				if (dc != null) dc.close();
			}
			
			if (responseWrapper.hasError())
			{
				String msg = responseWrapper.getMessage("There was an error in the selected exporter or importer");
				response.setError(msg, responseWrapper.getErrors());
			}
			else
			{
				response.setContinue(ExternalProgramExecutor.COMMAND_CONFIGURE_EXTERNAL_PARAMETERS);
			}
		}
		else if (ExternalProgramExecutor.COMMAND_CONFIGURE_EXTERNAL_PARAMETERS.equals(command))
		{
			// Store the created parameters needed by the external plugin.			
			RequestInformation ri = getConfigureExternalParameters();
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;					
			}
			// Validates the parameters
			ParameterDefinition.parseXml((String)request.getParameterValue(PARAMETER_EXTERNAL_PARAMETERS), true);
			
			storeValue(configuration, request, ri.getParameter(PARAMETER_EXTERNAL_PARAMETERS));
			response.setDone("Configuration is now complete.");
		}		
		else if (Request.COMMAND_CONFIGURE_JOB.equals(command))
		{
			RequestInformation ri = getConfigureJobParameters();
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);				
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;
			}

			// Store the job-configuration for the exporter and importer. 
			DbControl dc = sc.newDbControl();
			ResponseWrapper responseWrapper = new ResponseWrapper();
			RequestWrapper requestWrapper = new RequestWrapper(IOSupport.COMMAND_CONFIGURE_IO_JOB, request);
			try
			{
				IOSupport exporter = getExporter(dc, false, false);
				exporter.configure(null, new NamespaceRequestWrapper(requestWrapper, NAMESPACE_EXPORTER), responseWrapper);
				IOSupport importer = getImporter(dc, false, false);
				importer.configure(null, new NamespaceRequestWrapper(requestWrapper, NAMESPACE_IMPORTER), responseWrapper);
			}
			finally
			{
				if (dc != null) dc.close();
			}
			if (responseWrapper.hasError())
			{
				String msg = responseWrapper.getMessage("There was an error in the selected exporter or importer");
				response.setError(msg, responseWrapper.getErrors());
				return;
			}
			
			// Sava program parameters
			for (PluginParameter<?> pp : ri.getParameters())
			{
				if (pp.getName().startsWith(NAMESPACE_EXTERNAL + "."))
				{
					storeValues(job, request, pp);
				}
			}

			// And last... the source bioassay set
			storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
			storeValue(job, request, ri.getParameter(PARAMETER_DEBUG_MODE));

			BioAssaySet bioAssaySet = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
			PluginConfiguration config = configuration.getPluginConfiguration();
			response.setSuggestedJobName("Run external program '" + config.getName() +
				"' on bioassay set '" + bioAssaySet.getName() + "'");
			response.setDone("Job configuration is complete");
		}
		else 
		{
			throw new BaseException("Unknown command: "+command);
		}		
	}
	// -----------------------------------------------
	
	/**
		Get the bioassay set that is used as the source for the job.
		This method should only be called at runtime or job configuration 
		time.
		@return The source bioassay set
	*/
	@Override
	public BioAssaySet getCurrentBioAssaySet(DbControl dc)
	{
		return super.getCurrentBioAssaySet(dc);
	}

	/**
		Get the current working directory were data files for the
		external program are stored. This method should only be called
		at runtime.
	*/
	public java.io.File getWorkingDirectory()
	{
		return workDir;
	}

	/**
		Check the working directory for new files. Each call to
		this method will mark the existing files as "known" files.
		Subsequent calls will only find files that has been added
		to the working directory since that last call. The
		executor calls this method after the exporter plug-in has
		been executed just before the external program is about
		to be started. This allows the importer plug-in to call
		this method to find out which files that were created by
		the external program.
		@return A list with new files that has been added to the
			working directory since the last call (the list is empty
			if no new files exists)
		@since 2.16
	*/
	public List<java.io.File> inspectWorkingDirectory()
	{
		if (knownFiles == null) knownFiles = new HashSet<java.io.File>();
		java.io.File workDir = getWorkingDirectory();
		List<java.io.File> newFiles = FileUtil.findFiles(workDir, new FileFilter()
			{
				@Override
				public boolean accept(java.io.File file)
				{
					return !knownFiles.contains(file);
				}
			});
		knownFiles.addAll(newFiles);
		return newFiles;
	}
	
	/**
		Get the current program parameter definitions. This method should only
		be called at runtime or job configuration time.
	*/
	public List<ParameterDefinition> getParameterDefinitions(boolean validate)
	{
		String xml = (String)configuration.getValue(PARAMETER_EXTERNAL_PARAMETERS);
		return ParameterDefinition.parseXml(xml, validate);
	}
	
	/**
		Get all program parameters and their values. This method should only
		be called at runtime.
	*/
	public List<ProgramParameter> getProgramParameters()
	{
		List<ParameterDefinition> defs = getParameterDefinitions(true);
		List<ProgramParameter> parameters = new ArrayList<ProgramParameter>(defs.size());
		for (ParameterDefinition pdef : defs)
		{
			ProgramParameter pp = new ProgramParameter();
			pp.setParameterDefinition(pdef);
			pp.setValues(job.getValues(NAMESPACE_EXTERNAL + "." + pdef.getName()));
			parameters.add(pp);
		}
		return parameters;
	}
	
	/**
		Create a new file in the working directory. If a file with the given 
		name already exists the existing file is used.
		
		@param name The name of the file to create
		@return A new file object pointing to the file
		@throws IOException 
	*/
	public java.io.File createFile(String name)
		throws IOException
	{
		java.io.File f = new java.io.File(getWorkingDirectory(), name);
		f.createNewFile();
		return f;
	}
	
	/**
		Create a temporary working directory in the local file
		system. The exporter should export data to this directory.
		The external program will use the data and store it's output
		in the same directory. It is recommended that the directory
		is empty and not used by any other process at the same time.
		Existing files may be overwritten without notice.
		@return A file object representing a directory
	*/
	protected java.io.File createTemporaryWorkingDirectory()
		throws IOException
	{
		String now = new DateFormatter("-yyyyMMdd-HHmmss-").format(new Date());
		return FileUtil.createTempDirectory("executor"+now, null, null);
	}
	
	/**
		Execute the exporter plug-in using the supplied parameters.
	*/
	protected void runExporter(Request request, Response response, ProgressReporter progress)
	{
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			IOSupport exporter = getExporter(dc, true, true);
			dc.close();
			exporter.run(request, response, progress);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/**
		Execute the importer plug-in using the supplied parameters.
	*/
	protected void runImporter(Request request, Response response, ProgressReporter progress)
	{
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			IOSupport importer = getImporter(dc, true, true);
			dc.close();
			importer.run(request, response, progress);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	/**
		Run the external program. The program and active working directory
		is given by the parameters. No command line options can be sent to
		the program. It is possible to name one (existing) file that should
		be piped to standard input to the plug-in and one (non-existing)
		file that should receive output from the program. Anything the program 
		writes to the standard error stream should cause an error and be
		reported back to the caller via en exception.
		
		@param program The path to the external program
		@param workDir The active working directory
		@param stdin If given, the named file within the working directory
			should be piped to the standard input of the external program. 
		@param stdout If given, output from the external program should be
			piped the named file. If not given any output from the program
			needs to be read and discarded
	*/
	protected void runExternalProgram(String program, String cmdLine, java.io.File workDir,
		String stdin, String stdout, ProgressReporter progress)
	{
		// Build the external process
		ProcessBuilder builder = new ProcessBuilder();
		List<String> cmds = new ArrayList<String>();
		cmds.add(program);
		if (cmdLine != null) 
		{
			// Split on whitespace but not inside quotes. Quotes are removed after the split
			String[] options = StringUtil.tokenize(cmdLine, "\\s+", true, true);
			if (options != null)
			{
				cmds.addAll(Arrays.asList(options));
			}
		}
		builder.command(cmds);
		builder.directory(workDir);
		Process process = null;
		
		// This thread group is for thread readin/writing stream to program
		ThreadGroup fileCopyGroup = new ThreadGroup("File copy");		
		RuntimeException error = null;
		try
		{
			// Start the external process
			process = builder.start();

			// If stdin is given it is sent to the program
			java.io.File stdinFile = stdin == null ? null : new java.io.File(workDir, stdin);
			if (stdinFile != null && stdinFile.exists()) 
			{
				InputStream in = FileUtil.getInputStream(stdinFile);
				FileCopyRunnable inCopy = new FileCopyRunnable(in, process.getOutputStream());
				new Thread(fileCopyGroup, inCopy, "stdin").start();
			}
			
			// Always read output from the program and save to stdout if given
			OutputStream out = null;
			if (stdout != null) out = FileUtil.getOutputStream(new java.io.File(workDir, stdout));
			FileCopyRunnable outCopy = new FileCopyRunnable(process.getInputStream(), out);
			new Thread(fileCopyGroup, outCopy, "stdout").start();
		
			// Always read stderr
			ByteArrayOutputStream err = new ByteArrayOutputStream();
			FileCopyRunnable errCopy = new FileCopyRunnable(process.getErrorStream(), err);
			Thread errThread = new Thread(fileCopyGroup, errCopy, "stderr");
			errThread.start();
			
			// Wait for the process to complete
			int exitValue = process.waitFor();
			if (exitValue != 0)
			{
				// exitValue != 0 indicates an error... check the error stream for messages
				long time = System.currentTimeMillis();
				long waited = 0; // How long we have waited in seconds
				while (errThread.isAlive() && waited < 60)
				{
					if (progress != null) 
					{
						progress.display(55, "Exit code: " + exitValue + 
								"; waiting for error message...");
					}
					Thread.sleep(1000);
					waited = (System.currentTimeMillis() - time) / 1000;
				}
				err.flush();
				String msg = err.toString();
				if (msg == null || msg.length() == 0) 
				{
					msg = "The program did not write any error message to stderr.";
				}
				throw new RuntimeException("[" + exitValue + "] " + msg);
			}
		}
		catch (InterruptedException ex)
		{
			// Aborted by user
			error = new SignalException("Aborted by user", ex);
		}
		catch (IOException ex)
		{
			error = new RuntimeException(ex);
		}
		catch (RuntimeException ex)
		{
			error = ex;
		}
		finally
		{
			fileCopyGroup.interrupt();
		}
		if (error != null)
		{
			if (process != null) process.destroy();
			throw error;
		}
	}
	
	private IOSupport exporter;
	private IOSupport importer;
	
	/**
		Get an instance of the exporter plug-in.
		@param dc A DbControl to use for database access
		@param forceNew If TRUE a new instance must always be created, otherwise
			an already existing instance may be re-used
		@param runtime TRUE if the importer is created at runtime, which
			means that job agent configuration must be taking into account
		@return An exporter instance
	*/
	protected IOSupport getExporter(DbControl dc, boolean forceNew, boolean runtime)
	{
		if (exporter == null || forceNew)
		{
			String exporterClass = (String)configuration.getValue(PARAMETER_EXPORTER_CLASS);
			PluginDefinition plugin = PluginDefinition.getByClassName(dc, exporterClass);
			exporter = createInstance(dc, plugin, NAMESPACE_EXPORTER, runtime);
		}
		return exporter;
	}
	
	/**
		Get an instance of the importer plug-in.
		@param dc A DbControl to use for database access
		@param forceNew If TRUE a new instance must always be created, otherwise
			an already existing instance may be re-used
		@param runtime TRUE if the importer is created at runtime, which
			means that job agent configuration must be taking into account
		@return An importer instance
	*/
	protected IOSupport getImporter(DbControl dc, boolean forceNew, boolean runtime)
	{
		if (importer == null || forceNew)
		{
			String importerClass = (String)configuration.getValue(PARAMETER_IMPORTER_CLASS);
			PluginDefinition plugin = PluginDefinition.getByClassName(dc, importerClass);
			importer = createInstance(dc, plugin, NAMESPACE_IMPORTER, runtime);
		}
		return importer;
	}

	/**
		Create and initialize a plug-in instance for the given plug-in 
		definition. Job and configuration parameters should be wrapped
		in the given namespace.
		@param plugin The plug-in definition we should instantiate
		@param namespace The namespace, or null to not wrap the parameters
			in a namespace
		@param runtime TRUE if the importer is created at runtime, which
			means that job agent configuration must be taking into account
		@return An initialized plug-in instance
	*/
	protected IOSupport createInstance(DbControl dc, PluginDefinition plugin, String namespace, boolean runtime)
	{
		ParameterValues configWrapper = configuration;
		if (configuration != null && namespace != null)
		{
			configWrapper = new NamespaceParameterValuesWrapper(configuration, namespace);
		}
		ParameterValues jobWrapper = job;
		if (job != null && namespace != null)
		{
			jobWrapper = new NamespaceParameterValuesWrapper(job, namespace);
		}
		Job currentJob = getCurrentJob(dc);
		SessionControl scWrapper = sc.getPluginSessionControl(plugin, currentJob);
		
		IOSupport ioSupport = plugin.newInstance(IOSupport.class);
		ioSupport.init(scWrapper, configWrapper, jobWrapper, this);
		return ioSupport;		
	}
	
	/* 
	 	Gets the parameters needed to configure the plug-in.	 	
	 	* Path to external plug-in
	 	* Which exporter to use
	 	* Which importer to use	 	
	*/
	private RequestInformation getConfigurePluginParameters()
	{
		if (configurePlugin == null)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
				
				// The path to the external program
				parameters.add(new PluginParameter<String>(
						PARAMETER_EXTERNAL_PATH,
						"External program path",
						"Enter the complete absolute path to the external program that " +
						"should be executed.",
						new StringParameterType(255, null, true))
					);

				// Command line options
				parameters.add(new PluginParameter<String>(
						PARAMETER_EXTERNAL_CMDLINE,
						"Command line options",
						"Specify any command line options for the external program.",
						new StringParameterType(255, null, false))
					);
				
				// Exporter and importer -- must implement ExternalExecutorIOSupport
				ItemQuery<PluginDefinition> pluginQuery = PluginDefinition.getQuery();
				pluginQuery.include(Include.MINE, Include.IN_PROJECT, Include.SHARED, Include.OTHERS);
				pluginQuery.join(Hql.innerJoin("pluginTypes", "pt"));
				pluginQuery.restrict(
						Restrictions.eq(
								Hql.property("pt", "interfaceName"), 
								Expressions.string(IOSupport.class.getName())
					));
				Enumeration<String, String> exporters = new Enumeration<String, String>();		
				Enumeration<String, String> importers = new Enumeration<String, String>();
				for (PluginDefinition plugin : pluginQuery.list(dc))
				{
					Plugin.MainType pluginType = plugin.getMainType();
					IOSupport ioSupport = plugin.newInstance(IOSupport.class);
					if (pluginType == Plugin.MainType.EXPORT)
					{
						exporters.add(plugin.getClassName(), ioSupport.getShortName());
					}
					if (pluginType == Plugin.MainType.IMPORT)
					{
						importers.add(plugin.getClassName(), ioSupport.getShortName());
					}
				}
				parameters.add(new PluginParameter<String>(
					PARAMETER_EXPORTER_CLASS,
					"Export plug-in",
					"Select which export plug-in to use for exporting data to the external " +
					"program. In the next step you can configure the selected plug-in.",
					new StringParameterType(PluginDefinition.MAX_CLASSNAME_LENGTH, null, true, 1, 0, 0, exporters))
				);
				
				parameters.add(new PluginParameter<String>(
					PARAMETER_IMPORTER_CLASS,
					"Import plug-in",
					"Select which import plug-in to use for importing data from the external " +
					"program. In the next step you can configure the selected plug-in.",
					new StringParameterType(PluginDefinition.MAX_CLASSNAME_LENGTH, null, true, 1, 0, 0, importers))
				);
				
				parameters.add(new PluginParameter<Boolean>(
					PARAMETER_DEBUG_MODE,
					"Allow debug mode",
					"In debug mode the plug-in doesn't delete some temporary working files used " +
					"by the external program. This is useful for locating export/import problems.",
					new BooleanParameterType()
				));
				
				parameters.add(new PluginParameter<String>(
					"prerequisitesSection", 
					"Prerequisites", 
					"In this section you can set some prerequisites for the data that the " +
					"external program needs.", 
					null)
				);

				// Raw data type
				Enumeration<String, String> rawDataTypes = new Enumeration<String, String>();
				for (RawDataType rdt : RawDataTypes.getRawDataTypes())
				{
					if (rdt.isStoredInDb())
					{
						rawDataTypes.add(rdt.getId(), rdt.getName());
					}
				}
				rawDataTypes.sortValues();
				parameters.add(new PluginParameter<String>(
						PARAMETER_RAW_DATA_TYPE,
						"Raw data type",
						"The raw data type of the experiment that is required for the " +
						"external program to work correctly. If not specified the program " +
						"can work with any kind of data.",
						new StringParameterType(255, null, false, 1, 0, 0, rawDataTypes)
					));
				
				// Number of channels
				parameters.add(new PluginParameter<Integer>(
						PARAMETER_NUM_CHANNELS,
						"Number of channels",
						"The number of channels the experiment must have for the external " +
						"program to work correctly. This parameter is only used if no raw data " +
						"type has been selected. If not specified the program can work with " +
						"any number of channels.",
						new IntegerParameterType(1, null, null, false)
					));
				
				// Intensity transform of the source bioassay set
				Enumeration<String, String> transforms = new Enumeration<String, String>();
				for (IntensityTransform t : IntensityTransform.values())
				{
					transforms.add(t.name(), t.toString());
				}
				parameters.add(new PluginParameter<String>(
						PARAMETER_SOURCE_TRANSFORM,
						"Source intensities",
						"The type of spot intensity values that the source bioassay set must have " +
						"for the external to work correctly. If not specified the program can work " +
						"with any kind of data.",
						new StringParameterType(255, null, false, 1, 0, 0, transforms)
					));
				
				configurePlugin = new RequestInformation(
						Request.COMMAND_CONFIGURE_PLUGIN, 
						"Select program, importer/exporter and prerequisites", 
						"Step 1 of 3. Configure which external program to use " +
						"and in which format the data is transferred between it " +
						"and BASE. Specify requirements on the experiment that must " +
						"be met for the external program to work.", 
						parameters
						); 	
			}
			finally
			{
				if (dc != null) dc.close();
			}			
		}		
		return configurePlugin;
	}
	
	/*
	 	Configure the parameters needed by the export-plugin and import-plugin.
	 */
	private RequestInformation getConfigureExportImportPluginParameters()
	{
		DbControl dc = null;
		try
		{
			if (configureExportImportPluginParameters == null)
			{
				dc = sc.newDbControl();
	
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
				IOSupport exporter = getExporter(dc, false, false);
				RequestInformation riExporter = 
					exporter.getRequestInformation(null, IOSupport.COMMAND_CONFIGURE_IO_PLUGIN);
				if (riExporter != null)
				{
					parameters.add(new PluginParameter<String>(
							"exportSection", 
							"Export parameters", 
							"This section contains parameters for the selected exporter plug-in: " + 
							exporter.getShortName() + "\n" + 
							riExporter.getDescription(), 
							null
						));
					addParametersWithNamespace(parameters, riExporter.getParameters(), NAMESPACE_EXPORTER);
				}
				
				IOSupport importer = getImporter(dc, false, false);
				RequestInformation riImporter = 
					importer.getRequestInformation(null, IOSupport.COMMAND_CONFIGURE_IO_PLUGIN);
				if (riImporter != null)
				{
					parameters.add(new PluginParameter<String>(
							"importSection", 
							"Import parameters", 
							"This section contains parameters for the selected importer plug-in: " + 
							importer.getShortName() + "\n" + 
							riImporter.getDescription(), 
							null
						));
					addParametersWithNamespace(parameters, riImporter.getParameters(), NAMESPACE_IMPORTER);
				}

				configureExportImportPluginParameters = new RequestInformation(
						IOSupport.COMMAND_CONFIGURE_IO_PLUGIN,
						"Configure export and import plug-ins",
						"Configure the plugin parameters for the exporter and the importer",
						parameters);			
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureExportImportPluginParameters;
	}
	
	/*
	 	Create the parameters that should be used with the external plugin.
	 */
	private RequestInformation getConfigureExternalParameters()
	{
		if (configureExternalParameters == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(new PluginParameter<String>(
				PARAMETER_EXTERNAL_PARAMETERS,
				"Parameter definition",
				"This holds and XML string with parameter definitions.",
				new StringParameterType(65536, null, false, 0, 0, 0))
			);
			
			configureExternalParameters = new RequestInformation(
				ExternalProgramExecutor.COMMAND_CONFIGURE_EXTERNAL_PARAMETERS,
				"External program parameters",
				"Define parameters for the external program.",
				parameters,
				"external_plugin_parameters.jsp"
			);
		}
		return configureExternalParameters;
	}
	
	/*
	 	Configure the job parameters
	 */
	private RequestInformation getConfigureJobParameters()
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				dc = sc.newDbControl();
				BioAssaySet source = getCurrentBioAssaySet(dc);
				
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
				
				// Source bioassay set
				parameters.add(getSourceBioAssaySetParameter(null, null));
				
				if (Boolean.TRUE.equals(configuration.getValue(PARAMETER_DEBUG_MODE)))
				{
					parameters.add(new PluginParameter<Boolean>(
						PARAMETER_DEBUG_MODE,
						"Debug mode",
						"In debug mode the plug-in doesn't delete some temporary working files used " +
						"by the external program. This is useful for locating export/import problems.",
						new BooleanParameterType(true, false)
					));
				}
				
				// Parameters for the external program
				List<PluginParameter<?>> externalParameters = getExernalPluginParameters(dc, source);
				if (externalParameters.size() > 0)
				{
					parameters.add(new PluginParameter<String>(
						"externalParameters", 
						"Program parameters", 
						"This section contains all parameters to the external program.", 
						null)
					);
					addParametersWithNamespace(parameters, externalParameters, NAMESPACE_EXTERNAL);
				}
				
				// Parameters for the exporter
				IOSupport exporter = getExporter(dc, false, false);
				RequestInformation riExporter = 
					exporter.getRequestInformation(null, IOSupport.COMMAND_CONFIGURE_IO_JOB);
				if (riExporter != null)
				{
					parameters.add(new PluginParameter<String>(
							"exportSection", 
							"Export parameters", 
							"This section contains parameters for the selected exporter plug-in: " + 
							exporter.getShortName() + "\n" + 
							riExporter.getDescription(), 
							null
						));
					addParametersWithNamespace(parameters, riExporter.getParameters(), NAMESPACE_EXPORTER);
				}
				
				// Parameters for the importer
				IOSupport importer = getImporter(dc, false, false);
				RequestInformation riImporter = 
					importer.getRequestInformation(null, IOSupport.COMMAND_CONFIGURE_IO_JOB);
				if (riImporter != null)
				{
					parameters.add(new PluginParameter<String>(
							"importSection", 
							"Import parameters", 
							"This section contains parameters for the selected importer plug-in: " + 
							importer.getShortName() + "\n" + 
							riImporter.getDescription(), 
							null
						));
					addParametersWithNamespace(parameters, riImporter.getParameters(), NAMESPACE_IMPORTER);
				}
				
				configureJob = new RequestInformation(
						Request.COMMAND_CONFIGURE_JOB,
						"Configure job",					
						"Configure the parameters used to run the external program. " +
						"These parameters will be exported by the exporter to a file that " +
						"the program can access.",
						parameters);			
	
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	

	/*
	 	Get the external parameters as plug-in parameters.
	*/
	private List<PluginParameter<?>> getExernalPluginParameters(DbControl dc, BioAssaySet source)
	{
		List<ParameterDefinition> parameters = getParameterDefinitions(true);
		List<PluginParameter<?>> pluginParameters = 
			new ArrayList<PluginParameter<?>>(parameters.size());
		for (ParameterDefinition pdef : parameters)
		{
			pluginParameters.add(pdef.toPluginParameter(dc, source));
		}
		return pluginParameters;		
	}
	
	/**
		Add more parameters to a list of parameters with a different
		namespace.
		@param to The parameters list to add the parameters to
		@param from The parameters list to copy parameters from
		@param namespace The namespace to use for the added parameter
	*/
	@SuppressWarnings("unchecked")
	private void addParametersWithNamespace(List<PluginParameter<?>> to, 
			List<PluginParameter<?>> from, String namespace)
	{
		for (PluginParameter pp : from)
		{
			to.add(new NamespacePluginParameter(pp, namespace));
		}
	}
	

	static class RequestWrapper
		implements Request
	{

		private final String command;
		private final Request parent;
		private final Map<String, Object> parameters;
		
		RequestWrapper(String command, Request parent)
		{
			this.command = command;
			this.parent = parent;
			this.parameters = new HashMap<String, Object>();
		}
		
		void setParameter(String key, Object value)
		{
			parameters.put(key, value);
		}
		
		@Override
		public String getCommand()
		{
			return command;
		}

		@Override
		public Object getParameterValue(String name) 
			throws ParameterException
		{
			if (parameters.containsKey(name))
			{
				return parameters.get(name);
			}
			return parent.getParameterValue(name);
		}

		@Override
		public List<?> getParameterValues(String name)
		{
			return parent.getParameterValues(name);
		}

		@Override
		public boolean isAllowedImmediateExecution()
		{
			return false;
		}
		
	}
	
	static class ResponseWrapper
		implements Response
	{

		private String message;
		private List<Throwable> errors;
		private boolean hasError;
		
		ResponseWrapper()
		{
			this.errors = new ArrayList<Throwable>();
		}
		
		boolean hasError()
		{
			return hasError;
		}
		
		List<Throwable> getErrors()
		{
			return errors;
		}
		
		String getMessage()
		{
			return message;
		}
		
		String getMessage(String defaultMessage)
		{
			return message == null ? defaultMessage : message;
		}
		
		@Override
		public void setContinue(String nextCommand)
			throws PermissionDeniedException
		{
			setError("Not allowed to continue with command: " + nextCommand, null);
		}

		@Override
		public void setDone(String message, ExecutionTime estimatedExecutionTime)
		{
			if (hasError) return;
			this.message = message;
		}

		@Override
		public void setDone(String message)
		{
			if (hasError) return;
			this.message = message;
		}

		@Override
		public void setDownloadImmediately(String message,	ExecutionTime estimatedExecutionTime, boolean failIfDenied)
			throws PermissionDeniedException
		{
			setError("Not allowed to download immediately", null);
		}

		@Override
		public void setDryRun(boolean dryRun)
		{}

		
		@Override
		public void setSuggestedJobName(String jobName)
		{}
		
		@Override
		public void setError(String message, List<? extends Throwable> errorList)
		{
			if (hasError && this.message != null)
			{
				message = this.message + "\n" + message;
			}
			this.message = message;
			this.hasError = true;
			if (errorList != null) errors.addAll(errorList);
		}

		@Override
		public void setExecuteImmediately(String message, ExecutionTime estimatedExecutionTime, boolean failIfDenied)
			throws PermissionDeniedException
		{
			setError("Not allowed immediate execution", null);
		}
		
	}

}
